#ifndef AVERAGEINTERCLUSTERDISTANCE_H
#define AVERAGEINTERCLUSTERDISTANCE_H

#include <cmath>

#include "../base/dissimilaritymeasure.h"
#include "../base/euclideanspaceprovider.h"
#include "../clustering/cfentry.h"

namespace CluE
{
/**
 * @brief Calculates the average inter-cluster distance D2 of a given CFEntry.
 * @sa T. Zhang, R. Ramakrishan and M. Livny. "BIRCH: A New Data Clustering Algorithm and Its Applications", p. 146. <em>Data Mining and Knowledge Discovery</em>, 10.1023/A:1009783824328, 1997.
 */
template<typename T> class CFAverageInterClusterDistance : public DissimilarityMeasure<CFEntry<T> >
{
public:
	CFAverageInterClusterDistance(EuclideanSpaceProvider<T>* esp);
	CFAverageInterClusterDistance(const CFAverageInterClusterDistance<T>&);
	CFAverageInterClusterDistance<T>& operator= (const CFAverageInterClusterDistance<T>&);
	virtual ~CFAverageInterClusterDistance();
	
	virtual CFAverageInterClusterDistance* clone() const;

	/**
	 * @brief Computes the average inter cluster distance between the two given CFEntry objects.
	 */
	virtual double dissimilarity(CFEntry<T> const&, CFEntry<T> const&) const;
	
private:
	EuclideanSpaceProvider<T>* esp;
};

template<typename T> CFAverageInterClusterDistance<T>::CFAverageInterClusterDistance(EuclideanSpaceProvider<T>* esp) :
	esp(esp->clone())
{
}

template<typename T> CFAverageInterClusterDistance<T>::CFAverageInterClusterDistance(const CFAverageInterClusterDistance<T>& aicd) :
	esp(aicd.esp->clone())
{
}

template<typename T> CFAverageInterClusterDistance<T>& CFAverageInterClusterDistance<T>::operator= (const CFAverageInterClusterDistance<T>& aicd)
{
	if(esp != 0)
		delete esp;

	DissimilarityMeasure<CFEntry<T> >::operator= (aicd);

	esp = aicd.esp == 0 ? 0 : aicd.esp->clone();

	return *this;
}

template<typename T> CFAverageInterClusterDistance<T>::~CFAverageInterClusterDistance()
{
	if(esp != 0)
		delete esp;
}

template<typename T> CFAverageInterClusterDistance<T>* CFAverageInterClusterDistance<T>::clone() const
{
	return new CFAverageInterClusterDistance<T>(*this);
}

template<typename T> double CFAverageInterClusterDistance<T>::dissimilarity(CFEntry<T> const& cf1, CFEntry<T> const& cf2) const
{	
	double numerator = (cf2.number*cf1.SS - 2*(cf1.LS*cf2.LS) + cf1.number*cf2.SS);
	double denominator = cf1.number * cf2.number;
	return std::pow(numerator / denominator, 0.5);
}

}

#endif