#ifndef AGGLOMERATIVE_H
#define AGGLOMERATIVE_H

#include "../base/algorithm.h"
#include "../base/clusterdissimilaritymeasure.h"
#include "../base/solutionprovider.h"
#include "../base/inputsetter.h"
#include "../base/clustermeasuresetter.h"
#include "../datastructure/partitionsolution.h"
#include "../exception/invalidruntimeconfigurationexception.h"

#include <vector>
#include <limits>

namespace CluE
{

/**
 * @brief Agglomerative clustering algorithm
 * 
 * @ingroup clustering_algorithms
 */
template<typename T> class Agglomerative : public Algorithm, public InputSetter<T>, public ClusterMeasureSetter<T>
{
public:
	/**
	 * @brief On demand computation / pre-computation
	 * 
	 * This enum is used to tell the algorithm via the constructor if the dissimilarities should be
	 * computed on demand everytime they are needed or if they should be precomputed, held in memory
	 * and be updated whenever a dissimilarity may have changed.
	 */
	enum
	{
		OnDemand, // slow, needs no extra memory
		Precompute // fast, needs more memory
	};

	Agglomerative(std::vector<T*> const* data, ClusterDissimilarityMeasure<T>* measure = NULL,
		int calculationTime = OnDemand);

	Agglomerative(const Agglomerative<T>&);
	Agglomerative<T>& operator=(const Agglomerative<T>&);
	virtual ~Agglomerative();

	/**
	 * @throw InvalidRuntimeException [0] Input is NULL.
	 * @throw InvalidRuntimeException [1] Dissimilarity measure is NULL.
	 */
	virtual PartitionSolution<T>* compute();
	
	virtual void setInput(std::vector<T*> const*);
	virtual void setMeasure(ClusterDissimilarityMeasure<T> const*);
	void setMode(int);

	/**
	* @brief does a dynamic cast of the given Algorithm to Agglomerative<T> 
	* @return NULL if the Algorithm is not a Agglomerative<T> instance
	*/
	static Agglomerative<T>* toAgglomerative(Algorithm* s);	
	
private:
	std::vector<T*> const* input;
	ClusterDissimilarityMeasure<T>* measure;
	int mode;
};

template<typename T> Agglomerative<T>::Agglomerative(std::vector<T*> const* data,
	ClusterDissimilarityMeasure<T>* m, int cm) : input(data),
	measure(m==NULL?NULL:m->clone()), mode(cm)
{
}

template<typename T> Agglomerative<T>::Agglomerative(const Agglomerative<T>& rhs) :	Algorithm(rhs),
	input(rhs.input), measure(rhs.measure==NULL?NULL:rhs.measure->clone()),	mode(rhs.mode)
{	
}

template<typename T> Agglomerative<T>& Agglomerative<T>::operator=(
	const Agglomerative& rhs)
{
	Algorithm::operator=(rhs);
	DissimilarityMeasure<T>* dm = rhs.measure==NULL?NULL:rhs.measure->clone();
	delete this->measure;
	this->measure = dm;
	this->input = rhs.input;
	this->mode = rhs.mode;
	return *this;
}

template<typename T> Agglomerative<T>::~Agglomerative()
{
	delete this->measure;
}

template<typename T> PartitionSolution<T>* Agglomerative<T>::compute()
{
	if (this->input==NULL)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if (this->measure==NULL)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL");
	
	clock_t start, end;
	start = clock();

	PartitionSolution<T>* solution = new PartitionSolution<T>();

	unsigned int N = this->input->size();

	// create a first partition (partitions[0]) with one cluster for input data object
	std::clog << "CluE::Agglomerative<T>::compute() - starting with " << N << " singleton clusters..." << std::endl;
	std::vector<std::vector<T*> > partitionzero;
	for (unsigned int i=0; i<N; i++)
	{
		std::vector<T*> singleton;
		singleton.push_back((*this->input)[i]);
		partitionzero.push_back(singleton);
	}
	solution->partitions.push_back(partitionzero);

	// precompute dissimilarities
	double* dis = NULL;

	if (this->mode==Precompute)
	{
		std::clog << "CluE::Agglomerative<T>::compute() - precomputing dissimilarities..." << std::endl;
		dis = new double[N*N];

		for (unsigned int i=0; i<N; i++)
			for (unsigned int j=0; j<N; j++)
				dis[N*i+j] = measure->dissimilarity(solution->partitions[0][i], solution->partitions[0][j]);
	}

	for (unsigned int n=1; n<N; n++)
	{
		std::clog << "CluE::Agglomerative<T>::compute() - currently computing round " << n << std::endl;
		// find the two clusters with minimum dissimilarity
		int first = -1;
		int second = -1;
		double min = std::numeric_limits<double>::infinity();
		for (unsigned int i=0; i<N-n+1; i++)
			for (unsigned int j=0; j<N-n+1; j++)
			{
				//cout << "i=" << i << ", j=" << j << std::endl;
				if (i!=j)
				{
					double d;
					if (this->mode==Precompute)
						d = dis[N*i+j];
					else
						d = measure->dissimilarity(solution->partitions[n-1][i], solution->partitions[n-1][j]);

					if (d < min)
					{
						min = d;
						first = i;
						second = j;
					}
				}
			}

		if (second<first)
		{
			int i = first;
			first = second;
			second = i;
		}
		// cout << "merging cluster " << first << " and " << second << std::endl;

		// merge found clusters at next granularity (merged cluster at index first and last cluster moved to index second)
		solution->partitions.push_back(solution->partitions[n-1]);
		while (!solution->partitions[n][second].empty())
		{
			solution->partitions[n][first].push_back(solution->partitions[n][second].back());
			solution->partitions[n][second].pop_back();
		}
		solution->partitions[n][second] = solution->partitions[n].back();
		solution->partitions[n].pop_back();

		if (this->mode==Precompute)
		{
			// update dissimilarities
			for (unsigned int i=0; i<N-n; i++)
			{
				dis[N*second+i] = dis[N*(N-n)+i];
				dis[N*i+second] = dis[N*i+(N-n)];
			}

			for (unsigned int i=0; i<N-n; i++)
			{
				dis[N*first+i] = this->measure->dissimilarity(solution->partitions[n][first], solution->partitions[n][i]);
				dis[N*i+first] = this->measure->dissimilarity(solution->partitions[n][i], solution->partitions[n][first]);
			}
		}
	}

	end = clock();
	solution->seconds=double(end-start)/CLOCKS_PER_SEC;
	std::clog << "CluE::Agglomerative<T>::compute() - finished" << std::endl;
	return solution;
}

template<typename T> void Agglomerative<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void Agglomerative<T>::setMeasure(ClusterDissimilarityMeasure<T> const* m)
{
	this->measure = m==NULL?NULL:m->clone();
}

template<typename T> void Agglomerative<T>::setMode(int m)
{
	this->mode = m;
}

template<typename T> Agglomerative<T>* Agglomerative<T>::toAgglomerative(Algorithm* s)
{
	return dynamic_cast<Agglomerative<T>*>(s);
}
			
}

#endif
