#ifndef BALCANBLUMGUPTA_H
#define BALCANBLUMGUPTA_H

#include <map>
#include <lemon/list_graph.h>

#include "../base/algorithm.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../datastructure/partitionsolution.h"

#include "../exception/invalidruntimeconfigurationexception.h"
#include "../exception/invalidargumentexception.h"

namespace CluE
{	
/**
 * @brief BalcanBlumGupta algorithm.
 *
 * Approximate clustering algorithm as described in
 * M-F. Balcan, A. Blum and A. Gupta. "Approximate Clustering without the Approximation". <em>Proceedings of the twentieth Annual ACM-SIAM Symposium on Discrete Algorithms (SODA '09)</em>, 2009.
 *
 * @ingroup clustering_algorithms
 */
template<typename T> class BalcanBlumGupta : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:

	BalcanBlumGupta(DissimilarityMeasure<T> const *measure = 0, std::vector<T*> const* input = 0, unsigned int numOfPartitons = 1);

	BalcanBlumGupta(const BalcanBlumGupta<T>&);
	BalcanBlumGupta<T>& operator= (const BalcanBlumGupta<T>&);
	virtual ~BalcanBlumGupta();
	
	/**
	 * @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	 * @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	 * @throw InvalidRuntimeConfigurationException [2] Threshold graph: Empty node set.
	 */
	virtual PartitionSolution<T>* compute();
	
	virtual void setInput(std::vector<T*> const*);
	
	virtual void setMeasure(DissimilarityMeasure<T> const *measure);
	
	/**
	 * @sa "Approximate Clustering without the Approximation"
	 * @throw InvalidArgumentException [0] The given value of alpha is invalid. No changes were made.
	 */
	void setAlpha(double alpha);

	/**
	 * @sa "Approximate Clustering without the Approximation"
	 * @throw InvalidArgumentException [0] The given value of beta is invalid. No changes were made.
	 */
	void setBeta(double beta);
	
	/**
	 * @sa "Approximate Clustering without the Approximation"
	 * @throw InvalidArgumentException [0] The given value of epsilon is invalid. No changes were made.
	 */
	void setEpsilon(double epsilon);
	
	/**
	 * @sa "Approximate Clustering without the Approximation"
	 */
	void setOmega(double omega);
	
	/**
	 * @brief Sets the desired number of partitions.
	 * @sa "Approximate Clustering without the Approximation"
	 */
	void setNumberOfPartitions(unsigned int number);
	
	/**
	* @brief does a dynamic cast of the given Algorithm to BalcanBlumGupta<T>
	* @return NULL if the Algorithm is not a BalcanBlumGupta<T> instance
	*/
	static BalcanBlumGupta<T>* toBalcanBlumGupta(Algorithm* s);
	
private:
	DissimilarityMeasure<T> *measure;
	std::vector<T*> const *input;
	unsigned int numPartitions;
	double alpha;
	double beta;
	double epsilon;
	double omega;
};

template<typename T> BalcanBlumGupta<T>::BalcanBlumGupta(DissimilarityMeasure<T> const *measure, std::vector<T*> const* input, unsigned int numOfPartitons) :
	measure(measure==0 ? 0 : measure->clone()),
	input(input),
	numPartitions(numOfPartitons),
	alpha(1),
	beta(4),
	epsilon(1),
	omega(1)	
{
}

template<typename T> BalcanBlumGupta<T>::BalcanBlumGupta(const BalcanBlumGupta<T>& bbg) :
	measure(bbg.measure==0 ? 0 : bbg.measure->clone()),
	input(bbg.input),
	numPartitions(bbg.numPartitions),
	alpha(bbg.alpha),
	beta(bbg.beta),
	epsilon(bbg.epsilon),
	omega(1)
{
}

template<typename T> BalcanBlumGupta<T>& BalcanBlumGupta<T>::operator= (const BalcanBlumGupta<T>& bbg)
{
	if(measure != 0)
		delete measure;

	Algorithm::operator= (bbg);
	
	measure = bbg.measure == 0 ? 0 : bbg.measure->clone();
	input = bbg.input;
	epsilon = bbg.epsilon;
	alpha = bbg.alpha;
	beta = bbg.beta;
	numPartitions = bbg.numPartitions;
	omega = bbg.omega;
	
	return *this;
}

template<typename T> BalcanBlumGupta<T>::~BalcanBlumGupta()
{
	if(measure != 0)
		delete measure;
}

template<typename T> PartitionSolution<T>* BalcanBlumGupta<T>::compute()
{
	using namespace lemon;
	
	if(this->input == 0)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if(this->measure == 0)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");
	
	size_t inputSize = input->size();
	double tau = (alpha * omega) / (5 * beta * epsilon);
	
	//Step 1
	ListGraph threshGraph;
	ListGraph::NodeMap<T*> tMap(threshGraph);
	std::map<T*, ListGraph::Node> tNodeMap;
	for(size_t i = 0; i < inputSize; i++)
	{
		ListGraph::Node u = threshGraph.addNode();
		tMap[u] = (*input)[i];
		tNodeMap[(*input)[i]] = u;
	}
	for(size_t i = 0; i < inputSize; i++)
	{
		for(size_t j = i+1; j < inputSize; j++)
		{
			T* p = (*input)[i];
			T* q = (*input)[j];
			if(measure->dissimilarity(*p, *q) < tau)
				threshGraph.addEdge(tNodeMap[p], tNodeMap[q]);
		}
	}
	
	//Step 2
	PartitionSolution<T> *solution = new PartitionSolution<T>();
	solution->partitions.push_back(std::vector<std::vector<T*> >());
	for(size_t i = 0; i < numPartitions-1; i++)
	{
		solution->partitions[0].push_back(std::vector<T*>());
		std::vector<T*>& partition = solution->partitions[0].back();
		
		int maxEdges = 0;
		ListGraph::Node maxNode = INVALID;
		for(ListGraph::NodeIt u(threshGraph); u != INVALID; ++u)
		{
			int edges = countIncEdges(threshGraph, u);
			if(edges > maxEdges)
			{
				maxEdges = edges;
				maxNode = u;
			}
		}
		if(maxNode == INVALID)
			throw InvalidRuntimeConfigurationException(2, "Threshold graph: Empty node set.");
		
		std::vector<ListGraph::Node> nodesToErase;
		for(ListGraph::IncEdgeIt e(threshGraph, maxNode); e != INVALID; ++e)
		{
			ListGraph::Node n;
			if(threshGraph.u(e) != maxNode)
				n = threshGraph.u(e);
			else
				n = threshGraph.v(e);
			partition.push_back(tMap[n]);
			nodesToErase.push_back(n);
		}
		partition.push_back(tMap[maxNode]);
		for(std::vector<ListGraph::Node>::iterator it = nodesToErase.begin(); it != nodesToErase.end(); it++)
			threshGraph.erase(*it);
		threshGraph.erase(maxNode);
	}
	
	solution->partitions[0].push_back(std::vector<T*>());
	std::vector<T*>& partition = solution->partitions[0].back();
	for(ListGraph::NodeIt u(threshGraph); u != INVALID; ++u)
		partition.push_back(tMap[u]);
	
	return solution;
}

template<typename T> void BalcanBlumGupta<T>::setNumberOfPartitions(unsigned int numOfPartitions)
{
	this->numPartitions = numOfPartitions;
}

template<typename T> void BalcanBlumGupta<T>::setAlpha(double alpha)
{
	if(alpha > 0)
		this->alpha = alpha;
	else
		throw InvalidArgumentException(0, "The given value of alpha is invalid. No changes were made.", "alpha");
}

template<typename T> void BalcanBlumGupta<T>::setBeta(double beta)
{
	if(beta > 0)
		this->beta = beta;
	else
		throw InvalidArgumentException(0, "The given value of beta is invalid. No changes were made.", "beta");
}

template<typename T> void BalcanBlumGupta<T>::setEpsilon(double epsilon)
{
	if(epsilon > 0)
		this->epsilon = epsilon;
	else
		throw InvalidArgumentException(0, "The given value of epsilon is invalid. No changes were made.", "epsilon");
}

template<typename T> void BalcanBlumGupta<T>::setOmega(double omega)
{
	this->omega = omega;
}

template<typename T> void BalcanBlumGupta<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void BalcanBlumGupta<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(this->measure != 0)
		delete this->measure;
	
	if(measure == 0)
		this->measure = 0;
	else
		this->measure = measure->clone();
}

template<typename T> BalcanBlumGupta<T>* BalcanBlumGupta<T>::toBalcanBlumGupta(Algorithm* s)
{
	return dynamic_cast<BalcanBlumGupta<T>*>(s);
}

}
#endif
