#ifndef BIRCH_H
#define BIRCH_H

#include <lemon/list_graph.h>

#include "../base/streamingalgorithm.h"
#include "../base/inputsetter.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/solutionprovider.h"
#include "../base/partitionprovider.h"
#include "../datastructure/partitionsolution.h"
#include "../clustering/cftree.h"
#include "../clustering/birchconfig.h"

namespace CluE
{

/**
 * @brief BIRCH clustering algorithm
 *
 * Simple adaption of CFTree realizing the BIRCH algorithm as described in
 * T. Zhang, R. Ramakrishan and M. Livny. "BIRCH: A New Data Clustering Algorithm and Its Applications". <em>Data Mining and Knowledge Discovery</em>, 10.1023/A:1009783824328, 1997.
 *
 * @warning Meet requirements of CFTree!
 *
 * @ingroup clustering_algorithms
 */
template<typename T> class Birch : public StreamingAlgorithm<T>
{
public:
	Birch(BirchConfig<T> const * config);

	virtual PartitionSolution<T>* compute();
	virtual Birch<T>& operator<<(T const & element);

private:
	class ThreshFuncRedirector;

	BirchConfig<T> const * config;
	ThreshFuncRedirector tfr;
	CFTree<T> cft;

  /**
   * @brief Wraps around a BirchConfig object to provide threshold calculation
   */
	class ThreshFuncRedirector : public CFTree<T>::ThresholdCalculator
	{
	public:
		ThreshFuncRedirector(BirchConfig<T> const * config) : config(config) { }

		virtual double operator()(double t)
		{
			return config->calcNewThreshold(t);
		}

	private:
		BirchConfig<T> const * config;
	};
};

template<typename T> Birch<T>::Birch(BirchConfig<T> const * config) :
	config(config),
	tfr(config),
	cft(config->euclidianProvider, config->clusterDistanceMeasure, config->thresholdAttribute, &tfr,
	    config->innerBranching, config->leafBranching, config->threshold, config->maxSize)
{
}

template<typename T> PartitionSolution<T>* Birch<T>::compute()
{
	//Phase 2

	//Phase 3
	std::vector<CFEntry<T> > entries = cft.leafEntries();
	std::vector<CFEntry<T>*> entriesPointer;
	for(typename std::vector<CFEntry<T> >::iterator it = entries.begin(); it != entries.end(); ++it)
		entriesPointer.push_back(&(*it));
	Algorithm* alg3 = config->phase3(&entriesPointer);

	SolutionProvider* alg3sol = alg3->compute();
	PartitionProvider<CFEntry<T> >* alg3part = dynamic_cast<PartitionProvider<CFEntry<T> >*>(alg3sol);
	std::vector<std::vector<CFEntry<T>*> > partitions = alg3part->clustering(0);

	std::vector<T*> features;
	for(typename std::vector<std::vector<CFEntry<T>*> >::iterator it1 = partitions.begin(); it1 != partitions.end(); ++it1)
	{
		CFEntry<T>* cog = new CFEntry<T>(0, config->euclidianProvider->nullVector(), 0);
		for(typename std::vector<CFEntry<T>*>::iterator it2 = it1->begin(); it2 != it1->end(); ++it2)
			*cog += **it2;
		features.push_back(new T((1/cog->number) * cog->LS));
	}

	//Phase 4
	Algorithm* alg4 = config->phase4(&features);

	SolutionProvider* alg4sol = alg4->compute();
	PartitionProvider<T>* alg4part = dynamic_cast<PartitionProvider<T >*>(alg4sol);

	//Prepare result
	std::vector<std::vector<std::vector<T*> > > solvec;
	unsigned int max = alg4part->number_of_solutions();
	for(unsigned int i = 0; i < max; i++)
		solvec.push_back(alg4part->clustering(i));
	PartitionSolution<T>* sol = new PartitionSolution<T>();
	sol->partitions = solvec;

	//Clean up
	for(typename std::vector<T*>::iterator it = features.begin(); it != features.end(); ++it)
		delete *it;
	delete alg3;
	delete alg4;
	delete alg3sol;
	delete alg4sol;

	return sol;
}

template<typename T> Birch<T>& Birch<T>::operator<<(T const & element)
{
	cft.insert(element);
}

}

#endif
