#ifndef BIRCHCONFIG_H
#define BIRCHCONFIG_H

#include "../base/algorithm.h"
#include "../clustering/cfentry.h"

namespace CluE
{

/**
 * @brief BIRCH configuration class
 * 
 * Derive a class from BirchConfig to specify the phase 3 and 4 clustering algorithms
 * and other BIRCH parameters.
 */
template<typename T> struct BirchConfig
{
	/**
	 * @brief Returns a ready-to-compute Algorithm object of the phase 3 algorithm.
	 * 
	 * Expected to derive from: PartitionProvider<CFEntry<T> >
	 */
	virtual Algorithm* phase3(std::vector<CFEntry<T>*> const * input) const = 0;
	
	/**
	 * @brief Returns a ready-to-compute Algorithm object of the phase 4 algorithm.
	 * 
	 * Expected to derive from: PartitionProvider<T>
	 */
	virtual Algorithm* phase4(std::vector<T*> const * input) const = 0;
	
	/**
	 * @brief Returns the new threshold for rebuilding the clustering feature tree.
	 */
	virtual double calcNewThreshold(double oldThreshold) const = 0;
	
  /**
   * @brief Euclidean space provider
   */
	EuclideanSpaceProvider<T> * euclidianProvider;
  
  /**
   * @brief Cluster distance (e.g. average inter-cluster distance) calculator
   */
	DissimilarityMeasure<CFEntry<T> > *clusterDistanceMeasure;
  
  /**
   * @brief Cluster size (e.g. diameter) calculator
   */
	AttributeCalculator<CFEntry<T> > *thresholdAttribute;
  
  /**
   * @brief Maximum number of siblings of inner nodes
   */
	int innerBranching;
  
  /**
   * @brief Maximum number of siblings of leaf nodes
   */
	int leafBranching;
	
  /**
   * @bried Node threshold
   */
  double threshold;
  
  /**
   * @brief Maximum number of nodes
   */
	int maxSize;
};

}

#endif
