#ifndef BRUTEFORCECLUSTERING_H
#define BRUTEFORCECLUSTERING_H

#include <memory>
#include <utility>

#include "../base/algorithm.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../datastructure/proxysolution.h"
#include "../evaluation/kmeansevaluator.h"

#include "../exception/invalidruntimeconfigurationexception.h"
#include "../exception/invalidargumentexception.h"

namespace CluE
{

/**
 * @brief Brute force k-median / k-means clustering
 * 
 * Selects the best k centers from the input set
 * 
 * @warning Running time is exponential in the size of input!
 */
template<typename T> class BruteForceClustering : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	BruteForceClustering(DissimilarityMeasure<T> const *measure = 0, std::vector<T*> const* input = 0, unsigned int numOfCenters = 1);

	/**
	* @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	* @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	* @throw InvalidRuntimeConfigurationException [2] Input contains not enough elements.
	*/
	virtual ProxySolution<T>* compute();

	virtual void setInput(std::vector<T*> const*);

	virtual void setMeasure(DissimilarityMeasure<T> const *measure);

	/**
	 * @brief Sets the desired number of clusters.
	 */
	void setNumberOfClusters(unsigned int number);

private:
	std::pair<std::vector<size_t>, double> bruteCompute(std::vector<size_t> centers);

	KMeansEvaluator<T> eval;
	std::unique_ptr<DissimilarityMeasure<T>> measure;
	std::vector<T*> const *input;
	unsigned int numClusters;
};

template<typename T> BruteForceClustering<T>::BruteForceClustering(DissimilarityMeasure<T> const *measure, std::vector<T*> const* input, unsigned int numOfCenters) :
	eval(measure),
	measure(measure==0 ? 0 : measure->clone()),
	input(input),
	numClusters(numOfCenters)
{
}

template<typename T> ProxySolution<T>* BruteForceClustering<T>::compute()
{
	if(this->input == 0)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if(this->input->size() < numClusters)
		throw InvalidRuntimeConfigurationException(2, "Input contains not enough elements.");
	if(this->measure == 0)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");

	std::pair<std::vector<size_t>, double> centersInd(bruteCompute(std::vector<size_t>()));
	ProxySolution<T>* solution = new ProxySolution<T>();
	
	std::vector<T> centers;
	centers.reserve(centersInd.first.size());
	for(size_t i = 0; i < centersInd.first.size(); ++i)
		centers.push_back(*(*input)[centersInd.first[i]]);
	solution->proxysets.push_back(centers);

	return solution;
}

template<typename T> std::pair<std::vector<size_t>, double> BruteForceClustering<T>::bruteCompute(std::vector<size_t> centers)
{
	if(centers.size() == numClusters)
	{
		std::vector<T*> centerPointers;
		centerPointers.reserve(centers.size());
		for(size_t i = 0; i < centers.size(); ++i)
			centerPointers.push_back((*input)[centers[i]]);
		return std::pair<std::vector<size_t>, double>(centers, eval.proxycost(*input, centerPointers));
	}
	else
	{
		std::pair<std::vector<size_t>, double> min;
		size_t lowerBound = centers.size()==0 ? 0 : centers[centers.size()-1] + 1;
		size_t upperBound = input->size() - (numClusters-centers.size());
		for(size_t i = lowerBound; i <= upperBound; ++i)
		{
			std::vector<size_t> newCenters(centers);
			newCenters.push_back(i);
			std::pair<std::vector<size_t>, double> tmp(bruteCompute(newCenters));
			if(tmp.second < min.second || i == lowerBound)
				min = tmp;
		}
		return min;
	}
}

template<typename T> void BruteForceClustering<T>::setNumberOfClusters(unsigned int numOfClusters)
{
	this->numClusters = numOfClusters;
}

template<typename T> void BruteForceClustering<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void BruteForceClustering<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(measure == 0)
		this->measure = 0;
	else
		this->measure = std::unique_ptr<DissimilarityMeasure<T>>(measure->clone());
}

}

#endif
