#ifndef CFTREE_H
#define CFTREE_H

#include <iostream>
#include <stack>
#include <utility>

#include <lemon/list_graph.h>

#include "cfentry.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/attributecalculator.h"
#include "../base/euclideanspaceprovider.h"
#include "../exception/invalidargumentexception.h"
#include "../exception/invalidruntimeconfigurationexception.h"

namespace CluE
{

/**
 * @brief Clustering feature tree
 *
 * Clustering tree as described in
 * T. Zhang, R. Ramakrishan and M. Livny. "BIRCH: A New Data Clustering Algorithm and Its Applications". <em>Data Mining and Knowledge Discovery</em>, 10.1023/A:1009783824328, 1997.
 *
 * @note This class expects the following overloaded operators:
 * - + : T x T -> T and +=: (vector) sum
 * - - : T x T -> T and -=: (vector) subtraction
 * - * : double x T -> T: scalar multiplication
 * - * : T x T -> double: dot product
 *
 * @ingroup clustering_algorithms
 */
template<typename T> class CFTree
{
public:
	class ThresholdCalculator;

	/**
	 * @brief Constructs a CFTree for a given vector space
	 * @param euclidianProvider Euclidiean vector space provider
	 * @param clusterDistanceMeasure To calculate the distance between to given CFEntries
	 * @param thresholdAttribute To calculate the value of the threshold attribute of a given CFEntry
	 * @param innerBranching Inner branching factor
	 * @param leafBranching Leaf branching factor
	 * @param threshold CFEntry size threshold
	 * @sa "BIRCH: A New Data Clustering Algorithm and Its Applications", pp. 147-148 (see above)
	 */
	CFTree(EuclideanSpaceProvider<T> * euclidianProvider,
	       DissimilarityMeasure<CFEntry<T> > *clusterDistanceMeasure,
	       AttributeCalculator<CFEntry<T> > *thresholdAttribute,
	       ThresholdCalculator *thresholdFunctor,
	       int innerBranching, int leafBranching, double threshold, int maxSize);

	CFTree(const CFTree<T>&);

	CFTree<T>& operator= (const CFTree<T>&);

	virtual ~CFTree();

	/**
	 * @brief Inserts a point
	 * @sa "BIRCH: A New Data Clustering Algorithm and Its Applications", p. 149 (see above)
	 * @warning Be sure that properly overloaded operators exist! (see note above)
	 */
	void insert(T point);

	/**
	 * @brief Inserts a CFEntry (clustering feature)
	 * @sa "BIRCH: A New Data Clustering Algorithm and Its Applications", p. 149 (see above)
	 * @warning Be sure that properly overloaded operators exist! (see note above)
	 */
	void insert(CFEntry<T> feature);

	/**
	 * @brief Rebuild the tree
	 * @param newThreshold New threshold
	 */
	void rebuild(double newThreshold);

	/**
	 * @brief Returns the leaf entries
	 */
	std::vector<CFEntry<T> > leafEntries();

	/**
	 * @brief Counts the number of nodes
	 */
	int size() const;

	/**
	 * @brief Counts the number of contained points
	 */
	size_t numberOfPoints() const;

	enum ExceptionIds
	{
	    CANNOT_MERGE_NODE_ITSELF,
	    CANNOT_MERGE_NODE_DIFFERENT_PARENTS,
	    CANNOT_MERGE_LEAF_NODES
	};

	/**
	 * @brief Threshold calculation functor class
	 */
	struct ThresholdCalculator : public std::unary_function<double, double>
	{
		virtual double operator()(double oldThreshold) = 0;
	};

#ifdef CLUE_GRAPH
	/**
	 * @brief Write the CFTree as GraphViz source into a stream
	 */
	void print(std::ostream& os);
#endif

private:
	/**
	 * @brief Wrapper to maintain the graph maps
	 */
	lemon::ListDigraph::Node addNode();

	/**
	 * @brief Wrapper to maintain the graph maps
	 */
	void eraseNode(lemon::ListDigraph::Node node);

	/**
	 * @brief Used by rebuild()
	 */
	void rebuildTraverse(lemon::ListDigraph::Node currentOldNode, CFTree<T>& newTree, lemon::ListDigraph::Node currentNewNode);

	/**
	 * @bried Used by leafEntries()
	 */
	void leafEntriesTraverse(lemon::ListDigraph::Node currentNode, std::vector<CFEntry<T> >* vec);

	/**
	 * @brief Remove empty clustering features
	 */
	void clean(lemon::ListDigraph::Node node);

	/**
	 * @brief Splits a node
	 * @param node Node to be splitted
	 * @param keepTrack Return the node containing this CFEntry<T>* as first element
	 * @param fillUp Fill up mode (for Merging Refinements)
	 */
	std::pair<lemon::ListDigraph::Node, lemon::ListDigraph::Node> split(lemon::ListDigraph::Node node, CFEntry<T>* keepTrack = 0, bool fillUp = false);

	/**
	 * @brief Merges two node entries
	 * @throw InvalidArgumentException [CANNOT_MERGE_NODE_ITSELF] Cannot merge node with itself.
	 * @throw InvalidArgumentException [CANNOT_MERGE_NODE_DIFFERENT_PARENTS] Cannot merge nodes with different parents.
	 * @throw InvalidRuntimeConfigurationException [CANNOT_MERGE_LEAF_NODES] Cannot merge leaf entries.
	 */
	lemon::ListDigraph::Node merge(lemon::ListDigraph::Node node1, lemon::ListDigraph::Node node2);

#ifdef CLUE_GRAPH
	void print(std::ostream& os, lemon::ListDigraph::Node node);
#endif

	int innerBranching;
	int leafBranching;
	double threshold;
	int maxSize;

	EuclideanSpaceProvider<T> * euclidianProvider;
	DissimilarityMeasure<CFEntry<T> > *clusterDistanceMeasure;
	AttributeCalculator<CFEntry<T> > *thresholdAttribute;
	ThresholdCalculator *thresholdFunctor;

	lemon::ListDigraph tree;
	lemon::ListDigraph::ArcMap<CFEntry<T>*> innerFeatures;
	lemon::ListDigraph::NodeMap<std::vector<CFEntry<T>*>*> leafFeatures;
	lemon::ListDigraph::Node root;
};

//TODO Require valid EuclideanSpaceProvider etc.
template<typename T> CFTree<T>::CFTree(EuclideanSpaceProvider<T> * euclidianProvider,
                                       DissimilarityMeasure<CFEntry<T> > *clusterDistanceMeasure,
                                       AttributeCalculator<CFEntry<T> > *thresholdAttribute,
                                       ThresholdCalculator *thresholdFunctor,
                                       int innerBranching, int leafBranching, double threshold, int maxSize) :
	euclidianProvider(euclidianProvider->clone()),
	clusterDistanceMeasure(clusterDistanceMeasure->clone()),
	thresholdAttribute(thresholdAttribute->clone()),
	thresholdFunctor(thresholdFunctor),
	innerBranching(innerBranching),
	leafBranching(leafBranching),
	threshold(threshold),
	maxSize(maxSize),
	tree(),
	innerFeatures(tree),
	leafFeatures(tree),
	root(addNode())
{
}

template<typename T> CFTree<T>::CFTree(const CFTree<T>& x) :
	euclidianProvider(x.euclidianProvider->clone()),
	clusterDistanceMeasure(x.clusterDistanceMeasure->clone()),
	thresholdAttribute(x.thresholdAttribute->clone()),
	thresholdFunctor(x.thresholdFunctor),
	innerBranching(x.innerBranching),
	leafBranching(x.leafBranching),
	maxSize(x.maxSize),
	threshold(x.threshold),
	tree(),
	innerFeatures(tree),
	leafFeatures(tree)
{
	using namespace lemon;

	DigraphCopy<ListDigraph, ListDigraph> cg(x.tree, tree);
	cg.arcMap(x.innerFeatures, innerFeatures);
	cg.nodeMap(x.leafFeatures, leafFeatures);
	cg.node(x.root, root);
	cg.run();

	for (ListDigraph::ArcIt a(tree); a != INVALID; ++a)
	{
		if(innerFeatures[a] != 0)
			innerFeatures[a] = new CFEntry<T>(*innerFeatures[a]);
	}
	for (ListDigraph::NodeIt n(tree); n != INVALID; ++n)
	{
		if(leafFeatures[n] != 0)
		{
			leafFeatures[n] = new std::vector<CFEntry<T>*>(*leafFeatures[n]);
			for(typename std::vector<CFEntry<T>*>::iterator it = leafFeatures[n]->begin(); it != leafFeatures[n]->end(); ++it)
				*it = new CFEntry<T>(**it);
		}
	}
}

template<typename T> CFTree<T>& CFTree<T>::operator= (const CFTree<T>& x)
{
	using namespace lemon;

	if(this == &x)
		return *this;

	// Delete...
	for (ListDigraph::ArcIt a(tree); a != INVALID; ++a)
	{
		if(innerFeatures[a] != 0)
		{
			delete innerFeatures[a];
			innerFeatures[a] = 0;
		}
	}
	for (ListDigraph::NodeIt n(tree); n != INVALID; ++n)
	{
		if(leafFeatures[n] != 0)
		{
			for(typename std::vector<CFEntry<T>*>::iterator it = leafFeatures[n]->begin(); it != leafFeatures[n]->end(); ++it)
				delete *it;
			delete leafFeatures[n];
			leafFeatures[n] = 0;
		}
	}
	delete euclidianProvider;
	delete clusterDistanceMeasure;
	delete thresholdAttribute;

	// Copy...
	euclidianProvider = x.euclidianProvider->clone();
	clusterDistanceMeasure = x.clusterDistanceMeasure->clone();
	thresholdAttribute = x.thresholdAttribute->clone();
	thresholdFunctor = x.thresholdFunctor;
	innerBranching = x.innerBranching;
	leafBranching = x.leafBranching;
	maxSize = x.maxSize;
	threshold = x.threshold;

	DigraphCopy<ListDigraph, ListDigraph> cg(x.tree, tree);
	cg.arcMap(x.innerFeatures, innerFeatures);
	cg.nodeMap(x.leafFeatures, leafFeatures);
	cg.node(x.root, root);
	cg.run();

	for (ListDigraph::ArcIt a(tree); a != INVALID; ++a)
	{
		if(innerFeatures[a] != 0)
			innerFeatures[a] = new CFEntry<T>(*innerFeatures[a]);
	}
	for (ListDigraph::NodeIt n(tree); n != INVALID; ++n)
	{
		if(leafFeatures[n] != 0)
		{
			leafFeatures[n] = new std::vector<CFEntry<T>*>(*leafFeatures[n]);
			for(typename std::vector<CFEntry<T>*>::iterator it = leafFeatures[n]->begin(); it != leafFeatures[n]->end(); ++it)
				*it = new CFEntry<T>(**it);
		}
	}
}

template<typename T> CFTree<T>::~CFTree()
{
	using namespace lemon;

	for (ListDigraph::ArcIt a(tree); a != INVALID; ++a)
	{
		if(innerFeatures[a] != 0)
		{
			delete innerFeatures[a];
			innerFeatures[a] = 0;
		}
	}
	for (ListDigraph::NodeIt n(tree); n != INVALID; ++n)
	{
		if(leafFeatures[n] != 0)
		{
			for(typename std::vector<CFEntry<T>*>::iterator it = leafFeatures[n]->begin(); it != leafFeatures[n]->end(); ++it)
				delete *it;
			delete leafFeatures[n];
			leafFeatures[n] = 0;
		}
	}
	delete euclidianProvider;
	delete clusterDistanceMeasure;
	delete thresholdAttribute;
}

template<typename T> void CFTree<T>::insert(T point)
{
	CFEntry<T> pointFeature(1, point, point*point);
	insert(pointFeature);
}

template<typename T> void CFTree<T>::insert(CFEntry<T> pointFeature)
{
	using namespace lemon;

	// 1. Identify matching leaf
	ListDigraph::Node currentNode = root;
	while(countOutArcs(tree, currentNode) > 0)
	{
		ListDigraph::Node minNode = INVALID;
		ListDigraph::Arc minArc = INVALID;
		double minDist = -1;
		for (ListDigraph::OutArcIt arc(tree, currentNode); arc != INVALID; ++arc)
		{
			double dist = clusterDistanceMeasure->dissimilarity(*innerFeatures[arc], pointFeature);
			if(dist < minDist || minNode == INVALID)
			{
				minNode = tree.target(arc);
				minArc = arc;
				minDist = dist;
			}
		}
		currentNode = minNode;
	}

	// 2. Modify leaf
	std::vector<CFEntry<T>*>& leafEntries = *leafFeatures[currentNode];
	CFEntry<T>* minEntry = 0;
	double minDist = -1;
	for (typename std::vector<CFEntry<T>*>::iterator it = leafEntries.begin(); it!=leafEntries.end(); it++)
	{
		double dist = clusterDistanceMeasure->dissimilarity(**it, pointFeature);
		if(dist < minDist || minEntry == 0)
		{
			minEntry = *it;
			minDist = dist;
		}
	}

	bool lastSplitAtLeaf = true;
	bool splitted = false;
	std::pair<lemon::ListDigraph::Node, lemon::ListDigraph::Node> lastSplitPair;
	if(minEntry != 0 && thresholdAttribute->calculate(*minEntry + pointFeature) <= threshold)
	{
		// Update CFEntry
		*minEntry += pointFeature;
	}
	else
	{
		// Add new CFEntry
		CFEntry<T>* newEntry = new CFEntry<T>(pointFeature);
		leafEntries.push_back(newEntry);
		if(leafEntries.size() > leafBranching)
		{
			lastSplitPair = split(currentNode, newEntry);
			currentNode = lastSplitPair.first;
			splitted = true;
		}
	}

	// 3. Modify path
	while(countInArcs(tree, currentNode) > 0)
	{
		ListDigraph::InArcIt arc(tree, currentNode);

		currentNode = tree.source(arc);
		if(splitted && countOutArcs(tree, currentNode) > innerBranching)
		{
			lastSplitPair = split(currentNode, innerFeatures[arc]);
			currentNode = lastSplitPair.first;
		}
		else if(countOutArcs(tree, currentNode) > innerBranching)
		{
			// DEBUG
			throw 0;
		}
		else if(splitted)
		{
			splitted = false;

			if(countOutArcs(tree, currentNode) > 2 && !lastSplitAtLeaf)
			{
				ListDigraph::Arc minArc1 = INVALID;
				ListDigraph::Arc minArc2 = INVALID;
				double minDist = -1;
				for (ListDigraph::OutArcIt arc1(tree, currentNode); arc1 != INVALID; ++arc1)
				{
					for (ListDigraph::OutArcIt arc2(arc1); arc2 != INVALID; ++arc2)
					{
						if(arc1 == arc2)
							continue;
						double dist = clusterDistanceMeasure->dissimilarity(*innerFeatures[arc1], *innerFeatures[arc2]);
						if(dist < minDist || minArc1 == INVALID)
						{
							minArc1 = arc1;
							minArc2 = arc2;
							minDist = dist;
						}
					}
				}
				if(!(tree.target(minArc1) == lastSplitPair.first  && tree.target(minArc2) == lastSplitPair.second) &&
				        !(tree.target(minArc1) == lastSplitPair.second && tree.target(minArc2) == lastSplitPair.first))
				{
					ListDigraph::Node mergedNode = merge(tree.target(minArc1), tree.target(minArc2));
					if(countOutArcs(tree, mergedNode) > innerBranching)
						lastSplitPair = split(mergedNode, 0, true);
				}
			}
		}
		else
		{
			// !splitted
			*innerFeatures[arc] += pointFeature;
		}
		lastSplitAtLeaf = false;
	}

	//Rebuild tree if needed
	while(this->size() > maxSize)
	{
		double th = (*thresholdFunctor)(threshold);
		rebuild(th);
	}
}

template<typename T> void CFTree<T>::rebuild(double newThreshold)
{
	threshold = newThreshold;
	if(countNodes(tree) > 1)
	{
		CFTree<T> newTree(euclidianProvider, clusterDistanceMeasure, thresholdAttribute, thresholdFunctor, innerBranching, leafBranching, newThreshold, maxSize);
		rebuildTraverse(root, newTree, newTree.root);
		lemon::ListDigraph::Node oldRoot = root;
		*this = newTree;
		int s = size();
		clean(oldRoot);
		int t = size();
		int a = 0;
	}
}

template<typename T> void CFTree<T>::rebuildTraverse(lemon::ListDigraph::Node oldNode, CFTree<T>& newCFTree, lemon::ListDigraph::Node newNode)
{
	using namespace lemon;

	if(countOutArcs(tree, oldNode) > 0)
	{
		// Inner node
		for (ListDigraph::OutArcIt arc(tree, oldNode); arc != INVALID; ++arc)
		{
			ListDigraph::Node oldChild = tree.target(arc);
			ListDigraph::Node newChild = newCFTree.addNode();
			ListDigraph::Arc newArc = newCFTree.tree.addArc(newNode, newChild);
			newCFTree.innerFeatures[newArc] = new CFEntry<T>(0, newCFTree.euclidianProvider->nullVector(), 0);
			rebuildTraverse(oldChild, newCFTree, newChild);
		}
	}
	else
	{
		// Leaf
		for (typename std::vector<CFEntry<T>*>::iterator currentEntryIt = leafFeatures[oldNode]->begin(); currentEntryIt!=leafFeatures[oldNode]->end(); currentEntryIt++)
		{
			// 1. Identify NewClosestPath leaf
			ListDigraph::Node newTravNode = newCFTree.root;
			while(countOutArcs(newCFTree.tree, newTravNode) > 0)
			{
				ListDigraph::Node minNode = INVALID;
				ListDigraph::Arc minArc = INVALID;
				double minClosestDist = -1;
				for (ListDigraph::OutArcIt arc(newCFTree.tree, newTravNode); arc != INVALID; ++arc)
				{
					double dist = clusterDistanceMeasure->dissimilarity(*newCFTree.innerFeatures[arc], **currentEntryIt);
					if((dist < minClosestDist || minNode == INVALID) && (*newCFTree.innerFeatures[arc]).number != 0) //number != 0 --> don't add to empty NewCurrentPath
					{
						minNode = newCFTree.tree.target(arc);
						minArc = arc;
						minClosestDist = dist;
					}
				}
				if(minNode != INVALID)
					newTravNode = minNode;
				else
					break;
			}

			// 2. Identify best CFEntrys
			std::vector<CFEntry<T>*>& leafClosestEntries = *newCFTree.leafFeatures[newTravNode];
			CFEntry<T>* minClosestEntry = 0;
			double minClosestDist = -1;
			for (typename std::vector<CFEntry<T>*>::iterator newEntryCandidateIt = leafClosestEntries.begin(); newEntryCandidateIt!=leafClosestEntries.end(); newEntryCandidateIt++)
			{
				double dist = clusterDistanceMeasure->dissimilarity(**newEntryCandidateIt, **currentEntryIt);
				if(dist < minClosestDist || minClosestEntry == 0)
				{
					minClosestEntry = *newEntryCandidateIt;
					minClosestDist = dist;
				}
			}
			std::vector<CFEntry<T>*>& leafCurrentEntries = *newCFTree.leafFeatures[newNode];
			CFEntry<T>* minCurrentEntry = 0;
			double minCurrrentDist = -1;
			for (typename std::vector<CFEntry<T>*>::iterator newEntryCandidateIt = leafCurrentEntries.begin(); newEntryCandidateIt!=leafCurrentEntries.end(); newEntryCandidateIt++)
			{
				double dist = clusterDistanceMeasure->dissimilarity(**newEntryCandidateIt, **currentEntryIt);
				if(dist < minCurrrentDist || minCurrentEntry == 0)
				{
					minCurrentEntry = *newEntryCandidateIt;
					minCurrrentDist = dist;
				}
			}

			// 3. Decide where to add the OldCurrentPath's entry
			if(minClosestEntry != 0 && newTravNode != newNode && minClosestEntry != 0 && newCFTree.thresholdAttribute->calculate(*minClosestEntry + **currentEntryIt) <= threshold)
			{
				// Update CFEntry at new position (NewClosestPath)
				*minClosestEntry += **currentEntryIt;
			}
			else if(minClosestEntry != 0 && newTravNode != newNode && leafClosestEntries.size() < newCFTree.leafBranching)
			{
				// Add new CFEntry at new position (NewClosestPath)
				leafClosestEntries.push_back(new CFEntry<T>(**currentEntryIt));
			}
			else if(minCurrentEntry != 0 && newCFTree.thresholdAttribute->calculate(*minCurrentEntry + **currentEntryIt) <= threshold)
			{
				// Update CFEntry at old position (NewCurrentPath)
				*minCurrentEntry += **currentEntryIt;
				newTravNode = newNode;
			}
			else
			{
				// Add new CFEntry at old position (NewCurrentPath)
				newCFTree.leafFeatures[newNode]->push_back(new CFEntry<T>(**currentEntryIt));
				newTravNode = newNode;
			}

			// 4. Update parent path
			while(newTravNode != newCFTree.root)
			{
				ListDigraph::InArcIt parentArc(newCFTree.tree, newTravNode);
				*newCFTree.innerFeatures[parentArc] += **currentEntryIt;
				newTravNode = newCFTree.tree.source(parentArc);
			}
		}
	}
}

template<typename T> std::vector<CFEntry<T> > CFTree<T>::leafEntries()
{
	std::vector<CFEntry<T> > ret;
	leafEntriesTraverse(root, &ret);
	return ret;
}

template<typename T> void CFTree<T>::leafEntriesTraverse(lemon::ListDigraph::Node currentNode, std::vector<CFEntry<T> >* vec)
{
	using namespace lemon;

	if(countOutArcs(tree, currentNode) > 0)
	{
		// Inner node
		for (ListDigraph::OutArcIt arc(tree, currentNode); arc != INVALID; ++arc)
			leafEntriesTraverse(tree.target(arc), vec);
	}
	else
	{
		// Leaf
		std::vector<CFEntry<T>*>& leafEntries = *leafFeatures[currentNode];
		for (typename std::vector<CFEntry<T>*>::iterator it = leafEntries.begin(); it!=leafEntries.end(); it++)
		{
			CFEntry<T> ins(**it);
			vec->push_back(ins);
		}
	}
}

template<typename T> void CFTree<T>::clean(lemon::ListDigraph::Node node)
{
	using namespace lemon;

	if(countOutArcs(tree, node) > 0)
	{
		// Inner node
		for (ListDigraph::OutArcIt arc(tree, node); arc != INVALID; ++arc)
			clean(tree.target(arc));
		if(countOutArcs(tree, node) == 0)
		{
			ListDigraph::InArcIt parentArc(tree, node);
			delete innerFeatures[parentArc];
			innerFeatures[parentArc] = 0;
			tree.erase(parentArc);
			eraseNode(node);
		}
	}
	else
	{
		// Leaf
		if(leafFeatures[node]->size() == 0)
		{
			ListDigraph::InArcIt parentArc(tree, node);
			tree.erase(parentArc);
			eraseNode(node);
		}
	}
}

template<typename T> int CFTree<T>::size() const
{
	return lemon::countNodes(tree);
}

template<typename T> size_t CFTree<T>::numberOfPoints() const
{
	size_t num = 0;
	for (lemon::ListDigraph::OutArcIt arc(tree, root); arc != lemon::INVALID; ++arc)
		num += innerFeatures[arc]->number;
	return num;
}

template<typename T> std::pair<lemon::ListDigraph::Node, lemon::ListDigraph::Node> CFTree<T>::split(lemon::ListDigraph::Node node, CFEntry<T>* keepTrack, bool fillUp)
{
	using namespace lemon;

	std::pair<lemon::ListDigraph::Node, lemon::ListDigraph::Node> ret;

	// Split of inner node or leaf?
	if(countOutArcs(tree, node) > 0)
	{
		// Inner node
		// Identify farthest pair
		ListDigraph::Arc maxArc1 = INVALID;
		ListDigraph::Arc maxArc2 = INVALID;
		double maxDist = -1;
		for (ListDigraph::OutArcIt arc1(tree, node); arc1 != INVALID; ++arc1)
		{
			for (ListDigraph::OutArcIt arc2(arc1); arc2 != INVALID; ++arc2)
			{
				if(arc1 == arc2)
					continue;
				double dist = clusterDistanceMeasure->dissimilarity(*innerFeatures[arc1], *innerFeatures[arc2]);
				if(dist > maxDist)
				{
					maxArc1 = arc1;
					maxArc2 = arc2;
					maxDist = dist;
				}
			}
		}

		// Decide to which of the two new child nodes the clustering features have to be assigned
		ListDigraph::Node child1 = addNode();
		ListDigraph::Node child2 = addNode();
		T nullVector = euclidianProvider->nullVector();
		CFEntry<T> ent1(0, nullVector, 0);
		CFEntry<T> ent2(0, nullVector, 0);
		std::vector<ListDigraph::Arc> to1;
		std::vector<ListDigraph::Arc> to2;
		// Default pair / check for tracking maxArcs
		bool trackChild1 = innerFeatures[maxArc1] == keepTrack ? true : false;
		// Insert seeds
		to1.push_back(maxArc1);
		to2.push_back(maxArc2);
		ent1 += *innerFeatures[maxArc1];
		ent2 += *innerFeatures[maxArc2];
		for (ListDigraph::OutArcIt arc(tree, node); arc != INVALID; ++arc)
		{
			if(arc == maxArc1 || arc == maxArc2)
				continue;
			bool isNearerToMaxArc1 = clusterDistanceMeasure->dissimilarity(*innerFeatures[arc], *innerFeatures[maxArc1]) < clusterDistanceMeasure->dissimilarity(*innerFeatures[arc], *innerFeatures[maxArc2]);
			bool hasTo1Space = !fillUp || to1.size() < innerBranching;
			bool hasTo2Space = !fillUp || to2.size() < innerBranching;
			if((isNearerToMaxArc1 && hasTo1Space) || (!hasTo2Space))
			{
				to1.push_back(arc);
				ent1 += *innerFeatures[arc];
				if(innerFeatures[arc] == keepTrack)
					trackChild1 = true;
			}
			else
			{
				to2.push_back(arc);
				ent2 += *innerFeatures[arc];
				if(innerFeatures[arc] == keepTrack)
					trackChild1 = false;
			}
		}
		//Prepare return pair
		if(trackChild1)
		{
			ret.first = child1;
			ret.second = child2;
		}
		else
		{
			ret.first = child2;
			ret.second = child1;
		}

		// Change sources from old to new nodes
		for (typename std::vector<ListDigraph::Arc>::iterator it = to1.begin(); it != to1.end(); it++)
			tree.changeSource(*it, child1);
		for (typename std::vector<ListDigraph::Arc>::iterator it = to2.begin(); it != to2.end(); it++)
			tree.changeSource(*it, child2);

		// DEBUG Consistency check
		if(countOutArcs(tree, node) != 0)
			throw 0;

		// Update tree
		ListDigraph::Node parent = INVALID;
		if(node == root)
		{
			// Root
			parent = addNode();
			this->root = parent;
			eraseNode(node);
		}
		else
		{
			// Not root
			ListDigraph::InArcIt arc(tree, node);
			parent = tree.source(arc);
			tree.erase(arc);
			eraseNode(node);
		}

		ListDigraph::Arc childArc1 = tree.addArc(parent, child1);
		ListDigraph::Arc childArc2 = tree.addArc(parent, child2);
		innerFeatures[childArc1] = new CFEntry<T>(ent1);
		innerFeatures[childArc2] = new CFEntry<T>(ent2);
	}
	else
	{
		// Leaf
		// Identify farthest pair
		CFEntry<T>* maxEnt1 = 0;
		CFEntry<T>* maxEnt2 = 0;
		double maxDist = -1;
		std::vector<CFEntry<T>*>& entries = *leafFeatures[node];
		for (typename std::vector<CFEntry<T>*>::iterator ent1 = entries.begin(); ent1 != entries.end(); ++ent1)
		{
			for (typename std::vector<CFEntry<T>*>::iterator ent2(ent1); ent2 != entries.end(); ++ent2)
			{
				if(*ent1 == *ent2)
					continue;
				double dist = clusterDistanceMeasure->dissimilarity(**ent1, **ent2);
				if(dist > maxDist)
				{
					maxEnt1 = *ent1;
					maxEnt2 = *ent2;
					maxDist = dist;
				}
			}
		}

		// Assign clustering features to the two new child nodes
		ListDigraph::Node child1 = addNode();
		ListDigraph::Node child2 = addNode();
		T nullVector = euclidianProvider->nullVector();
		CFEntry<T> ent1(0, nullVector, 0);
		CFEntry<T> ent2(0, nullVector, 0);
		std::vector<CFEntry<T>*>& to1 = * new std::vector<CFEntry<T>*>();
		std::vector<CFEntry<T>*>& to2 = * new std::vector<CFEntry<T>*>();
		for (typename std::vector<CFEntry<T>*>::iterator ent = entries.begin(); ent != entries.end(); ++ent)
		{
			bool isMaxEnt1 = *ent == maxEnt1;
			bool isNotMaxEnt2 = *ent != maxEnt2;
			bool isNearerToMaxEnt1 = clusterDistanceMeasure->dissimilarity(**ent, *maxEnt1) < clusterDistanceMeasure->dissimilarity(**ent, *maxEnt2);
			bool isTo1Full = to1.size() >= innerBranching;
			bool isTo2Full = to2.size() >= innerBranching;
			if((isMaxEnt1 || (isNotMaxEnt2 && isNearerToMaxEnt1)) && (!fillUp || !isTo1Full || isTo2Full))
			{
				to1.push_back(*ent);
				ent1 += **ent;
				if(*ent == keepTrack)
				{
					ret.first = child1;
					ret.second = child2;
				}
			}
			else
			{
				to2.push_back(*ent);
				ent2 += **ent;
				if(*ent == keepTrack)
				{
					ret.first = child2;
					ret.second = child1;
				}
			}
		}
		entries.clear();

		// Update tree
		ListDigraph::Node parent = INVALID;
		if(node == root)
		{
			// Root
			parent = addNode();
			this->root = parent;
			eraseNode(node);
		}
		else
		{
			// Not root
			ListDigraph::InArcIt arc(tree, node);
			parent = tree.source(arc);
			tree.erase(arc);
			eraseNode(node);
		}

		leafFeatures[child1] = &to1;
		leafFeatures[child2] = &to2;
		ListDigraph::Arc childArc1 = tree.addArc(parent, child1);
		ListDigraph::Arc childArc2 = tree.addArc(parent, child2);
		innerFeatures[childArc1] = new CFEntry<T>(ent1);
		innerFeatures[childArc2] = new CFEntry<T>(ent2);
	}

	return ret;
}

template<typename T> lemon::ListDigraph::Node CFTree<T>::merge(lemon::ListDigraph::Node node1, lemon::ListDigraph::Node node2)
{
	using namespace lemon;

	//Same node?
	if(node1 == node2)
		throw InvalidArgumentException(CANNOT_MERGE_NODE_ITSELF, "Cannot merge node with itself.", "node1/node2");
	//Same parent?
	if(tree.source(ListDigraph::InArcIt(tree, node1)) != tree.source(ListDigraph::InArcIt(tree, node2)))
		throw InvalidArgumentException(CANNOT_MERGE_NODE_DIFFERENT_PARENTS, "Cannot merge nodes with different parents.", "node1/node2");

	// Mergingth of inner node or leaf?
	if(countOutArcs(tree, node1) > 0)
	{
		// Inner node
		ListDigraph::Node child = addNode();
		T nullVector = euclidianProvider->nullVector();
		CFEntry<T> ent(0, nullVector, 0);
		ent += *innerFeatures[ListDigraph::InArcIt(tree, node1)];
		ent += *innerFeatures[ListDigraph::InArcIt(tree, node2)];
		std::vector<ListDigraph::Arc> toChange;
		for (ListDigraph::OutArcIt arc(tree, node1); arc != INVALID; ++arc)
			toChange.push_back(arc);
		for (ListDigraph::OutArcIt arc(tree, node2); arc != INVALID; ++arc)
			toChange.push_back(arc);
		for (typename std::vector<ListDigraph::Arc>::iterator it = toChange.begin(); it != toChange.end(); it++)
			tree.changeSource(*it, child);

		// Update tree
		ListDigraph::InArcIt arc1(tree, node1);
		ListDigraph::InArcIt arc2(tree, node2);
		ListDigraph::Node parent = tree.source(arc1);
		tree.erase(arc1);
		tree.erase(arc2);
		eraseNode(node1);
		eraseNode(node2);
		ListDigraph::Arc childArc = tree.addArc(parent, child);
		innerFeatures[childArc] = new CFEntry<T>(ent);

		return child;
	}
	else
	{
		// Leaf
		throw InvalidRuntimeConfigurationException(CANNOT_MERGE_LEAF_NODES, "Cannot merge leaf entries.");
	}
}

template<typename T> lemon::ListDigraph::Node CFTree<T>::addNode()
{
	lemon::ListDigraph::Node node = tree.addNode();
	leafFeatures[node] = new std::vector<CFEntry<T>*>();
	return node;
}

template<typename T> void CFTree<T>::eraseNode(lemon::ListDigraph::Node node)
{
	if(leafFeatures[node] != 0)
	{
		for (typename std::vector<CFEntry<T>*>::iterator ent = leafFeatures[node]->begin(); ent != leafFeatures[node]->end(); ++ent)
			delete *ent;
		delete leafFeatures[node];
		leafFeatures[node] = 0;
	}
	tree.erase(node);
}

#ifdef CLUE_GRAPH
template<typename T> void CFTree<T>::print(std::ostream& os)
{
	os << "digraph G{\n";
	os << "node [shape=record];\n";
	print(os, root);
	os << "}\n";
}

template<typename T> void CFTree<T>::print(std::ostream& os, lemon::ListDigraph::Node node)
{
	using namespace lemon;

	int id = tree.id(node);
	int outArcs = countOutArcs(tree, node);

	os << "node" << id << "[label=\"";
	if(outArcs > 0)
	{
		// Inner node
		int fvalue = 0;
		for (ListDigraph::OutArcIt arc(tree, node); arc != INVALID; ++arc)
		{
			if(fvalue > 0)
				os << "|";
			os << "<f" << fvalue << "> " << innerFeatures[arc]->number << "," << innerFeatures[arc]->LS << "," << innerFeatures[arc]->SS;
			fvalue++;
		}
	}
	else
	{
		// Leaf
		std::vector<CFEntry<T>*>& entries = *leafFeatures[node];
		int fvalue = 0;
		for (typename std::vector<CFEntry<T>*>::iterator ent = entries.begin(); ent != entries.end(); ++ent)
		{
			if(fvalue > 0)
				os << "|";
			os << "<f" << fvalue << "> " << (*ent)->number << "," << (*ent)->LS << "," << (*ent)->SS;
			fvalue++;
		}
	}
	os << "\"];\n";

	if(outArcs > 0)
	{
		// Inner node
		int fvalue = 0;
		for (ListDigraph::OutArcIt arc(tree, node); arc != INVALID; ++arc)
		{
			print(os, tree.target(arc));
			os << "node" << id << ":f" << fvalue << " -> node" << tree.id(tree.target(arc)) << ";\n";
			fvalue++;
		}
	}
	else
	{
		// Leaf
	}
}
#endif

}

#endif
