#ifndef DASGUPTALONG_H
#define DASGUPTALONG_H

#include <cmath>

#include "../base/algorithm.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../datastructure/partitionsolution.h"
#include "../sampling/farthestfirsttraversal.h"
#include "../misc/dynamicbitset.h"

#include "../exception/invalidruntimeconfigurationexception.h"
#include "../exception/invalidargumentexception.h"

namespace CluE
{

/**
 * @brief DasguptaLong algorithm.
 *
 * Hierarchical clustering algorithm as described in
 * S. Dasgupta and P. M. Long. "Performance guarantees for hierarchical clustering". <em>Journal of Computer and System Sciences</em>, 70(4):555-569, 2005.
 *
 * @ingroup clustering_algorithms
 */
template<typename T> class DasguptaLong : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	/**
	 * @brief Constructor for general usage.
	 * @note Remember to set granularity.
	 * @sa setMeasure
	 * @sa setInput
	 * @sa setFirstPoint
	 * @sa setGranularity
	 */
	DasguptaLong(DissimilarityMeasure<T> const *measure = 0, std::vector<T*> const* input = 0, unsigned int firstPoint = 0);

	DasguptaLong(const DasguptaLong<T>&);
	DasguptaLong<T>& operator= (const DasguptaLong<T>&);
	virtual ~DasguptaLong();

	/**
	 * @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	 * @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	 */
	virtual SolutionProvider* compute();
	
	virtual void setInput(std::vector<T*> const*);
	
	virtual void setMeasure(DissimilarityMeasure<T> const *measure);

	/**
	 * @brief Sets the starting point for farthest first traversal.
	 */
	void setFirstPoint(unsigned int index);

	/**
	 * @brief Sets the granularity.
	 * @sa "Performance guarantees for hierarchical clustering"  (see above)
	 * @throw InvalidArgumentException [0] The given value of alpha is invalid. No changes were made.
	 * @throw InvalidArgumentException [1] The given value of beta is invalid. No changes were made.
	 */
	void setGranularity(double alpha, double beta);

	/**
	* @brief does a dynamic cast of the given Algorithm to DasguptaLong<T>
	* @return NULL if the Algorithm is not a DasguptaLong<T> instance
	*/
	static DasguptaLong<T>* toDasguptaLong(Algorithm* s);

private:
	DissimilarityMeasure<T> *measure;
	std::vector<T*> const *input;
	unsigned int first_point;
	double alpha;
	double beta;
};

template<typename T> DasguptaLong<T>::DasguptaLong(DissimilarityMeasure<T> const *measure, std::vector<T*> const* data, unsigned int fp) :
	measure(measure==0 ? 0 : measure->clone()),
	input(data),
	first_point(fp),
	alpha(1),
	beta(2)
{
}

template<typename T> DasguptaLong<T>::DasguptaLong(const DasguptaLong<T>& dl) :
	measure(dl.measure == 0 ? 0 : dl.measure->clone()),
	input(dl.input),
	first_point(first_point),
	alpha(dl.alpha),
	beta(dl.beta)
{
}

template<typename T> DasguptaLong<T>& DasguptaLong<T>::operator= (const DasguptaLong<T>& dl)
{
	if(measure != 0)
		delete measure;

	Algorithm::operator= (dl);

	measure = dl.measure == 0 ? 0 : dl.measure->clone();
	input = dl.input;
	first_point = first_point;
	alpha = dl.alpha;
	beta = dl.beta;

	return *this;
}

template<typename T> DasguptaLong<T>::~DasguptaLong()
{
	if(measure != 0)
		delete measure;
}

template<typename T> SolutionProvider* DasguptaLong<T>::compute()
{
	PartitionSolution<T>* solution = new PartitionSolution<T>();
	unsigned int numOfPoints = this->input->size();

	if(this->input == 0)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if(this->measure == 0)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");


	time_t start, end;
	start = time(0);

	FarthestFirstTraversal<T> fft(this->measure, this->input, numOfPoints, first_point);
	DiscreteProxySolution<T> *dps = dynamic_cast<DiscreteProxySolution<T>*>(fft.compute());
	std::vector<T*> *numberedPoints = &dps->proxysets[0];

	//Build R
	std::vector<double> rValues;
	std::vector<unsigned int> rPoints;
	rValues.push_back(0);  //dummy value
	for(unsigned int i = 1; i < numOfPoints; i++)
	{
		double minDist = measure->dissimilarity(* (*numberedPoints)[i], * (*numberedPoints)[0]);
		for(unsigned int j = 1; j < i; j++)
		{
			double candDist = measure->dissimilarity(* (*numberedPoints)[i], * (*numberedPoints)[j]);
			if(candDist < minDist)
				minDist = candDist;
		}
		rValues.push_back(minDist);
	}
	double r = alpha * rValues[1];

	//Build L
	DynamicBitset<> isLeveled(numOfPoints);
	isLeveled.set(first_point, true);

	std::vector<std::vector<unsigned int> > lValues;
	lValues.push_back(std::vector<unsigned int>());
	lValues[0].push_back(first_point);
	for(unsigned int i = 1; isLeveled.count() < numOfPoints; i++)
	{
		lValues.push_back(std::vector<unsigned int>());

		double lowerBound = r / std::pow(beta, (double) i);
		double upperBound = r / std::pow(beta, (double)(i-1));
		for(unsigned int j = 0; j < numOfPoints; j++)
		{
			if(isLeveled[j])
				continue;

			if(lowerBound < rValues[j] && rValues[j] <= upperBound)
			{
				isLeveled.set(j, true);
				lValues[i].push_back(j);
			}
		}
	}

	//Pi
	unsigned int pi[numOfPoints];
	pi[first_point] = 0;
	unsigned int levels = lValues.size();
	for(unsigned int level = 1; level < levels; level++)
	{
		//Iterate over all levels...
		unsigned int levelSize = lValues[level].size();
		for(unsigned int levelIndex = 0; levelIndex < levelSize; levelIndex++)
		{
			//...to assign to each point...
			T* currentPoint = (*numberedPoints)[lValues[level][levelIndex]];
			double minDist = measure->dissimilarity(*currentPoint, * (*numberedPoints)[first_point]);
			unsigned int minPoint = first_point;
			for(unsigned int lowerLevel = 1; lowerLevel < level; lowerLevel++)
			{
				//...from a lower level...
				unsigned int lowerLevelSize = lValues[lowerLevel].size();
				for(unsigned int lowerLevelIndex = 0; lowerLevelIndex < lowerLevelSize; lowerLevelIndex++)
				{
					//...the nearest point.
					double candDist = measure->dissimilarity(*currentPoint, * (*numberedPoints)[lValues[lowerLevel][lowerLevelIndex]]);
					if(candDist < minDist)
					{
						minDist = candDist;
						minPoint = lValues[lowerLevel][lowerLevelIndex];
					}
				}
			}
			pi[lValues[level][levelIndex]] = minPoint;
		}
	}

	/*
	std::cout << "Pi:" << std::endl;
	int n = lValues.size();
	for(int i = 0; i < n; i++)
	{
		int m = lValues[i].size();
		for(int j = 0; j < m; j++)
		{
			std::cout << lValues[i][j] << " -> " << pi[lValues[i][j]] << std::endl;
		}
	}
	*/

	//Build clusterings
	std::vector<std::vector<std::vector<T*> > >& sets = solution->partitions;
	for(unsigned int i = 0; i < levels; i++)
	{
		sets.push_back(std::vector<std::vector<T*> >());
		std::vector<std::vector<T*> >& partitions = sets.back();

		unsigned int recursivePi[numOfPoints];
		//In-granularity-level points
		for(unsigned int j = 0; j <= i; j++)
		{
			unsigned int size = lValues[j].size();
			for(unsigned int k = 0; k < size; k++)
			{
				partitions.push_back(std::vector<T*>());
				recursivePi[lValues[j][k]] = partitions.size()-1;
			}
		}
		//Out-of-granularity-level points
		for(unsigned int j = i+1; j < levels; j++)
		{
			unsigned int size = lValues[j].size();
			for(unsigned int k = 0; k < size; k++)
			{
				unsigned int index = lValues[j][k];
				recursivePi[index] = recursivePi[pi[index]];
			}
		}

		//Assign points to clusters
		for(unsigned int i = 0; i < numOfPoints; i++)
		{
			partitions[recursivePi[i]].push_back((*numberedPoints)[i]);
		}
	}

	delete dps;

	end = time(0);
	solution->seconds = end-start;

	return solution;
}

template<typename T> void DasguptaLong<T>::setFirstPoint(unsigned int s)
{
	this->first_point = s;
}

template<typename T> void DasguptaLong<T>::setGranularity(double a, double b)
{
	if(b > 1)
	{
		if(a >= 1 && a < b)
		{
			this->alpha = a;
			this->beta = b;
		}
		else
		{
			throw InvalidArgumentException(0, "The given value of alpha is invalid. No changes were made.", "alpha");
		}
	}
	else
	{
		throw InvalidArgumentException(1, "The given value of beta is invalid. No changes were made.", "beta");
	}
}

template<typename T> void DasguptaLong<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void DasguptaLong<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(this->measure != 0)
		delete this->measure;
	
	if(measure == 0)
		this->measure = 0;
	else
		this->measure = measure->clone();
}

template<typename T> DasguptaLong<T>* DasguptaLong<T>::toDasguptaLong(Algorithm* s)
{
	return dynamic_cast<DasguptaLong<T>*>(s);
}

}
#endif
