#ifndef GONZALES_H
#define GONZALES_H

#include "../base/algorithm.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../datastructure/discretedoublesolution.h"
#include "../sampling/farthestfirsttraversal.h"

#include "../exception/invalidruntimeconfigurationexception.h"

namespace CluE
{

/**
 * @brief Gonzales algorithm.
 *
 * Clustering algorithm as described in:
 * T. Feder and D. Greene. Optimal algorithms for approximate clustering. <em>Proceedings of the 20th ACM Symposium on the Theory of Computing</em>, 1988. \n
 * Also mentioned as "farthest first traversal" in:
 * S. Dasgupta and P. M. Long. "Performance guarantees for hierarchical clustering". <em>Journal of Computer and System Sciences</em>, 70(4):555-569, 2005.
 *
 * @sa DasguptaLong is a hierarchical clustering algorithm based on the idea of Gonzales.
 *
 * @ingroup clustering_algorithms
 */
template<typename T> class Gonzales : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	/**
	 * @brief Constructor for general use.
	 * @sa setMeasure
	 * @sa setInput
	 * @sa setNumberOfClusters
	 * @sa setFirstPoint
	 */
	Gonzales(DissimilarityMeasure<T> const *measure = 0, std::vector<T*> const* input = 0, unsigned int numberOfClusters = 1, unsigned int firstPoint = 0);

	Gonzales(const Gonzales<T>&);
	Gonzales<T>& operator= (const Gonzales<T>&);
	virtual ~Gonzales();

	/**
	 * @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	 * @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	 * @throw InvalidRuntimeConfigurationException [2] Desired number of clusters is larger than size of input.
	 */
	virtual SolutionProvider* compute();

	virtual void setInput(std::vector<T*> const*);

	virtual void setMeasure(DissimilarityMeasure<T> const *measure);

	/**
	 *@brief Sets the desired number of clusters.
	 */
	void setNumberOfClusters(unsigned int number);

	/**
	 * @brief Sets the starting point for farthest first traversal.
	 */
	void setFirstPoint(unsigned int index);

	/**
	* @brief does a dynamic cast of the given Algorithm to Gonzales<T>
	* @return NULL if the Algorithm is not a Gonzales<T> instance
	*/
	static Gonzales<T>* toGonzales(Algorithm* s);

private:
	DissimilarityMeasure<T>* measure;
	std::vector<T*> const* input;
	unsigned int numberOfClusters;
	unsigned int first_point;
};

template<typename T> Gonzales<T>::Gonzales(DissimilarityMeasure<T> const *measure, std::vector<T*> const* data, unsigned int nc, unsigned int fp) :
	measure(measure==0 ? 0 : measure->clone()),
	input(data),
	numberOfClusters(nc),
	first_point(fp)
{
}

template<typename T> Gonzales<T>::Gonzales(const Gonzales<T>& gon) :
	measure(gon.measure == 0 ? 0 : gon.measure->clone()),
	input(gon.input),
	numberOfClusters(gon.numberOfClusters),
	first_point(gon.first_point)
{
}

template<typename T> Gonzales<T>& Gonzales<T>::operator= (const Gonzales<T>& gon)
{
	if(measure != 0)
		delete measure;

	Algorithm::operator= (gon);

	measure = gon.measure == 0 ? 0 : gon.measure->clone();
	input = gon.input;
	numberOfClusters = gon.numberOfClusters;
	first_point = gon.first_point;

	return *this;
}

template<typename T> Gonzales<T>::~Gonzales()
{
	if(measure != 0)
		delete measure;
}

template<typename T> SolutionProvider* Gonzales<T>::compute()
{
	DiscreteDoubleSolution<T>* solution = new DiscreteDoubleSolution<T>();
	unsigned int numOfPoints = this->input->size();

	if(this->input == 0)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if(this->measure == 0)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");
	if(this->numberOfClusters > numOfPoints)
		throw InvalidRuntimeConfigurationException(2, "Desired number of clusters is larger than size of input.");

	time_t start, end;
	start = time(0);

	FarthestFirstTraversal<T> fft(this->measure, this->input, numberOfClusters, first_point);
	DiscreteProxySolution<T> *dps = dynamic_cast<DiscreteProxySolution<T>*>(fft.compute());

	unsigned int numOfProxies = dps->proxysets[0].size();

	//ProxySolution
	solution->proxysets = dps->proxysets;

	//Prepare PartitionSolution
	solution->partitions.push_back(std::vector<std::vector<T*> >());
	for(unsigned int i = 0; i < numOfProxies; i++)
		solution->partitions[0].push_back(std::vector<T*>());

	//Assign points to proxies
	for(unsigned int i = 0; i < numOfPoints; i++)
	{
		T* point = (*input)[i];

		double min = this->measure->dissimilarity(*point, * (dps->proxysets[0][0]));
		int assignedProxy = 0;
		for(unsigned int j = 1; j < numOfProxies; j++)
		{
			T* proxy = dps->proxysets[0][j];
			double candidate = this->measure->dissimilarity(*point, *proxy);
			if(candidate < min)
			{
				min = candidate;
				assignedProxy = j;
			}
		}

		solution->partitions[0][assignedProxy].push_back(point);
	}

	delete dps;

	end = time(0);
	solution->seconds = end-start;

	return solution;
}

template<typename T> void Gonzales<T>::setFirstPoint(unsigned int s)
{
	this->first_point = s;
}

template<typename T> void Gonzales<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void Gonzales<T>::setNumberOfClusters(unsigned int nc)
{
	this->numberOfClusters = nc;
}

template<typename T> void Gonzales<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(this->measure != 0)
		delete this->measure;

	if(measure == 0)
		this->measure = 0;
	else
		this->measure = measure->clone();
}

template<typename T> Gonzales<T>* Gonzales<T>::toGonzales(Algorithm* s)
{
	return dynamic_cast<Gonzales<T>*>(s);
}
}

#endif
