#ifndef KUMARSABHARWALSEN_H
#define KUMARSABHARWALSEN_H

#include <algorithm>
#include <map>

#include "../base/algorithm.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../datastructure/doublesolution.h"
#include "../base/proxygenerator.h"
#include "../base/dissimilaritymeasure.h"
#include "../sampling/uniformsampling.h"
#include "../misc/fixedsizesubsetiterator.h"
#include "../measure/pointsetdistance.h"

#include "../exception/invalidargumentexception.h"
#include "../exception/invalidruntimeconfigurationexception.h"

namespace CluE
{

/**
 * @brief KumarSabharwalSen algorithm.
 *
 * Approxmative (1+epsilon) clustering algorithm as described in
 * A. Kumar, Y. Sabharwal and S. Sen. "A Simple Linear Time (1+epsilon)-Approximation Algorithm for k-Means Clustering in Any Dimensions". <em>???</em>, 2005.
 *
 * @ingroup clustering_algorithms
 */
template<typename T> class KumarSabharwalSen : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	/**
	 * @brief Constructor for general usage.
	 * @note Remember to set the set sizes.
	 * @sa setSetSizes
	 */
	KumarSabharwalSen(DissimilarityMeasure<T> const *measure = 0, ProxyGenerator<T> const *centroidGenerator = 0, unsigned int numberOfClusters = 2, std::vector<T*> const* input = 0);

	KumarSabharwalSen(const KumarSabharwalSen<T>&);
	KumarSabharwalSen<T>& operator= (const KumarSabharwalSen<T>&);
	virtual ~KumarSabharwalSen();

	/**
	 * @throw InvalidRuntimeConfigurationException [0] Number of desired clusters is 0.
	 * @throw InvalidRuntimeConfigurationException [1] Input size must be larger than number of desired clusters.
	 * @throw InvalidRuntimeConfigurationException [2] Set sizes are uninitialized.
	 */
	virtual DoubleSolution<T>* compute();

	/**
	 * @brief Returns the cost of the last computation.
	 */
	double getLastCost();

	virtual void setInput(std::vector<T*> const*);

	virtual void setMeasure(DissimilarityMeasure<T> const *measure);

	/**
	 * @brief Set the centroid generator used for cluster center approximation.
	 * @sa [KumarSabharwelSen], 2. Preliminaries.
	 */
	void setCentroidGenerator(ProxyGenerator<T> const *centroidGenerator);

	/**
	 * @brief Set the superset and subset sizes used for guessing clusters.
	 * @sa [KumarSabharwalSen], 4.2 The algorithm.
	 * @throw InvalidArgumentException [0] subsetSize has to be smaller than superSetSize. subsetSize.
	 */
	void setSetSizes(unsigned int superSetSize, unsigned int subsetSize);

	/**
	* @brief does a dynamic cast of the given Algorithm to FarthestFirstTraversal<T>
	* @return NULL if the Algorithm is not a FarthestFirstTraversal<T> instance
	*/
	static KumarSabharwalSen<T>* toKumarSabharwalSen(Algorithm* s);

private:
	/**
	 * @brief Used to return a DoubleSolution and its cost.
	 */
	struct Tupel
	{
		Tupel(DoubleSolution<T>* ds, double cost) :
			ds(ds),
			cost(cost)
		{
		}

		Tupel() :
			ds(0),
			cost(std::numeric_limits<double>::max())
		{
		}
		DoubleSolution<T>* ds;
		double cost;
	};

	Tupel irredKMeans(std::set<T*> Q, size_t m, size_t k, std::map<size_t, T> C, double Sum);

	DissimilarityMeasure<T>* measure;
	std::set<T*> input;
	ProxyGenerator<T> *centroidGenerator;
	unsigned int superSetSize;
	unsigned int subsetSize;
	unsigned int numberOfClusters;
	double lastCost;
};

template<typename T> KumarSabharwalSen<T>::KumarSabharwalSen(DissimilarityMeasure<T> const *measure, ProxyGenerator<T> const *centGen, unsigned int numberOfClusters, std::vector<T*> const* data) :
	measure(measure==0 ? 0 : measure->clone()),
	centroidGenerator(centGen==0 ? 0 : centGen->clone()),
	superSetSize(0),
	subsetSize(0),
	numberOfClusters(numberOfClusters),
	lastCost(0)
{
	if(data != 0)
		input = std::set<T*>(data->begin(), data->end());
}

template<typename T> KumarSabharwalSen<T>::KumarSabharwalSen(const KumarSabharwalSen<T>& kss) :
	measure(kss.measure == 0 ? 0 : kss.measure->clone()),
	input(kss.data),
	centroidGenerator(kss.centroidGenerator==0 ? 0 : kss.centroidGenerator->clone()),
	superSetSize(kss.superSetSize),
	subsetSize(kss.subsetSize),
	numberOfClusters(kss.numberOfClusters),
	lastCost(0)
{
}

template<typename T> KumarSabharwalSen<T>& KumarSabharwalSen<T>::operator= (const KumarSabharwalSen<T>& kss)
{
	if(this->measure != 0)
		delete this->measure;
	if(this->centroidGenerator != 0)
		delete this->centroidGenerator;

	Algorithm::operator= (kss);

	this->measure = kss.measure == 0 ? 0 : kss.measure->clone();
	this->input = kss.input;
	this->centroidGenerator = kss.centroidGenerator == 0 ? 0 : kss.centroidGenerator->clone();
	this->superSetSize = kss.superSetSize;
	this->subsetSize = kss.subsetSize;
	this->numberOfClusters = kss.numberOfClusters;
	this->lastCost = kss.lastCost;

	return *this;
}

template<typename T> KumarSabharwalSen<T>::~KumarSabharwalSen()
{
	if(this->measure != 0)
		delete this->measure;
	if(this->centroidGenerator != 0)
		delete this->centroidGenerator;
}

template<typename T> DoubleSolution<T>* KumarSabharwalSen<T>::compute()
{
	if(numberOfClusters == 0)
		throw InvalidRuntimeConfigurationException(0, "Number of desired clusters is 0.");
	if(numberOfClusters > input.size())
		throw InvalidRuntimeConfigurationException(1, "Input size must be larger than number of desired clusters.");
	if(subsetSize == 0 || superSetSize == 0)
		throw InvalidRuntimeConfigurationException(2, "Set sizes are uninitialized.");

	std::map<size_t, T> empty;
	Tupel result;
	for(size_t i = 1; i <= numberOfClusters; i++)
	{
		Tupel candidate(irredKMeans(input, i, i, empty, 0));
		if(candidate.cost < result.cost)
			result = candidate;
	}
	lastCost = result.cost;
	return result.ds;
}

template<typename T> typename KumarSabharwalSen<T>::Tupel KumarSabharwalSen<T>::irredKMeans(std::set<T*> Q, size_t m, size_t k, std::map<size_t, T> C, double Sum)
{
	//C stores cluster centers by their index (used to access PartitionSolution's partitions)
	
	//STEP 1
	if(m == 0)
	{
		//Prepare result
		DoubleSolution<T> *step1ds = new DoubleSolution<T>();
		step1ds->partitions.push_back(std::vector<std::vector<T*> >());
		step1ds->proxysets.push_back(std::vector<T>());

		typename std::map<size_t, T>::iterator endC = C.end();
		for(typename std::map<size_t, T>::iterator itC = C.begin(); itC != endC; itC++)
		{
			step1ds->partitions[0].push_back(std::vector<T*>());
			step1ds->proxysets[0].push_back((*itC).second);
		}

		//Assign points
		typename std::set<T*>::iterator endQ = Q.end();
		for(typename std::set<T*>::iterator itQ = Q.begin(); itQ != endQ; itQ++)
		{
			double minDist = measure->dissimilarity(**itQ, C.begin()->second);
			size_t minIndex = 0;
			for(typename std::map<size_t, T>::iterator itC = C.begin()++; itC != endC; itC++)
			{
				double tmpDist = measure->dissimilarity(**itQ, itC->second);
				if(tmpDist < minDist)
				{
					minDist = tmpDist;
					minIndex = itC->first;
				}
			}
			Sum += minDist;
			step1ds->partitions[0][minIndex].push_back(*itQ);
		}

		Tupel step1result(step1ds, Sum);
		return step1result;
	}

	//STEP 2
	size_t actualSuperSetSize = superSetSize;
	if(actualSuperSetSize > Q.size())
		actualSuperSetSize = Q.size();

	std::vector<T*> qVector(Q.begin(), Q.end());
	UniformSampling<T> supersetSampler(&qVector, superSetSize);
	DiscreteProxySolution<T>* supersetSolution = supersetSampler.compute();

	size_t actualSubsetSize = subsetSize;
	if(actualSubsetSize > supersetSolution->proxysets[0].size())
		actualSubsetSize = supersetSolution->proxysets[0].size();
	FixedSizeSubsetIterator<T> subsetIterator(supersetSolution->proxysets[0], actualSubsetSize);

	Tupel step2result;
	while(true)
	{
		std::vector<T*> currentSubset = subsetIterator.vector();
		T centroid = centroidGenerator->generate(currentSubset);
		std::map<size_t, T> cCentroid(C);
		cCentroid.insert(std::pair<size_t, T>(C.size(), centroid));
		Tupel candidate(irredKMeans(Q, m-1, k, cCentroid, Sum));
		if(candidate.cost < step2result.cost)
		{
			delete step2result.ds;
			step2result = candidate;
		}
		else
			delete candidate.ds;

		if(subsetIterator.hasMore())
			subsetIterator.next();
		else
			break;
	}

	//STEP 3
	if(C.size() > 0 && Q.size() > 1)
	{
		//Found centers yet? Then STEP 3.

		//Compute U
		std::set<T*> setC;
		for(typename std::map<size_t, T>::iterator it = C.begin(); it != C.end(); it++)
			setC.insert(&(it->second));
		PointSetDistance<T> dts(*measure, setC);
		std::vector<T*> U(Q.begin(), Q.end());
		typename std::vector<T*>::iterator U_half = U.begin() + U.size()/2;
		std::nth_element(U.begin(), U_half, U.end(), dts);
		U.erase(U_half++, U.end());

		//Assign points
		std::set<T*> setU(U.begin(), U.end());
		Tupel assignedU(irredKMeans(setU, 0, C.size(), C, Sum)); //use irredKMeans to assign points to centers

		//Compute irredKMeans(Q-U, ...)
		for(typename std::vector<T*>::iterator it = U.begin(); it != U.end(); it++)
			Q.erase(*it);
		Tupel step3result(irredKMeans(Q, m, k, C, Sum));

		//Merge
		step3result.cost += assignedU.cost;
		for(size_t cluster = 0; cluster < assignedU.ds->partitions[0].size(); cluster++)
		{
			for(size_t point = 0; point < assignedU.ds->partitions[0][cluster].size(); point++)
				step3result.ds->partitions[0][cluster].push_back(assignedU.ds->partitions[0][cluster][point]);
			step3result.ds->proxysets[0].push_back(assignedU.ds->proxysets[0][cluster]);
		}
		delete assignedU.ds;

		//Return minimum result
		if(step2result.cost < step3result.cost)
		{
			delete step3result.ds;
			return step2result;
		}
		else
		{
			delete step2result.ds;
			return step3result;
		}
	}
	else
	{
		//No centers so far? No STEP 3.
		return step2result;
	}
}

template<typename T> double KumarSabharwalSen<T>::getLastCost()
{
	return lastCost;
}

template<typename T> void KumarSabharwalSen<T>::setInput(std::vector<T*> const* data)
{
	if(data != 0)
		input = std::set<T*>(data->begin(), data->end());
}

template<typename T> void KumarSabharwalSen<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(this->measure == 0)
		delete this->measure;

	this->measure = measure==0 ? 0 : measure->clone();
}

template<typename T> void KumarSabharwalSen<T>::setCentroidGenerator(ProxyGenerator<T> const *centroidGenerator)
{
	if(this->centroidGenerator == 0)
		delete this->centroidGenerator;

	this->centroidGenerator = centroidGenerator==0 ? 0 : centroidGenerator->clone();
}

template<typename T> void KumarSabharwalSen<T>::setSetSizes(unsigned int superSetSize, unsigned int subsetSize)
{
	if(subsetSize >= superSetSize)
		throw InvalidArgumentException(0, "subsetSize has to be smaller than superSetSize.", "subsetSize");

	this->superSetSize = superSetSize;
	this->subsetSize = subsetSize;
}


template<typename T> KumarSabharwalSen<T>* KumarSabharwalSen<T>::toKumarSabharwalSen(Algorithm* s)
{
	return dynamic_cast<KumarSabharwalSen<T>*>(s);
}

}

#endif
