#ifndef LLOYDTYPE_H
#define LLOYDTYPE_H

#include "../base/algorithm.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../base/proxygenerator.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/solutionprovider.h"
#include "../datastructure/doublesolution.h"

#include "../exception/invalidruntimeconfigurationexception.h"

#include <set>
#include <ctime>
#include <vector>

namespace CluE
{

/**
* @brief Lloyd type algorithm
* 
* @ingroup clustering_algorithms
*/
template<typename T>
class Lloydtype : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	/**
	* @brief Constructor initializing input, seeding the centers and configuring the algorithm
	* @remarks Remember to set seeding.
	*/
	Lloydtype(std::vector<T*> const* data = NULL, ProxyGenerator<T>* g = NULL,
		DissimilarityMeasure<T>* m = NULL, unsigned int maxIterations = 0);

	/**
	* @brief Constructor initializing input, seeding the centers and configuring the algorithm
	* @param seeds seed for centers as vector of objects
	*/
	Lloydtype(std::vector<T> const& seeds, std::vector<T*> const* data = NULL,
		ProxyGenerator<T>* g = NULL, DissimilarityMeasure<T>* m = NULL, unsigned int maxIterations = 0);

	/**
	* @overload
	*/
	Lloydtype(std::vector<T*> const& seeds, std::vector<T*> const* data = NULL,
		ProxyGenerator<T>* g = NULL, DissimilarityMeasure<T>* m = NULL, unsigned int i = 0);

	Lloydtype(const Lloydtype<T>&);
	Lloydtype<T>& operator=(const Lloydtype<T>&);
	virtual ~Lloydtype();

	/**
	 * @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	 * @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	 * @throw InvalidRuntimeConfigurationException [2] Proxy generator is NULL.
	 * @throw InvalidRuntimeConfigurationException [3] Empty seeding.
	 */
	virtual DoubleSolution<T>* compute();
	
	virtual void setInput(std::vector<T*> const*);
	void setSeeding(std::vector<T> const&);
	void setSeeding(std::vector<T*> const&);
	void setGenerator(ProxyGenerator<T>*);
	virtual void setMeasure(DissimilarityMeasure<T> const*);

	/**
	* @brief does a dynamic cast of the given Algorithm to Lloydtype<T> 
	* @return NULL if the Algorithm is not a Lloydtype<T> instance
	*/
	static Lloydtype<T>* toLloydtype(Algorithm* a);
	
private:
	std::vector<T*> const* input;
	std::vector<T> seeding;
	ProxyGenerator<T>* generator;
	DissimilarityMeasure<T>* measure;
	unsigned int maxiterations;
};

template<typename T> Lloydtype<T>::Lloydtype(std::vector<T*> const* data,
	ProxyGenerator<T>* g, DissimilarityMeasure<T>* m, unsigned int i) : input(data),
	generator(g==NULL?NULL:g->clone()), measure(m==NULL?NULL:m->clone()), maxiterations(i)
{
}

template<typename T> Lloydtype<T>::Lloydtype(std::vector<T> const& seeds,
	std::vector<T*> const* data, ProxyGenerator<T>* g, DissimilarityMeasure<T>* m,
	unsigned int i) : seeding(seeds), input(data), generator(g==NULL?NULL:g->clone()),
	measure(m==NULL?NULL:m->clone()), maxiterations(i)
{
}

template<typename T> Lloydtype<T>::Lloydtype(std::vector<T*> const& seeds,
	std::vector<T*> const* data, ProxyGenerator<T>* g, DissimilarityMeasure<T>* m, 
	unsigned int i) : input(data), generator(g==NULL?NULL:g->clone()),
	measure(m==NULL?NULL:m->clone()), maxiterations(i)
{
	this->setSeeding(seeds);
}

template<typename T> Lloydtype<T>::Lloydtype(const Lloydtype<T>& rhs) :
	Algorithm(rhs), input(rhs.input), seeding(rhs.seeding),
	generator(rhs.generator==NULL?NULL:rhs.generator->clone()),
	measure(rhs.measure==NULL?NULL:rhs.measure->clone()), maxiterations(rhs.maxiterations)
{	
}

template<typename T> Lloydtype<T>& Lloydtype<T>::operator=(const Lloydtype<T>& rhs)
{
	Algorithm::operator=(rhs);
	ProxyGenerator<T>* pg = rhs.generator==NULL?NULL:rhs.generator->clone();
	DissimilarityMeasure<T>* dm = rhs.measure==NULL?NULL:rhs.measure->clone();
	delete this->generator;
	this->generator = pg;
	delete this->measure;
	this->measure = dm;
	this->input = rhs.input;
	this->seeding = rhs.seeding;
	this->maxiterations = rhs.maxiterations;
	return *this;
}

template<typename T> Lloydtype<T>::~Lloydtype()
{
	delete this->measure;
	delete this->generator;
}

template<typename T> DoubleSolution<T>* Lloydtype<T>::compute()
{
	if (this->input==NULL)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if (this->generator==NULL)
		throw InvalidRuntimeConfigurationException(2, "Proxy generator is NULL.");
	if (this->measure==NULL)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");
	
	time_t start, end;
	start = time(0);

	DoubleSolution<T>* solution = new DoubleSolution<T>();

	unsigned int k = this->seeding.size();
	if (k==0)
		throw InvalidRuntimeConfigurationException(3, "Empty seeding.");
	
	std::set<unsigned int> altered_clusters;
	for (unsigned int i=0; i<k; i++)
		altered_clusters.insert(i);

	std::vector<std::vector<T*> > partitionzero(k);
	partitionzero[0] = *this->input;
	solution->partitions.push_back(partitionzero);
	solution->proxysets.push_back(this->seeding);

	bool in_motion;
	unsigned int iterations = 0;
	do
	{
		iterations++;
		in_motion = false;

		// reassign all elements to the nearest of the current proxies
		std::vector<std::vector<T*> > previous = solution->partitions[0];
		solution->partitions[0] = std::vector<std::vector<T*> >(k);
		//		int di = 0;
		for (unsigned int i=0; i<k; i++)
		{
			//			cout << "i=" << i << std::endl;

			unsigned int size = previous[i].size();
			for (unsigned int j=0; j<size; j++)
			{
				//				cout << "(i,j)=(" << i << "," << j << ")" << std::endl;

				double min = this->measure->dissimilarity(*previous[i][j], solution->proxysets[0][i]);
				unsigned int c = i;
				for (unsigned int l=0;l<k;l++)
				{
					double d = this->measure->dissimilarity(*previous[i][j], solution->proxysets[0][l]);
					if (d < min)
					{
						min = d;
						c = l;
					}
				}

				if (c!=i)
				{
					in_motion = true;
					altered_clusters.insert(i);
					altered_clusters.insert(c);
				}
				solution->partitions[0][c].push_back(previous[i][j]);
				//cout << ++di << ". element reassigned" << std::endl;
			}
		}

		/// compute new centers for non-empty altered clusters
		for (std::set<unsigned int>::iterator iter=altered_clusters.begin(); iter!=altered_clusters.end(); ++iter)
			if (solution->partitions[0][*iter].size()>0)
				solution->proxysets[0][*iter] = this->generator->generate(solution->partitions[0][*iter]);
				
		//std::clog << "CluE::LloydTypee<T>::compute() - " << altered_clusters.size()
		//	<< " altered clusters in iteration " << iterations << std::endl;
		altered_clusters.clear();
		
		if (this->maxiterations>0 && iterations>=this->maxiterations)
		{
			in_motion = false;
			//std::clog << "CluE::LloydTypee<T>::compute() - maximum number of iterations "
			//	<< altered_clusters.size() << " is reached" << std::endl;
		}
		
	} while (in_motion);

	end = time(0);
	solution->seconds=end-start;
	//std::clog << "CluE::LloydType<T>::compute() - finished after " << iterations << " iterations" << std::endl;
	return solution;
}

template<typename T> void Lloydtype<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void Lloydtype<T>::setSeeding(std::vector<T> const& seeds)
{
	this->seeding = seeds;
}

template<typename T> void Lloydtype<T>::setSeeding(std::vector<T*> const& seeds)
{
	this->seeding.clear();
	unsigned int size = seeds.size();
	for (unsigned int i=0;i<size;i++)
		this->seeding.push_back(*seeds[i]);
}

template<typename T> void Lloydtype<T>::setGenerator(ProxyGenerator<T>* g)
{
	this->generator = g==NULL?NULL:g->clone();
}

template<typename T> void Lloydtype<T>::setMeasure(DissimilarityMeasure<T> const* m)
{
	this->measure = m==NULL?NULL:m->clone();
}

template<typename T> Lloydtype<T>* Lloydtype<T>::toLloydtype(Algorithm* a)
{
	return dynamic_cast<Lloydtype<T>*>(a);
}
		
}

#endif
