#ifndef LloydtypeCFCF_H
#define LloydtypeCFCF_H

#include "../base/algorithm.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../base/proxygenerator.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/solutionprovider.h"
#include "../datastructure/doublesolution.h"
#include "../exception/invalidruntimeconfigurationexception.h"
#include "../base/euclideanspaceprovider.h"

#include "cfrentry.h"

#include <set>
#include <ctime>
#include <vector>

namespace CluE
{

/**
 * @brief LloydtypeCF specific DoubleSolution
 */
template<typename T>
class LloydtypeCFSolution : public DoubleSolution<T>
{
public:
	LloydtypeCFSolution() :
		DoubleSolution<T>(),
		costs(0)
	{

	}

	double costs;
};

/**
* @brief Lloyd type algorithm using clustering features
*
* @ingroup clustering_algorithms
*/
template<typename T>
class LloydtypeCF : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	/**
	* @brief Constructor initializing input, seeding the centers and configuring the algorithm
	* @remarks Remember to set seeding.
	*/
	LloydtypeCF(std::vector<T*> const* data = NULL, EuclideanSpaceProvider<T> const * s = NULL,
	            DissimilarityMeasure<T>* m = NULL, unsigned int maxIterations = 0);

	/**
	* @brief Constructor initializing input, seeding the centers and configuring the algorithm
	* @param seeds seed for centers as vector of objects
	*/
	LloydtypeCF(std::vector<T> const& seeds, std::vector<T*> const* data = NULL,
	            EuclideanSpaceProvider<T> const * s = NULL, DissimilarityMeasure<T>* m = NULL, unsigned int maxIterations = 0);

	/**
	* @overload
	*/
	LloydtypeCF(std::vector<T*> const& seeds, std::vector<T*> const* data = NULL,
	            EuclideanSpaceProvider<T> const * s = NULL, DissimilarityMeasure<T>* m = NULL, unsigned int i = 0);

	LloydtypeCF(const LloydtypeCF<T>&);
	LloydtypeCF<T>& operator=(const LloydtypeCF<T>&);
	virtual ~LloydtypeCF();

	/**
	 * @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	 * @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	 * @throw InvalidRuntimeConfigurationException [2] Euclidean space provider is NULL.
	 * @throw InvalidRuntimeConfigurationException [3] Empty seeding.
	 */
	virtual LloydtypeCFSolution<T>* compute();

	virtual void setInput(std::vector<T*> const*);
	void setSeeding(std::vector<T> const&);
	void setSeeding(std::vector<T*> const&);
	void setEuclideanSpaceProvider(EuclideanSpaceProvider<T> const *);
	virtual void setMeasure(DissimilarityMeasure<T> const*);

	/**
	* @brief does a dynamic cast of the given Algorithm to LloydtypeCF<T>
	* @return NULL if the Algorithm is not a LloydtypeCF<T> instance
	*/
	static LloydtypeCF<T>* toLloydtypeCF(Algorithm* a);

private:
	std::vector<T*> const* input;
	std::vector<T> seeding;
	EuclideanSpaceProvider<T>* espace;
	DissimilarityMeasure<T>* measure;
	unsigned int maxiterations;
};

template<typename T> LloydtypeCF<T>::LloydtypeCF(std::vector<T*> const* data,
        EuclideanSpaceProvider<T> const * s, DissimilarityMeasure<T>* m, unsigned int i) : input(data),
	espace(s==NULL?NULL:s->clone()), measure(m==NULL?NULL:m->clone()), maxiterations(i)
{
}

template<typename T> LloydtypeCF<T>::LloydtypeCF(std::vector<T> const& seeds,
        std::vector<T*> const* data, EuclideanSpaceProvider<T> const * s, DissimilarityMeasure<T>* m,
        unsigned int i) : seeding(seeds), input(data), espace(s==NULL?NULL:s->clone()),
	measure(m==NULL?NULL:m->clone()), maxiterations(i)
{
}

template<typename T> LloydtypeCF<T>::LloydtypeCF(std::vector<T*> const& seeds,
        std::vector<T*> const* data, EuclideanSpaceProvider<T> const * s, DissimilarityMeasure<T>* m,
        unsigned int i) : input(data), espace(s==NULL?NULL:s->clone()),
	measure(m==NULL?NULL:m->clone()), maxiterations(i)
{
	this->setSeeding(seeds);
}

template<typename T> LloydtypeCF<T>::LloydtypeCF(const LloydtypeCF<T>& rhs) :
	Algorithm(rhs), input(rhs.input), seeding(rhs.seeding),
	espace(rhs.espace==NULL?NULL:rhs.espace->clone()),
	measure(rhs.measure==NULL?NULL:rhs.measure->clone()), maxiterations(rhs.maxiterations)
{
}

template<typename T> LloydtypeCF<T>& LloydtypeCF<T>::operator=(const LloydtypeCF<T>& rhs)
{
	Algorithm::operator=(rhs);
	ProxyGenerator<T>* pg = rhs.generator==NULL?NULL:rhs.generator->clone();
	DissimilarityMeasure<T>* dm = rhs.measure==NULL?NULL:rhs.measure->clone();
	delete this->generator;
	this->generator = pg;
	delete this->measure;
	this->measure = dm;
	this->input = rhs.input;
	this->seeding = rhs.seeding;
	this->maxiterations = rhs.maxiterations;
	return *this;
}

template<typename T> LloydtypeCF<T>::~LloydtypeCF()
{
	delete this->measure;
	delete this->espace;
}

template<typename T> LloydtypeCFSolution<T>* LloydtypeCF<T>::compute()
{
	if (this->input==NULL)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if (this->measure==NULL)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");
	if (this->espace==NULL)
		throw InvalidRuntimeConfigurationException(2, "Euclidean space provider is NULL.");

	time_t start, end;
	start = time(0);

	unsigned int k = this->seeding.size();
	if (k==0)
		throw InvalidRuntimeConfigurationException(3, "Empty seeding.");

	// Create CFREntry for each cluster
	T nullVector = espace->nullVector();
	double nullVectorSq = nullVector * nullVector;
	std::vector<CFREntry<T>> clusters(k, CFREntry<T>(0, nullVector, nullVectorSq, nullVector));
	for(size_t i = 0; i < k; ++i)
		clusters[i].representative = seeding[i];

	bool in_motion;
	unsigned int iterations = 0;
	std::vector<size_t> assignment(input->size(), -1);
	do
	{
		iterations++;
		in_motion = false;

		// Assign nearest center (from last iteration) to each input point
		for(size_t i = 0; i < input->size(); ++i)
		{
			double minDist = std::numeric_limits<double>::infinity();
			size_t minIndex = 0;
			for(size_t j = 0; j < k; ++j)
			{
				double tmpDist = measure->dissimilarity(*(*input)[i], clusters[j].representative);
				if(tmpDist < minDist)
				{
					minDist = tmpDist;
					minIndex = j;
				}
			}

			// Check for change in assignment
			if(assignment[i] != minIndex)
			{
				if(assignment[i] != -1)
					clusters[assignment[i]].remove(*(*input)[i]);
				clusters[minIndex].insert(*(*input)[i]);
				assignment[i] = minIndex;
				in_motion = true;
			}
		}


		// Break after reassignment of all points to the centers of the maxiterations. iteration
		if (this->maxiterations > 0 && iterations > this->maxiterations)
		{
			// Current iteration is maxiterations+1

			in_motion = false;

			//std::clog << "CluE::LloydtypeCFe<T>::compute() - maximum number of iterations "
			//	<< altered_clusters.size() << " is reached" << std::endl;
		}
		
		// Don't calculate new centers when not in motion any more
		if(!in_motion)
			break;

		// Calculate new centers
		for(size_t i = 0; i < k; ++i)
			clusters[i].representative = clusters[i].cog();
	}
	while (in_motion);
	end = time(0);

	LloydtypeCFSolution<T>* solution = new LloydtypeCFSolution<T>();
	// Proxies
	std::vector<T> proxies;
	proxies.resize(k);
	for(size_t i = 0; i < k; ++i)
		proxies[i] = clusters[i].representative;
	solution->proxysets.push_back(proxies);
	// Partitions
	std::vector<std::vector<T*>> partitions(k, std::vector<T*>());
	for(size_t i = 0; i < input->size(); ++i)
		partitions[assignment[i]].push_back((*input)[i]);
	solution->partitions.push_back(partitions);
	// Costs
	for(size_t i = 0; i < k; ++i)
		solution->costs += clusters[i].kMeansCost(clusters[i].cog());
	// Time
	solution->seconds=end-start;
	//std::clog << "CluE::LloydtypeCF<T>::compute() - finished after " << iterations << " iterations" << std::endl;
	return solution;
}

template<typename T> void LloydtypeCF<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void LloydtypeCF<T>::setSeeding(std::vector<T> const& seeds)
{
	this->seeding = seeds;
}

template<typename T> void LloydtypeCF<T>::setSeeding(std::vector<T*> const& seeds)
{
	this->seeding.clear();
	unsigned int size = seeds.size();
	for (unsigned int i=0; i<size; i++)
		this->seeding.push_back(*seeds[i]);
}

template<typename T> void LloydtypeCF<T>::setMeasure(DissimilarityMeasure<T> const* m)
{
	this->measure = m==NULL?NULL:m->clone();
}

template<typename T> void LloydtypeCF<T>::setEuclideanSpaceProvider(EuclideanSpaceProvider<T> const * m)
{
	this->espace = m==NULL?NULL:m->clone();
}

template<typename T> LloydtypeCF<T>* LloydtypeCF<T>::toLloydtypeCF(Algorithm* a)
{
	return dynamic_cast<LloydtypeCF<T>*>(a);
}

}

#endif
