#ifndef PROXYBASEDAGGLO_H
#define PROXYBASEDAGGLO_H

#include "../base/algorithm.h"
#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/proxygenerator.h"
#include "../base/solutionprovider.h"
#include "../datastructure/doublesolution.h"

#include <vector>

namespace CluE
{

/**
 * @brief Agglomerative proxy-based clustering algorithm 
 * 
 * Algorithm<T> template implementing an agglomerative clustering algorithm based on the
 * computation and comparison of proxies for the computed clusters.
 * 
 * @ingroup clustering_algorithms
 */
template<typename T> class ProxyBasedAgglo : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:

	/**
	 * @brief On demand computation / pre-computation
	 * 
	 * This enum is used to tell the algorithm via the constructor if the dissimilarities should be
	 * computed on demand everytime they are needed or if they should be precomputed, held in memory
	 * and be updated whenever a dissimilarity may have changed.
	 */
	enum ComputationMode
	{
		OnDemand,	/// slow, needs no extra memory
		Precompute	/// fast, needs more memory
	};

	ProxyBasedAgglo(std::vector<T*> const* = NULL, ProxyGenerator<T>* = NULL,
		DissimilarityMeasure<T>* = NULL, ComputationMode = OnDemand);
	
	ProxyBasedAgglo(const ProxyBasedAgglo<T>&);
	ProxyBasedAgglo<T>& operator=(const ProxyBasedAgglo<T>&);
	virtual ~ProxyBasedAgglo();
	
	/**
	 * @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	 * @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	 * @throw InvalidRuntimeConfigurationException [2] Proxy generator is NULL.
	 */
	virtual DoubleSolution<T>* compute();
	
	virtual void setInput(std::vector<T*> const*);
	void setGenerator(ProxyGenerator<T>*);
	virtual void setMeasure(DissimilarityMeasure<T>*);
	void setMode(int);

	/**
	* @brief Does a dynamic cast of the given Algorithm to ProxyBasedAgglo<T>.
	* @return NULL if the Algorithm is not a ProxyBasedAgglo<T> instance
	*/
	static ProxyBasedAgglo<T>* toProxyBasedAgglo(Algorithm* s);
	
private:
	std::vector<T*> const* input;
	ProxyGenerator<T>* generator;
	DissimilarityMeasure<T>* measure;
	int mode;
};

template<typename T> ProxyBasedAgglo<T>::ProxyBasedAgglo(std::vector<T*> const* data,
	ProxyGenerator<T>* g, DissimilarityMeasure<T>* m, ComputationMode cm) : input(data),
	generator(g==NULL?NULL:g->clone()), measure(m==NULL?NULL:m->clone()), mode(cm)
{
}

template<typename T> ProxyBasedAgglo<T>::ProxyBasedAgglo(const ProxyBasedAgglo<T>& rhs) :
	Algorithm(rhs), input(rhs.input), generator(rhs.generator==NULL?NULL:rhs.generator->clone()),
	measure(rhs.measure==NULL?NULL:rhs.measure->clone()), mode(rhs.mode)
{	
}

template<typename T> ProxyBasedAgglo<T>& ProxyBasedAgglo<T>::operator=(
	const ProxyBasedAgglo<T>& rhs)
{
	Algorithm::operator=(rhs);
	ProxyGenerator<T>* pg = rhs.generator==NULL?NULL:rhs.generator->clone();
	DissimilarityMeasure<T>* dm = rhs.measure==NULL?NULL:rhs.measure->clone();
	delete this->generator;
	this->generator = pg;
	delete this->measure;
	this->measure = dm;
	this->input = rhs.input;
	this->mode = rhs.mode;
	return *this;
}

template<typename T> ProxyBasedAgglo<T>::~ProxyBasedAgglo()
{
	delete this->generator;
	delete this->measure;
}

template<typename T> DoubleSolution<T>* ProxyBasedAgglo<T>::compute()
{
	if (this->input==NULL)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if (this->generator==NULL)
		throw InvalidRuntimeConfigurationException(2, "ProxyGenerator is NULL.");
	if (this->measure==NULL)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");
	
	clock_t start, end;
	start = clock();

	DoubleSolution<T>* solution = new DoubleSolution<T>();

	unsigned int N = this->input->size();

	// create a first partition (partitions[0]) with one cluster for input data object
	std::clog << "CluE::ProxyBasedAgglo<T>::compute() - starting with " << N << " singleton clusters..." << std::endl;
	std::vector<std::vector<T*> > partitionzero;
	for (unsigned int i=0; i<N; i++)
	{
		std::vector<T*> singleton;
		singleton.push_back((*this->input)[i]);
		partitionzero.push_back(singleton);
	}
	solution->partitions.push_back(partitionzero);

	// compute a proxy for each cluster of the initial partition
	std::vector<T> zeroproxies;
	for (unsigned int i=0; i<N; i++)
		zeroproxies.push_back(generator->generate(solution->partitions[0][i]));
	solution->proxysets.push_back(zeroproxies);

	// precompute dissimilarities
	double* dis = NULL;

	if (this->mode==Precompute)
	{
		std::clog << "CluE::ProxyBasedAgglo<T>::compute() - precomputing dissimilarities..." << std::endl;
		dis = new double[N*N];

		for (unsigned int i=0; i<N; i++)
			for (unsigned int j=0; j<N; j++)
			{
				dis[N*i+j] = this->measure->dissimilarity(solution->proxysets[0][i], solution->proxysets[0][j]);
			}
	}

	for (unsigned int n=1; n<N; n++)
	{
		std::clog << "CluE::ProxyBasedAgglo<T>::compute() - currently computing round " << n << std::endl;
		// find the two clusterproxies with minimum dissimilarity
		int first = -1;
		int second = -1;
		double min = std::numeric_limits<double>::infinity();
		for (unsigned int i=0; i<N-n+1; i++)
			for (unsigned int j=0; j<N-n+1; j++)
			{
				//cout << "i=" << i << ", j=" << j << std::endl;
				if (i!=j)
				{
					double d;
					if (this->mode==Precompute)
						d = dis[N*i+j];
					else
						d = this->measure->dissimilarity(solution->proxysets[n-1][i], solution->proxysets[n-1][j]);

					if (d < min)
					{
						min = d;
						first = i;
						second = j;
					}
				}
			}

		if (second<first)
		{
			int i = first;
			first = second;
			second = i;
		}
		//cout << "merging cluster " << first << " and " << second << " (min. dissimilarity of "
		//     << min << " between " << solution->proxysets[n-1][first] << " and "
		//     << solution->proxysets[n-1][second] << std::endl;

		// merge found clusters at next granularity (merged cluster at index first and last cluster moved to index second)
		solution->partitions.push_back(solution->partitions[n-1]);
		while (!solution->partitions[n][second].empty())
		{
			solution->partitions[n][first].push_back(solution->partitions[n][second].back());
			solution->partitions[n][second].pop_back();
		}
		solution->partitions[n][second] = solution->partitions[n].back();
		solution->partitions[n].pop_back();

		// compute proxies for merged cluster at next granularity
		solution->proxysets.push_back(solution->proxysets[n-1]);
		solution->proxysets[n][first] = generator->generate(solution->partitions[n][first]);
		solution->proxysets[n][second] = solution->proxysets[n].back();
		solution->proxysets[n].pop_back();

		if (this->mode==Precompute)
		{
			// update dissimilarities
			for (unsigned int i=0; i<N-n; i++)
			{
				dis[N*second+i] = dis[N*(N-n)+i];
				dis[N*i+second] = dis[N*i+(N-n)];
			}

			for (unsigned int i=0; i<N-n; i++)
			{
				dis[N*first+i] = this->measure->dissimilarity(solution->proxysets[n][first], solution->proxysets[n][i]);
				dis[N*i+first] = this->measure->dissimilarity(solution->proxysets[n][i], solution->proxysets[n][first]);
			}
		}
	}

	end = clock();
	solution->seconds=double(end-start)/CLOCKS_PER_SEC;
	std::clog << "CluE::ProxyBasedAgglo<T>::compute() - finished" << std::endl;
	return solution;
}

template<typename T> void ProxyBasedAgglo<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void ProxyBasedAgglo<T>::setGenerator(ProxyGenerator<T>* g)
{
	this->generator = g==NULL?NULL:g->clone();
}

template<typename T> void ProxyBasedAgglo<T>::setMeasure(DissimilarityMeasure<T>* m)
{
	this->measure = m==NULL?NULL:m->clone();
}

template<typename T> void ProxyBasedAgglo<T>::setMode(int m)
{
	this->mode = m;
}

template<typename T> ProxyBasedAgglo<T>* ProxyBasedAgglo<T>::toProxyBasedAgglo(Algorithm* s)
{
	return dynamic_cast<ProxyBasedAgglo<T>*>(s);
}
	
}

#endif
