#ifndef FRAHLINGSOHLER_H
#define FRAHLINGSOHLER_H

#include <memory>
#include <unordered_map>

#include "../clustering/cfentry.h"
#include "../base/streamingalgorithm.h"
#include "../base/weightmodifier.h"
#include "../datastructure/proxysolution.h"
#include "../datastreams/distinctelementsprimitive.h"
#include "../hashing/hashfunction.h"
#include "../hashing/carterwegman.h"

namespace CluE
{

/**
 * @brief Coreset algorithm for k-median clustering
 *
 * G. Frahling, C. Sohler: Coresets in dynamic geometric data streams. STOC 2005.
 * @warning This implementation needs more testing.
 *
 * @ingroup coreset_algorithms
 */
template<typename VectorType, typename Hash, typename size_space>
class FrahlingSohler : public StreamingAlgorithm<VectorType>
{
public:
  /**
   * @brief Set of grids
   */
	struct GridInstance
	{
    /**
     * @brief A single grid
     */
		struct Grid
		{
			Grid(unsigned long long cellsPerRow, unsigned long long numberOfCells, CarterWegman<unsigned long long, size_t> bucketHash, CarterWegman<unsigned long long, unsigned long long> sampleHash) :
				cellsPerRow(cellsPerRow),
				numberOfCells(numberOfCells),
				distinctCells(0,0,0),
				bucketHash(bucketHash),
				sampleHash(sampleHash),
				buckets()
			{
			}

			unsigned long long cellsPerRow;
			unsigned long long numberOfCells;
			DistinctElementsPrimitive<unsigned long long, unsigned long long> distinctCells;
			CarterWegman<unsigned long long, size_t> bucketHash;
			CarterWegman<unsigned long long, unsigned long long> sampleHash;
			std::unordered_map<size_t, CFEntry<VectorType>> buckets;
		};

		unsigned long opt;
		std::vector<Grid> subgrids;
	};

	FrahlingSohler(DiscreteBoundedSpace<VectorType, size_space> * space, WeightModifier<VectorType> * weightModifier, std::function<Hash()> createHashfunction, double eps, double rho, unsigned int numOfCenters);

	void insert(VectorType const & element);
	void remove(VectorType const & element);

	virtual FrahlingSohler<VectorType, Hash, size_space>& operator<<(VectorType const & element);

	virtual ProxySolution<VectorType>* compute();

private:
	void shiftGrid(std::vector<size_space>& cell);
	void toParentCell(std::vector<size_space>& cell);
	std::vector<std::vector<size_space>> getChildCells(std::vector<size_space> cell);
	unsigned long long getCellIndex(typename GridInstance::Grid const & grid, std::vector<size_space>& cell, size_t layer);
	bool evaluateGrid(std::vector<VectorType>& coreset, size_t gridInstance, size_t gridIndex, std::vector<size_space> cell);
	bool isHeavy(size_t gridInstance, size_t gridIndex, size_t index);

	std::unique_ptr<DiscreteBoundedSpace<VectorType, size_space>> space;
	std::unique_ptr<WeightModifier<VectorType>> weightModifier;
	std::vector<GridInstance> grids;
	VectorHash<size_t> vechash;
	double eps;
	double rho;
	unsigned int numOfCenters; // = k
	double alpha;
	double delta;
	size_t dimension;
	unsigned long long nFitting; // = smallest 2^i >= n
	size_t numOfNestedLayers;
};

template<typename VectorType, typename Hash, typename size_space>
FrahlingSohler<VectorType, Hash, size_space>::FrahlingSohler(DiscreteBoundedSpace<VectorType, size_space> * space, WeightModifier<VectorType> * weightModifier, std::function<Hash()> createHashfunction, double eps, double rho, unsigned int numOfCenters) :
	space(space->clone()),
	weightModifier(weightModifier->clone()),
	grids(),
	vechash(),
	eps(eps),
	rho(rho),
	numOfCenters(numOfCenters),
	alpha(6 * std::pow(eps, -2) * std::log(std::pow(rho, -1)) + 1),
	delta(1000), //TODO delta
	dimension(space->dimension()),
	nFitting(std::pow(std::ceil(std::log2(space->n())), 2))
{
	// Number of Opt candidates
	size_t jBound = std::ceil((space->n()+1) * std::log2(space->dimension()));
	// Number of counter-hash-function-buckets
	size_t numOfBuckets = 3 * std::pow(16 * alpha / delta + numOfCenters * std::pow(2, dimension), 2);
	// Number of nested grid layers
	numOfNestedLayers = std::ceil(std::log2(space->n()))+1;

	// Prepare Opt candidate grids
	grids.resize(jBound);
	for(size_t j = 0; j < jBound; ++j)
	{
		// Prepare nested grids of an Opt candidate grid
		GridInstance& grid = grids[j];
		grid.opt = std::pow(2, j);
		grid.subgrids.reserve(numOfNestedLayers);
		for(size_t g = 0; g < numOfNestedLayers; ++g)
		{
			unsigned long long sampleBuckets = delta * std::pow(2, double(j) - double(g)) / alpha;
			if(sampleBuckets == 0) // Construct at least one bucket
				throw "TODO delta to small";
			unsigned long long cellsPerRow = std::ceil(nFitting / std::pow(2, g+1));
			unsigned long long numberOfCells = std::ceil(std::pow(nFitting / std::pow(2, g+1), dimension));
			CarterWegman<unsigned long long, size_t> bucketHash(numberOfCells, numOfBuckets);
			CarterWegman<unsigned long long, unsigned long long> sampleHash(std::pow(nFitting, dimension), sampleBuckets);
			grid.subgrids.push_back(typename GridInstance::Grid(cellsPerRow, numberOfCells, bucketHash, sampleHash));
		}
	}
}

template<typename VectorType, typename Hash, typename size_space>
FrahlingSohler<VectorType, Hash, size_space>& FrahlingSohler<VectorType, Hash, size_space>::operator<<(VectorType const & element)
{
	//TODO Insertion / deletion streaming
}

template<typename VectorType, typename Hash, typename size_space>
void FrahlingSohler<VectorType, Hash, size_space>::insert(VectorType const & element)
{
	// Get cell coordinates and consecutive index
	std::vector<size_space> cell(space->getCoordinates(element));
	unsigned long long originalIndex = getCellIndex(grids[0].subgrids[0], cell, 0);
	shiftGrid(cell); // Apply random shift
	// Iterate over all grid instances
	for(size_t i = 0; i < grids.size(); ++i)
	{
		GridInstance & instance = grids[i];
		// Iterate over all nested grids
		for(size_t g = 0; g < instance.subgrids.size(); ++g)
		{
			typename GridInstance::Grid & grid = instance.subgrids[g];
			if(grid.sampleHash(originalIndex) == 0) // Take element into account?
			{
				unsigned long long index = getCellIndex(grid, cell, g);
				grid.distinctCells << index;
				size_t bucket = grid.bucketHash(index);
				if(grid.buckets.count(index) == 0)
				{
					// Construct bucket if empty
					VectorType origin = space->origin();
					grid.buckets.insert(std::pair<unsigned long long, CFEntry<VectorType>>(index, CFEntry<VectorType>(0, origin, origin*origin)));
				}
				// Update the bucket's clustering feature
				CFEntry<VectorType> cfelement(1, element, element*element);
				grid.buckets.at(index) += cfelement;
				// Get parent cell
			}
		}
	}
}

template<typename VectorType, typename Hash, typename size_space>
void FrahlingSohler<VectorType, Hash, size_space>::remove(VectorType const & element)
{
	// Get cell coordinates and consecutive index
	std::vector<size_space> cell(space->getCoordinates(element));
	unsigned long long originalIndex = getCellIndex(grids[0].subgrids[0], cell, 0);
	shiftGrid(cell); // Apply random shift
	// Iterate over all grid instances
	for(size_t i = 0; i < grids.size(); ++i)
	{
		GridInstance & instance = grids[i];
		// Iterate over all nested grids
		for(size_t g = 0; g < instance.subgrids.size(); ++g)
		{
			typename GridInstance::Grid & grid = instance.subgrids[g];
			if(grid.sampleHash(originalIndex) == 0) // Take element into account?
			{
				unsigned long long index = getCellIndex(grid, cell, g);
				size_t bucket = grid.bucketHash(index);
				// Update the bucket's clustering feature
				CFEntry<VectorType> cfelement(1, element, element*element);
				grid.buckets.at(index) -= cfelement;
				if(grid.buckets.at(index).number == 0) // Destruct bucket if empty
					grid.buckets.erase(index);
				// Get parent cell
			}
		}
	}
}

template<typename VectorType, typename Hash, typename size_space>
ProxySolution<VectorType>* FrahlingSohler<VectorType, Hash, size_space>::compute()
{
	std::vector<size_space> startCell(dimension, 0);
	for(size_t i = 0; i < grids.size(); ++i)
	{
		std::vector<VectorType> coreset;
		bool success = evaluateGrid(coreset, i, numOfNestedLayers-1, startCell);
		if(success)
		{
			ProxySolution<VectorType>* solution = new ProxySolution<VectorType>();
			solution->proxysets.push_back(coreset);
			return solution;
		}
	}
}

template<typename VectorType, typename Hash, typename size_space>
bool FrahlingSohler<VectorType, Hash, size_space>::evaluateGrid(std::vector<VectorType>& coreset, size_t gridInstance, size_t gridIndex, std::vector<size_space> cell)
{
	bool success = true;
	double threshold = 16 * alpha / delta + numOfCenters * std::pow(2, dimension);
	unsigned long long distElements = grids[gridInstance].subgrids[gridIndex].distinctCells.numberOfDistinctElements();
	// Check for hash function collisions
	if(distElements <= threshold)
	{
		bool mark = true; // "Marked cell" indicator
		if(gridIndex < grids[gridInstance].subgrids.size())
		{
			// Mark a cell iff it is heavy but contains no heavy subcells
			std::vector<std::vector<size_space>> children(getChildCells(cell));
			for(size_t i = 0; i < children.size(); ++i)
			{
				unsigned long long index = getCellIndex(grids[gridInstance].subgrids[gridIndex-1], children[i], gridIndex);
				mark &= !isHeavy(gridInstance, gridIndex-1, index);
			}
			if(!mark)
			{
				// Decide whether to add or move a representative point
				unsigned long long toMove = 0;
				for(size_t i = 0; i < children.size(); ++i)
				{
					unsigned long long index = getCellIndex(grids[gridInstance].subgrids[gridIndex-1], children[i], gridIndex);
					if(isHeavy(gridInstance, gridIndex+1, index)) // Add
						evaluateGrid(coreset, gridInstance, gridIndex-1, children[i]);
					else // Move
						toMove += grids[gridInstance].subgrids[gridIndex-1].buckets.at(index).number;
				}
				VectorType& last = coreset[coreset.size()-1];
				weightModifier->setWeight(last, weightModifier->getWeight(last) + toMove);
			}
		}
		if(mark)
		{
			unsigned long long index = getCellIndex(grids[gridInstance].subgrids[gridIndex], cell, gridIndex);
			CFEntry<VectorType>& cf = grids[gridInstance].subgrids[gridIndex].buckets.at(index);
			VectorType point((1.0 / cf.number) * cf.LS);
			weightModifier->setWeight(point, cf.number);
			coreset.push_back(point);
		}
	}
	else
	{
		success = false;
	}
	return success;
}

template<typename VectorType, typename Hash, typename size_space>
bool FrahlingSohler<VectorType, Hash, size_space>::isHeavy(size_t gridInstance, size_t gridIndex, size_t index)
{
	if(grids[gridInstance].subgrids[gridIndex].buckets.count(index) > 0)
		return grids[gridInstance].subgrids[gridIndex].buckets.at(index).number >=
		       delta * grids[gridInstance].opt / std::pow(2, gridIndex+1);
	else
		return false;
}

template<typename VectorType, typename Hash, typename size_space>
void FrahlingSohler<VectorType, Hash, size_space>::shiftGrid(std::vector<size_space>& cell)
{
	//TODO Random shift
}

template<typename VectorType, typename Hash, typename size_space>
void FrahlingSohler<VectorType, Hash, size_space>::toParentCell(std::vector<size_space>& cell)
{
	for(size_t i = 0; i < cell.size(); ++i)
		cell[i] /= 2;
}

template<typename VectorType, typename Hash, typename size_space>
std::vector<std::vector<size_space>> FrahlingSohler<VectorType, Hash, size_space>::getChildCells(std::vector<size_space> cell)
{
	std::vector<std::vector<size_space>> cells;
	for(size_t i = 0; i < cell.size(); ++i)
		cell[i] *= 2;
	for(size_t i = 0; i < std::pow(2, dimension); ++i)
	{
		size_t inc = i;
		std::vector<size_space> childCell(cell);
		for(size_t j = 0; j < childCell.size(); ++j)
			childCell[j] += (inc & 1);
		cells.push_back(childCell);
		inc >>= 1;
	}
	return cells;
}

template<typename VectorType, typename Hash, typename size_space>
unsigned long long FrahlingSohler<VectorType, Hash, size_space>::getCellIndex(typename GridInstance::Grid const & grid, std::vector<size_space>& cell, size_t layer)
{
	unsigned long long index = 0;
	unsigned long long power = 1;
	for(size_t d = 0; d < dimension; ++d)
	{
		size_space coord = cell[d];
		for(size_t l = 0; l < layer; ++l)
			coord /= 2;
		index += coord * power;
		power *= grid.cellsPerRow;
	}
	return index;
}

}
#endif
