#ifndef DOUBLESOLUTION_H
#define DOUBLESOLUTION_H

#include "../base/solutionprovider.h"
#include "../base/proxyprovider.h"
#include "../base/partitionprovider.h"

#include <vector>

//TODO Derived documentation bug

namespace CluE
{

/**
 * @brief Data structure for partitions and proxies.
 *
 * This struct is for use in algorithms computing proxies.
 * 
 * @ingroup data_structures
 */
template<typename T> struct DoubleSolution : public SolutionProvider, public ProxyProvider<T>,
	public PartitionProvider<T>
{
public:
	DoubleSolution();

	virtual ~DoubleSolution()
	{
	}

	virtual double computationtime() const;
	virtual unsigned int number_of_solutions() const;
	virtual unsigned int size_of_solution(unsigned int) const;
	
	virtual T proxy(unsigned int n, unsigned int c) const;
	virtual std::vector<T> proxies(unsigned int n) const;

	virtual unsigned int clustersize(unsigned int, unsigned int) const;
	virtual T* element(unsigned int, unsigned int, unsigned int) const;
	virtual std::vector<T*> cluster(unsigned int, unsigned int) const;
	virtual std::vector<std::vector<T*> > clustering(unsigned int) const;

	double seconds;
	std::vector<std::vector<T> > proxysets;
	std::vector<std::vector<std::vector<T*> > > partitions;
};

template<typename T> DoubleSolution<T>::DoubleSolution() : seconds()
{
}

template<typename T> double DoubleSolution<T>::computationtime() const
{
	return seconds;
}

template<typename T> unsigned int DoubleSolution<T>::number_of_solutions() const
{
	unsigned int proxynum = this->proxysets.size();
	unsigned int partitionnum = this->partitions.size();
	return proxynum<partitionnum?proxynum:partitionnum;
}

template<typename T> unsigned int DoubleSolution<T>::size_of_solution(unsigned int i) const
{
	if (i<this->number_of_solutions())
	{
		unsigned int proxysize = this->proxysets[i].size();
		unsigned int partitionsize = this->partitions[i].size();
		return proxysize<partitionsize?proxysize:partitionsize;
	}
	return 0;
}

template<typename T> T DoubleSolution<T>::proxy(unsigned int n, unsigned int c) const
{
	if (n<this->proxysets.size())
		if (c<this->proxysets[n].size())
				return this->proxysets[n][c];
				
	std::cerr << "DoubleSolution<T>::proxy(" << n << "," << c << "): requested proxy not available"
		<< std::endl;
	throw "ILLEGAL STATE";
}

template<typename T> std::vector<T> DoubleSolution<T>::proxies(unsigned int n) const
{
	if (n<this->proxysets.size())
			return this->proxysets[n];
	return std::vector<T>();
}

template<typename T> unsigned int DoubleSolution<T>::clustersize(unsigned int g, unsigned int c) const
{
	if (g<this->partitions.size())
		if (c<this->partitions[g].size())
			return this->partitions[g][c].size();
	return 0;
}

template<typename T> T* DoubleSolution<T>::element(unsigned int g, unsigned int c, unsigned int n) const
{
	if (g<this->partitions.size())
		if (c<this->partitions[g].size())
			if (n<this->partitions[g][c].size())
				return this->partitions[g][c][n];
	return NULL;
}

template<typename T> std::vector<T*> DoubleSolution<T>::cluster(unsigned int g, unsigned int c) const
{
	if (g<this->partitions.size())
		if (c<this->partitions[g].size())
				return this->partitions[g][c];
	return std::vector<T*>();
}

template<typename T> std::vector<std::vector<T*> > DoubleSolution<T>::clustering(unsigned int g) const
{
	if (g<this->partitions.size())
			return this->partitions[g];
	return std::vector<std::vector<T*> >();
}

}

#endif
