#ifndef KCENTEREVALUAR_H
#define KCENTEREVALUAR_H

#include "../base/measuresetter.h"
#include "../base/partitionprovider.h"
#include "../base/proxyevaluation.h"
#include "../base/combinedevaluation.h"
#include "../base/dissimilaritymeasure.h"

namespace CluE
{

//TODO all=max(elem) or all=sum(elem) ?

/**
 * @brief Calculates the k-center weight.
 * @ingroup evaluation_algorithms
 */
template<typename T> class KCenterEvaluator : public ProxyEvaluation<T>, public CombinedEvaluation<T>, public MeasureSetter<T>
{
public:

	/**
	 * @brief Instantiates KCenterEvaluator, optionally with a DissimilarityMeasure to use when calculation the maximum diameter.
	 * @param measure Optional. Nevertheless, you have to set a DissimilarityMeasure before using this class.
	 * @see setMeasure
	 */
	KCenterEvaluator(DissimilarityMeasure<T> const *measure = 0);

	KCenterEvaluator(const KCenterEvaluator<T>&);
	KCenterEvaluator<T>& operator= (const KCenterEvaluator<T>&);
	virtual ~KCenterEvaluator();

	/**
	 * @brief Sets the DissimilarityMeasure used when calculating the radius.
	 */
	virtual void setMeasure(DissimilarityMeasure<T> const *measure);

//# ProxyEvaluation members
	/**
	 * @brief Assigns all points to a proxy and calculates the k-center weight of the resulting clustering.
	 */
	virtual double proxycost(std::vector<T*> const& points, std::vector<T> const& proxies) const;

	/**
	 * @overload
	 */
	virtual double proxycost(std::vector<T*> const& points, std::vector<T*> const& proxies) const;

	/**
	 * @brief Assigns all points to a proxy (provided by proxySource) and calculates the k-center weight of the resulting clustering.
	 */
	virtual double proxycost(std::vector<T*> const& points, ProxyProvider<T> const &proxySource, unsigned int solutionIndex) const;

	/**
	 * @overload
	 */
	virtual double proxycost(std::vector<T*> const& points, DiscreteProxyProvider<T> const &proxySource, unsigned int solutionIndex) const;


	/**
	 * @brief Assigns all points to a proxy and calculates the k-center weight of the cluster[index].
	 */
	virtual double proxycost(std::vector<T*> const& points, std::vector<T> const& proxies, unsigned int index) const;

	/**
	 * @overload
	 */
	virtual double proxycost(std::vector<T*> const& points, std::vector<T*> const& proxies, unsigned int index) const;

	/**
	 * @brief Assigns all points to a proxy (provided by proxySource) and calculates the k-center weight of the cluster[index].
	 */
	virtual double proxycost(std::vector<T*> const& points, ProxyProvider<T> const &proxySource, unsigned int solutionIndex, unsigned int proxyIndex) const;

	/**
	 * @overload
	 */
	virtual double proxycost(std::vector<T*> const& points, DiscreteProxyProvider<T> const &proxySource, unsigned int solutionIndex, unsigned int proxyIndex) const;

//# CombinedEvaluation members
	/**
	 * @brief Calculates the k-center weight of a given clustering.
	 */
	virtual double combinedcost(std::vector<std::vector<T*> > const& clusters, std::vector<T> const& proxies) const;

	/**
	 * @overload
	 */
	virtual double combinedcost(std::vector<std::vector<T*> > const& clusters, std::vector<T*> const& proxies) const;


	/**
	 * @brief Calculates the k-center weight of a given clustering (provided by clusteringSource and proxySource).
	 */
	virtual double combinedcost(PartitionProvider<T> const &clusteringSource, ProxyProvider<T> const &proxySource, unsigned int solutionIndex) const;

	/**
	 * @overload
	 */
	virtual double combinedcost(PartitionProvider<T> const &clusteringSource, DiscreteProxyProvider<T> const &proxySource, unsigned int solutionIndex) const;


	/**
	 * @brief Calculates the k-center weight of a given proxy and the corresponding points.
	 */
	virtual double combinedcost(std::vector<T*> const& cluster, T const& proxy) const;

	/**
	 * @brief Calculates the k-center weight of a given proxy (provided by proxySource) and the corresponding points (provided by clusteringSource).
	 */
	virtual double combinedcost(PartitionProvider<T> const &clusteringSource, ProxyProvider<T> const &proxySource, unsigned int solutionIndex, unsigned int proxyIndex) const;

	/**
	 * @overload
	 */
	virtual double combinedcost(PartitionProvider<T> const &clusteringSource, DiscreteProxyProvider<T> const &proxySource, unsigned int solutionIndex, unsigned int proxyIndex) const;

protected:
	/**
	 * @brief Provides a k-center weight result per cluster (may be added, chosen from, ...).
	 */
	std::vector<double> proxycostGeneric(std::vector<T*> const& points, std::vector<T> const& proxies) const;

	/**
	 * @overload
	 */
	std::vector<double> proxycostGeneric(std::vector<T*> const& points, std::vector<T*> const& proxies) const;

private:
	DissimilarityMeasure<T>* measure;
};

template<typename T> KCenterEvaluator<T>::KCenterEvaluator(DissimilarityMeasure<T> const *measure) :
	measure(measure==0 ? 0 : measure->clone())
{
	//empty
}

template<typename T> KCenterEvaluator<T>::KCenterEvaluator(const KCenterEvaluator<T>& kce) :
	measure(kce.measure == 0 ? 0 : kce.measure->clone())
{
}

template<typename T> KCenterEvaluator<T>& KCenterEvaluator<T>::operator= (const KCenterEvaluator<T>& kce)
{
	if(measure != 0)
		delete measure;

	ProxyEvaluation<T>::operator= (kce);
	CombinedEvaluation<T>::operator= (kce);

	measure = kce.measure == 0 ? 0 : kce.measure->clone();

	return *this;
}

template<typename T> KCenterEvaluator<T>::~KCenterEvaluator()
{
	if(measure != 0)
		delete measure;
}

//### ProxyEvaluation members

template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, std::vector<T> const& proxies) const
{
	std::vector<double> values = proxycostGeneric(points, proxies);
	int numOfValues = values.size();

	double result = 0;
	for(int i = 0; i < numOfValues; i++)
	{
		if(values[i] > result)
			result = values[i];
	}

	return result;
}

template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, std::vector<T*> const& proxies) const
{
	std::vector<double> values = proxycostGeneric(points, proxies);
	int numOfValues = values.size();

	double result = 0;
	for(int i = 0; i < numOfValues; i++)
	{
		if(values[i] > result)
			result = values[i];
	}

	return result;
}

template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, ProxyProvider<T> const &provider, unsigned int solutionIndex) const
{
	return proxycost(points, provider.proxies(solutionIndex));
}

template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, DiscreteProxyProvider<T> const &provider, unsigned int solutionIndex) const
{
	return proxycost(points, provider.discrete_proxies(solutionIndex));
}


template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, std::vector<T> const& proxies, unsigned int index) const
{
	std::vector<double> values = proxycostGeneric(points, proxies);
	return values.at(index);
}

template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, std::vector<T*> const& proxies, unsigned int index) const
{
	std::vector<double> values = proxycostGeneric(points, proxies);
	return values.at(index);
}

template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, ProxyProvider<T> const &provider, unsigned int solutionIndex, unsigned int proxyIndex) const
{
	return proxycost(points, provider.proxies(solutionIndex), proxyIndex);
}

template<typename T> double KCenterEvaluator<T>::proxycost(std::vector<T*> const& points, DiscreteProxyProvider<T> const &provider, unsigned int solutionIndex, unsigned int proxyIndex) const
{
	return proxycost(points, provider.discrete_proxies(solutionIndex), proxyIndex);
}

template<typename T> void KCenterEvaluator<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(measure)
		this->measure = 0;
	else
		this->measure = measure->clone();
}


template<typename T> std::vector<double> KCenterEvaluator<T>::proxycostGeneric(std::vector<T*> const& points, std::vector<T> const& proxies) const
{
	int numOfPoints = points.size();
	int numOfProxies = proxies.size();

	std::vector<double> result(numOfProxies, 0);

	for(int i = 0; i < numOfPoints; i++)
	{
		T* point = points[i];

		double min = this->measure->dissimilarity(*point, proxies[0]);
		int assignedProxy = 0;
		for(int j = 1; j < numOfProxies; j++)
		{
			T proxy = proxies[j];
			double candidate = this->measure->dissimilarity(*point, proxy);
			if(candidate < min)
			{
				min = candidate;
				assignedProxy = j;
			}
		}

		if(min > result[assignedProxy])
			result[assignedProxy] = min;
	}

	return result;
}

template<typename T> std::vector<double> KCenterEvaluator<T>::proxycostGeneric(std::vector<T*> const& points, std::vector<T*> const& proxies) const
{
	int numOfPoints = points.size();
	int numOfProxies = proxies.size();

	std::vector<double> result(numOfProxies, 0);

	for(int i = 0; i < numOfPoints; i++)
	{
		T* point = points[i];

		double min = this->measure->dissimilarity(*point, *proxies[0]);
		int assignedProxy = 0;
		for(int j = 1; j < numOfProxies; j++)
		{
			T* proxy = proxies[j];
			double candidate = this->measure->dissimilarity(*point, *proxy);
			if(candidate < min)
			{
				min = candidate;
				assignedProxy = j;
			}
		}

		if(min > result[assignedProxy])
			result[assignedProxy] = min;
	}

	return result;
}

//### CombinedEvaluation members

template<typename T> double KCenterEvaluator<T>::combinedcost(std::vector<std::vector<T*> > const& clusters, std::vector<T> const& proxies) const
{
	double sum = 0;

	int numOfClusters = clusters.size();
	int numOfProxies = proxies.size();
	int minNumOfClustersProxies = numOfClusters < numOfProxies ? numOfClusters : numOfProxies;

	for(int i = 0; i < minNumOfClustersProxies; i++)
	{
		double cost = combinedcost(clusters[i], proxies[i]);
		if(cost > sum)
			sum = cost;
	}

	if(numOfClusters < numOfProxies)
	{
		std::clog << "CluE::KCenterEvaluator<T>::combinedcost(std::vector<std::vector<T*> >, std::vector<T>) - WARNING: More proxies than clusters: ignoring redundant proxies." << std::endl;
	}
	else if(numOfClusters > numOfProxies)
	{
		std::clog << "CluE::KCenterEvaluator<T>::combinedcost(std::vector<std::vector<T*> >, std::vector<T>) - WARNING: Less proxies than clusters: assigning remaining points to proxies." << std::endl;
		for(int i = numOfProxies; i < numOfClusters; i++)
		{
			double cost =  proxycost(clusters[i], proxies);
			if(cost > sum)
				sum=cost;
		}
	}

	return sum;
}

template<typename T> double KCenterEvaluator<T>::combinedcost(std::vector<std::vector<T*> > const& clusters, std::vector<T*> const& proxies) const
{
	double sum = 0;

	int numOfClusters = clusters.size();
	int numOfProxies = proxies.size();
	int minNumOfClustersProxies = numOfClusters < numOfProxies ? numOfClusters : numOfProxies;

	for(int i = 0; i < minNumOfClustersProxies; i++)
	{
		double cost = CombinedEvaluation<T>::combinedcost(clusters[i], proxies[i]);
		if(cost > sum)
			sum = cost;
	}

	if(numOfClusters < numOfProxies)
	{
		std::clog << "CluE::KCenterEvaluator<T>::combinedcost(std::vector<std::vector<T*> >, std::vector<T*>) - WARNING: More proxies than clusters: ignoring redundant proxies." << std::endl;
	}
	else if(numOfClusters > numOfProxies)
	{
		std::clog << "CluE::KCenterEvaluator<T>::combinedcost(std::vector<std::vector<T*> >, std::vector<T*>) - WARNING: Less proxies than clusters: assigning remaining points to proxies." << std::endl;
		for(int i = numOfProxies; i < numOfClusters; i++)
		{
			double cost =  proxycost(clusters[i], proxies);
			if(cost > sum)
				sum=cost;
		}
	}

	return sum;
}

template<typename T> double KCenterEvaluator<T>::combinedcost(PartitionProvider<T> const &clusterProvider, ProxyProvider<T> const &proxyProvider, unsigned int solutionIndex) const
{
	return combinedcost(clusterProvider.clustering(solutionIndex), proxyProvider.proxies(solutionIndex));
}

template<typename T> double KCenterEvaluator<T>::combinedcost(PartitionProvider<T> const &clusterProvider, DiscreteProxyProvider<T> const &proxyProvider, unsigned int solutionIndex) const
{
	return combinedcost(clusterProvider.clustering(solutionIndex), proxyProvider.discrete_proxies(solutionIndex));
}

template<typename T> double KCenterEvaluator<T>::combinedcost(std::vector<T*> const& cluster, T const& proxy) const
{
	double sum = 0;

	int numOfPoints = cluster.size();
	for(int i = 0; i < numOfPoints; i++)
	{
		double cost = this->measure->dissimilarity(*cluster[i], proxy);
		if(cost > sum)
			sum =cost;
	}

	return sum;
}

template<typename T> double KCenterEvaluator<T>::combinedcost(PartitionProvider<T> const &clusterProvider, ProxyProvider<T> const &proxyProvider, unsigned int solutionIndex, unsigned int proxyIndex) const
{
	return combinedcost(clusterProvider.cluster(solutionIndex, proxyIndex), proxyProvider.proxy(solutionIndex, proxyIndex));
}

template<typename T> double KCenterEvaluator<T>::combinedcost(PartitionProvider<T> const &clusterProvider, DiscreteProxyProvider<T> const &proxyProvider, unsigned int solutionIndex, unsigned int proxyIndex) const
{
	return CombinedEvaluation<T>::combinedcost(clusterProvider.cluster(solutionIndex, proxyIndex), proxyProvider.discrete_proxy(solutionIndex, proxyIndex));
}

}

#endif
