#ifndef MAXDIAMEVALUATOR_H
#define MAXDIAMEVALUATOR_H

#include "../base/measuresetter.h"
#include "../base/partitionevaluation.h"
#include "../base/dissimilaritymeasure.h"

namespace CluE
{

/**
 * @brief Calculates the maximum diameter of one or more given clusters.
 * @note Arguments named \e firstSecondThird may be shortened \e fST in formulas.
 * @note The metric is referred to as \e d(x,y) in formulas.
 * @ingroup evaluation_algorithms
 */
template<typename T> class MaxDiamEvaluator : public PartitionEvaluation<T>, public MeasureSetter<T>
{
public:

	/**
	 * @brief Instantiates MaxDiamEvaluator, optionally with a DissimilarityMeasure to use when calculating the maximum diameter.
	 * @param measure Optional. Nevertheless, you have to set a DissimilarityMeasure before using this class.
	 * @see setMeasure
	 */
	MaxDiamEvaluator(DissimilarityMeasure<T> const *measure = 0);

	MaxDiamEvaluator(const MaxDiamEvaluator<T>&);
	MaxDiamEvaluator<T>& operator= (const MaxDiamEvaluator<T>&);
	virtual ~MaxDiamEvaluator();

	/**
	 * @brief Calculates the maximum diameter of a given clustering.
	 * @return \f[ \max\limits_{c \in \mathtt{clusters}} (\max\limits_{p,q \in c} d(p,q)) \f]
	 */
	virtual double partitioncost(std::vector<std::vector<T*> > const& clusters) const;

	/**
	 * @brief Calculates the maximum diameter of a PartitionProvider's clustering solution.
	 * @return \f[ \max\limits_{c \in \mathtt{cS}_{\mathtt{sI}}} (\max\limits_{p,q \in c} d(p,q)) \f]
	 */
	virtual double partitioncost(PartitionProvider<T> const & clusteringSource, unsigned int solutionIndex) const;

	/**
	 * @brief Calculates the maximum diameter of a given vector of points.
	 * @return \f[ \max\limits_{p,q \in \mathtt{points}} d(p,q) \f]
	 */
	virtual double partitioncost(std::vector<T*> const& points) const;

	/**
	 * @brief Calculates the maximum diameter of a PartitionProvider's clustering solution cluster/partition.
	 * @return \f[ \max\limits_{p,q \in \mathtt{cS}_{\mathtt{sI},\mathtt{pI}}} d(p,q) \f]
	 */
	virtual double partitioncost(PartitionProvider<T> const & clusteringSource, unsigned int solutionIndex, unsigned int partitionIndex) const;


	/**
	 * @brief Sets the DissimilarityMeasure used when calculating the maximum diameter.
	 */
	virtual void setMeasure(DissimilarityMeasure<T> const *measure);

private:
	DissimilarityMeasure<T>* measure;
};

template<typename T> MaxDiamEvaluator<T>::MaxDiamEvaluator(DissimilarityMeasure<T> const *measure) :
	measure(measure==0 ? 0 : measure->clone())
{
	//empty
}

template<typename T> MaxDiamEvaluator<T>::MaxDiamEvaluator(const MaxDiamEvaluator<T>& mde) :
	measure(mde.measure == 0 ? 0 : mde.measure->clone())
{
}

template<typename T> MaxDiamEvaluator<T>& MaxDiamEvaluator<T>::operator= (const MaxDiamEvaluator<T>& mde)
{
	if(this.measure != 0)
		delete this.measure;

	PartitionEvaluation<T>::operator= (mde);

	this.measure = mde.measure == 0 ? 0 : mde.measure->clone();

	return *this;
}

template<typename T> MaxDiamEvaluator<T>::~MaxDiamEvaluator()
{
	if(this->measure != 0)
		delete this->measure;
}

template<typename T> double MaxDiamEvaluator<T>::partitioncost(std::vector<std::vector<T*> > const &clusters) const
{
	double sum = 0;

	int size = clusters.size();
	for(int i = 0; i < size; i++)
	{
		sum += partitioncost(clusters[i]);
	}

	return sum;
}

template<typename T> double MaxDiamEvaluator<T>::partitioncost(PartitionProvider<T> const &provider, unsigned int solutionIndex) const
{
	return partitioncost(provider.clustering(solutionIndex));
}

template<typename T> double MaxDiamEvaluator<T>::partitioncost(std::vector<T*> const& points) const
{
	double max = 0;

	int size = points.size();
	for(int i = 0; i < size; i++)
	{
		for(int j = 0; j < i; j++)
		{
			double candidate = measure->dissimilarity(*points[i], *points[j]);
			if(max < candidate)
				max = candidate;
		}
	}

	return max;
}

template<typename T> double MaxDiamEvaluator<T>::partitioncost(PartitionProvider<T> const &provider, unsigned int solutionIndex, unsigned int partitionIndex) const
{
	return partitioncost(provider.cluster(solutionIndex, partitionIndex));
}

template<typename T> void MaxDiamEvaluator<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(measure == 0)
		this->measure = 0;
	else
		this->measure = measure->clone();
}

}

#endif
