#include "../frequency/frequencydistribution.h"

#include <vector>
#include <iostream>
#include <math.h>
#include <float.h>
#include <algorithm>
#include <numeric>
#include <limits>

using namespace CluE;

/**
* @brief constructs the centroid of the given vector of @ref FrequencyDistribution objects.
*/
FrequencyDistribution::FrequencyDistribution(std::vector<unsigned long> const& f)
{
	size_t fSize = f.size();
	unsigned long fSum = accumulate(f.begin(), f.end(), 0);
	this->frequencies = f;
	for(size_t i = 0; i < fSize; i++)
		probabilities.push_back(double(f[i]) / fSum);
	this->frequencysum = fSum;
}

/**
* @brief constructs optimal consolidated probability distribution
*/
FrequencyDistribution::FrequencyDistribution(std::vector<FrequencyDistribution*> const& distributions)
{
	this->frequencysum = 0;

	unsigned int num = distributions.size();
	unsigned int length = 0;
	for (unsigned int i=0; i<num; i++)
	{
		unsigned int l = distributions[i]->size();
		if (l>length)
			length=l;
	}
	this->frequencies.assign(length, 0);
	
	for (unsigned int i=0; i<num; i++)
	{
		unsigned int l = distributions[i]->size();
		for (unsigned int j=0; j<l; j++)
		{
			this->frequencies[j] += distributions[i]->frequencies[j];
       	}
	}
	this->frequencysum = accumulate(this->frequencies.begin(), this->frequencies.end(), 0);
    
	size_t fSize = this->frequencies.size();
	for(size_t i = 0; i < fSize; i++)
		probabilities.push_back(double(this->frequencies[i]) / this->frequencysum);
	
    if (num==0)
    	std::clog << "FrequencyDistribution::FrequencyDistribution - WARNING: returning empty frequency distribution" << std::endl;
}

unsigned int FrequencyDistribution::size() const
{
	return this->frequencies.size();
}

unsigned long FrequencyDistribution::frequency(unsigned int n) const
{
	if (n<this->frequencies.size())
		return this->frequencies[n];
	return 0;
}

unsigned long FrequencyDistribution::total() const
{
	return this->frequencysum;
}

double FrequencyDistribution::probability(unsigned int i) const
{
	if (i<this->frequencies.size())
		return probability(i);

   	std::clog << "FrequencyDistribution::probability - WARNING: returning zero, index out of bound" << std::endl;
	return .0;
}

double FrequencyDistribution::entropy() const
{
	if (this->frequencysum==0)
		return .0;
		
	std::vector<double> temp;
	for(unsigned int i=0;i < this->frequencies.size(); i++)
	{
		if (this->frequencies[i]>0)
		{
			double p = probability(i);
			if (p>.0)
				temp.push_back(p*log(p));
		}
	}
	
	std::sort(temp.begin(), temp.end());
	return -accumulate(temp.begin(), temp.end(), .0)/log(2);
}

double FrequencyDistribution::kullbackleibler(FrequencyDistribution const& fd) const
{
	unsigned int length = this->frequencies.size();
	unsigned int fdlength = fd.frequencies.size();
	if (fdlength!=length)
    {
		std::cerr << "FrequencyDistribution::kullbackleibler - ERROR: frequency distributions have different lengths! ("
		     << fdlength << "!=" << length << ")" << std::endl;
  		return std::numeric_limits<double>::quiet_NaN();
  	}
  	
  	if (this->frequencysum==0)
  		return .0;

  	if (fd.frequencysum==0)
  		return std::numeric_limits<double>::infinity();

  	std::vector<double> temp;
  	for(unsigned int i=0; i<length; i++)
  	{
  		if (this->frequencies[i]>0)
  		{
  			if (fd.frequencies[i]==0)
  				return std::numeric_limits<double>::infinity();
  			double p = probability(i);
  			if (p>.0)
  			{
  				double q = fd.probability(i);
  				if (q>.0)
  				{
  					double r = p/q;
  					if (r>.0)
  						temp.push_back(p*log(r));
  				} else return std::numeric_limits<double>::infinity();
  			}
  		}
  	}
  	std::sort(temp.begin(), temp.end());
  	return accumulate(temp.begin(), temp.end(), .0)/log(2);
 }
 
 std::ostream& operator << (std::ostream& os, FrequencyDistribution& fd)
 {
 	unsigned int dimension = fd.size();
    os << "FrequencyDistribution(";
    for (unsigned int i=0;i<dimension;i++)
    {
        os << fd.frequency(i) << ", ";
    }
    os << "sum=" << fd.total() << ")";
    return os;
 }
