#include <set>

#include "primegenerator.h"

namespace CluE
{

unsigned long long PrimeGenerator::getPrime(unsigned long long lowerBound, unsigned long long upperBound, double errorProbability)
{
	if(lowerBound < 2)
		lowerBound = 2;
	if(lowerBound > upperBound)
		throw "TODO";
	if(errorProbability <= 0 || errorProbability > 0.25)
		throw "TODO";

	std::set<unsigned long long> checkedNumbers;
	RandomGenerator gen = Randomness::getRandomGenerator();
    std::uniform_int_distribution<unsigned long long> dis(lowerBound, upperBound);
	
	// Sample from the given interval and check if the number is prime
	while(checkedNumbers.size() < upperBound - lowerBound + 1)
	{
		unsigned long long candidate = 0;
		while(candidate == 0 || checkedNumbers.count(candidate) > 0)
			candidate = dis(gen);
		checkedNumbers.insert(candidate);
		
		if(isMillerRabinPrime(candidate, errorProbability))
			return candidate;
	}
	
	throw "TODO";
}

bool PrimeGenerator::isMillerRabinPrime(unsigned long long candidate, double errorProbability)
{
	if(candidate == 2)
		return true;
	else if((candidate & 1) == 0) // mod 2
		return false;
	
	if(errorProbability <= 0 || errorProbability > 0.25)
		throw "TODO";

	// Write candidate = 2^s * r + 1
	unsigned long long r = candidate-1;
	int s = 0;
	while((r & 1) == 0) // mod 2
	{
		s++;
		r /= 2;
	}
	
	// Random number generator
	std::set<unsigned long long> usedNumbers;
	RandomGenerator gen = Randomness::getRandomGenerator();
    std::uniform_int_distribution<unsigned long long> dis(2, candidate-1);
	
	double currentErrorProbability = 0.5;
	for(unsigned long i = 1; errorProbability < currentErrorProbability; i++)
	{
		// Sample an unused number
		unsigned long long randomBase = 0;
		while(randomBase == 0 || usedNumbers.count(randomBase) > 0)
			randomBase = dis(gen);
		usedNumbers.insert(randomBase);
		
		// Calculate y=randomBase^r mod candidate
		unsigned long long y = potMod(randomBase, r, candidate);
		
		// 1-Condition derived from Femat's little theorem
		bool onlyOnes = (y == 1 || y == candidate-1);
		
		// Calculate y^2 mod candidate, (y^2 mod candidate)^2 mod candidate, ...
		for(int j = 1; j <= s; j++)
		{
			y = (y*y) % candidate;
			if(onlyOnes && y != 1)
				return false;
			else if(!onlyOnes)
			{
				if(y == 1)
					return false;
				else
					onlyOnes = (y == candidate-1);
			}
		}
		if(y != 1)
			return false;
		
		if(i >= candidate/4)
			return true;
		currentErrorProbability *= 0.5;
	}
	
	return true;
}

unsigned long long PrimeGenerator::potMod(unsigned long long x, unsigned long long k, unsigned long long m)
{
	unsigned long long result = 1;
	unsigned long long base = x;
	while(k > 0)
	{
		if((k & 1) == 1)
			result = (result * base) % m;
		k >>= 1;
		base = (base * base) % m;
	}
	
	return result;
}

}