#ifndef SINGLELINKAGE_H
#define SINGLELINKAGE_H

#include "../base/dissimilaritymeasure.h"
#include "../base/clusterdissimilaritymeasure.h"

#include <vector>
#include <limits>

namespace CluE
{

/**
 * @brief Single cluster linkage
 */
template<typename T> class SingleLinkage : public ClusterDissimilarityMeasure<T>
{
public:
	SingleLinkage(const DissimilarityMeasure<T>& measure);

	SingleLinkage(const SingleLinkage<T>&);
	SingleLinkage<T>& operator=(const SingleLinkage<T>&);
	virtual ~SingleLinkage();

	virtual SingleLinkage<T>* clone() const;

	/**
	* @brief Computes the single linkage dissimilarity between the two given clusters.
	*
	* This function computes the dissimilarities between all (ordered) pairs of points from the
	* two given clusters and returns the minimum.
	*/
	virtual double dissimilarity(std::vector<T*> const& c1, std::vector<T*> const& c2);

private:
	DissimilarityMeasure<T>* measure;
};

template<typename T> SingleLinkage<T>::SingleLinkage(const DissimilarityMeasure<T>& m) :
	measure(m.clone())
{
}

template<typename T> SingleLinkage<T>::SingleLinkage(const SingleLinkage<T>& rhs) :
	ClusterDissimilarityMeasure<T>(rhs), measure(rhs.measure->clone())
{	
}

template<typename T> SingleLinkage<T>& SingleLinkage<T>::operator=(const SingleLinkage<T>& rhs)
{
	ClusterDissimilarityMeasure<T>::operator=(rhs);
	DissimilarityMeasure<T>* dm = rhs.measure->clone();
	delete this->measure;
	this->measure = dm;
	return *this;
}

template<typename T> SingleLinkage<T>::~SingleLinkage()
{
	delete this->measure;
}

template<typename T> SingleLinkage<T>* SingleLinkage<T>::clone() const
{
	return new SingleLinkage<T>(*this);
}

template<typename T> double SingleLinkage<T>::dissimilarity(std::vector<T*> const& c1,
	std::vector<T*> const& c2)
{
	double max = std::numeric_limits<double>::infinity();
	for (unsigned int i=0; i<c1.size(); i++)
		for (unsigned int j=0; j<c2.size(); j++)
		{
			double cost = this->measure->dissimilarity(*c1[i], *c2[j]);
			if (cost < max)
				max = cost;
			cost = this->measure->dissimilarity(*c2[j], *c1[i]);
			if (cost < max)
				max = cost;
		}
	return max;
}

}

#endif
