#ifndef DYNAMICBITSET_H
#define DYNAMICBITSET_H

#include <vector>

namespace CluE
{
	
/**
 * @brief Dynamic bitset similiar to boost::dynamic_bitset
 * 
 * T is the integral type maintaining a bunch of bits,
 * i.e. sizeof(T)*8 bits.
 */
template<typename T = long long> class DynamicBitset
{
static_assert(std::is_integral<T>::value, "T has to be an integral type");

public:
	/**
	 * @brief Constructs a bitset which maintains numberOfBits bits.
	 * 
	 * @param numberOfBits The number of bits.
	 */
	DynamicBitset(unsigned long long numberOfBits);
	
	/**
	 * @brief Read access to the (position+1)-th bit
	 * @param Index of bit (starting with 0)
	 */
	bool operator[](size_t position) const;
	
	/**
	 * @brief Write access to the (position+1)-th bit
	 * 
	 * @param position Index of bit (starting with 0)
	 * @param value Value to write
	 */
	void set(size_t position, bool value);
	
	/**
	 * @brief Number of set bits
	 */
	unsigned long long count();
	
private:
	const unsigned long long numberOfBits;
	const int sizeOfT;
	std::vector<T> bitvector;
};


template<typename T> DynamicBitset<T>::DynamicBitset(unsigned long long numberOfBits) :
	numberOfBits(numberOfBits),
	sizeOfT(sizeof(T) * 8),
	bitvector(numberOfBits / sizeOfT + (numberOfBits % sizeOfT == 0 ? 0 : 1))
{
}

template<typename T> bool DynamicBitset<T>::operator [](size_t position) const
{
	size_t index = position / sizeOfT;
	int offset = position % sizeOfT;
	return (bitvector[index] >> offset) & 1;
}

template<typename T> void DynamicBitset<T>::set(size_t position, bool value)
{
	size_t index = position / sizeOfT;
	int offset = position % sizeOfT;
	if(value)
		bitvector[index] |=   T(1) << offset;
	else
		bitvector[index] &= ~(T(1) << offset);
}

//TODO Improve performance
template<typename T> unsigned long long DynamicBitset<T>::count()
{
	unsigned long long setBits = 0;
	for(unsigned long long i = 0; i < numberOfBits; i++)
	{
		if((*this)[i])
			setBits++;
	}
	return setBits;
}

}

#endif
