#ifndef FIXEDSIZESUBSETITERATOR_H
#define FIXEDSIZESUBSETITERATOR_H

#include <cmath>
#include <stdexcept>

#include "../exception/invalidargumentexception.h"
#include "../misc/setiterator.h"

namespace CluE
{
/**
 * @brief Iterates over all fixed-size subsets of a given superset.
 *
 * Generates all fixed-size subsets of a given superset.
 * The superset's size is @b not limited by the number of bits of the largest integral data type.
 * 
 * @ingroup helper_classes
 */
template<class T> class FixedSizeSubsetIterator : public SetIterator<T>
{
public:
	/**
	 * @brief Provide the superset as a set.
	 * @throw InvalidArgumentException [0] Value of size is invalid! size.
	 */
	FixedSizeSubsetIterator(std::set<T*> superset, size_t size);

	/**
	 * @brief Provide the superset as a vector.
	 * @throw InvalidArgumentException [0] Value of size is invalid! size.
	 */
	FixedSizeSubsetIterator(std::vector<T*> superset, size_t size);

	/**
	 * @brief Size of the current subset.
	 * @remarks Use this method instead of set().size() when you do not need the set itself!
	 */
	virtual size_t size() const;

	/**
	 * @brief Generates the next subset.
	 * @throw std::out_of_range No remaining subsets.
	 */
	virtual void next();

	/**
	 * @brief Returns if the Graycode sequence is incomplete yet or not.
	 */
	virtual bool hasMore() const;

	/**
	 * @brief Returns the current subset as a set.
	 */
	virtual std::set<T*> set() const;

	/**
	 * @brief Returns the current subset as a vector.
	 * @remarks set() is more efficient.
	 */
	virtual std::vector<T*> vector() const;

private:
	void buildSubsetIndices(std::vector<size_t> &buildVector);

	std::vector<T*> supersetVector;
	std::vector<std::vector<size_t> > subsetIndices;
	size_t fixedSize;
	size_t currentPosition;
};

template<class T> FixedSizeSubsetIterator<T>::FixedSizeSubsetIterator(std::set<T*> superset, size_t size) :
	fixedSize(size),
	currentPosition(0),
	supersetVector(superset.begin(), superset.end())
{
	if(size == 0 || size >= superset.size())
		throw InvalidArgumentException(0, "Value of size is invalid!", "size");
	
	std::vector<size_t> empty;
	buildSubsetIndices(empty);
}

template<class T> FixedSizeSubsetIterator<T>::FixedSizeSubsetIterator(std::vector<T*> superset, size_t size) :
	supersetVector(superset),
	fixedSize(size),
	currentPosition(0)
{
	if(size == 0 || size > superset.size())
		throw InvalidArgumentException(0, "Value of size is invalid!", "size");
	
	std::vector<size_t> empty;
	buildSubsetIndices(empty);
}

template<class T> void FixedSizeSubsetIterator<T>::buildSubsetIndices(std::vector<size_t> &buildVector)
{
	size_t start = 0;
	if(buildVector.size() > 0)
		start = buildVector[buildVector.size()-1]+1;

	for(; start < supersetVector.size(); start++)
	{
		buildVector.push_back(start);
		if(buildVector.size() < fixedSize)
			buildSubsetIndices(buildVector);
		else
			subsetIndices.push_back(buildVector);
		buildVector.erase(buildVector.end()-1);
	}
}

template<class T> size_t FixedSizeSubsetIterator<T>::size() const
{
	return subsetIndices[currentPosition].size();
}

template<class T> void FixedSizeSubsetIterator<T>::next()
{
	if(currentPosition < subsetIndices.size()-1)
		currentPosition++;
	else
		throw std::out_of_range("No more subsets.");
}

template<class T> bool FixedSizeSubsetIterator<T>::hasMore() const
{
	return currentPosition < subsetIndices.size()-1;
}

template<class T> std::set<T*> FixedSizeSubsetIterator<T>::set() const
{
	std::set<T*> result;
	size_t n = subsetIndices[currentPosition].size();
	for(size_t i = 0; i < n; i++)
		result.insert(supersetVector[subsetIndices[currentPosition][i]]);
		
	return result;
}

template<class T> std::vector<T*>  FixedSizeSubsetIterator<T>::vector() const
{
	std::vector<T*> result;
	size_t n = subsetIndices[currentPosition].size();
	for(size_t i = 0; i < n; i++)
		result.push_back(supersetVector[subsetIndices[currentPosition][i]]);
		
	return result;
}

}

#endif