#ifndef SUBSETITERATOR_H
#define SUBSETITERATOR_H

#include <stdexcept>

#include "../misc/setiterator.h"
#include "../misc/flippinggraycodebits.h"

namespace CluE
{
/**
 * @brief Iterates over all subsets of a given superset.
 *
 * Generates all subsets of a given superset.
 * The superset's size is @b not limited by the number of bits of the largest integral data type.
 * 
 * @ingroup helper_classes
 */
template<class T> class SubsetIterator : public SetIterator<T>
{
public:
	/**
	 * @brief Provide the superset as a set.
	 */
	SubsetIterator(std::set<T*> superset);

	/**
	 * @brief Provide the superset as a vector.
	 */
	SubsetIterator(std::vector<T*> superset);

	/**
	 * @brief Size of the current subset.
	 * @remarks Use this method instead of set().size() when you do not need the set itself!
	 */
	virtual size_t size() const;

	/**
	 * @brief Generates the next subset.
	 * @throw std::out_of_range No remaining subsets.
	 */
	virtual void next();

	/**
	 * @brief Returns if the Graycode sequence is incomplete yet or not.
	 */
	virtual bool hasMore() const;

	/**
	 * @brief Returns the current subset as a set.
	 */
	virtual std::set<T*> set() const;

	/**
	 * @brief Returns the current subset as a vector.
	 * @remarks set() is more efficient.
	 */
	virtual std::vector<T*> vector() const;

private:
	GraycodeSequence gcs;
	std::vector<T*> supersetVector;
	size_t supersetVectorSize;
	std::set<T*> currentSubset;
	size_t currentSize;
};

template<class T> SubsetIterator<T>::SubsetIterator(std::set<T*> superset) :
	supersetVectorSize(superset.size()),
	currentSize(0),
	supersetVector(superset.begin(), superset.end())
{
}

template<class T> SubsetIterator<T>::SubsetIterator(std::vector<T*> superset) :
	supersetVector(superset),
	supersetVectorSize(superset.size()),
	currentSize(0)
{
}

template<class T> size_t SubsetIterator<T>::size() const
{
	return currentSize;
}

template<class T> void SubsetIterator<T>::next()
{
	size_t element = gcs.current();
	if(element < supersetVectorSize)
	{
		if(currentSubset.find(supersetVector[element]) != currentSubset.end())
		{
			currentSubset.erase(supersetVector[element]);
			currentSize--;
		}
		else
		{
			currentSubset.insert(supersetVector[element]);
			currentSize++;
		}
		gcs.next();
	}
	else
	{
		throw std::out_of_range("No more subsets.");
	}
}

template<class T> bool SubsetIterator<T>::hasMore() const
{
	if(gcs.current() < supersetVectorSize)
		return true;
	else
		return false;
}

template<class T> std::set<T*> SubsetIterator<T>::set() const
{
	return currentSubset;
}

template<class T> std::vector<T*>  SubsetIterator<T>::vector() const
{
	return std::vector<T*> (currentSubset.begin(), currentSubset.end());
}

}

#endif