#include "../point/matrix.h"

#include "../exception/invalidargumentexception.h"
#include "../exception/invalidruntimeconfigurationexception.h"

#include <math.h>
#include <iostream>
#include <numeric>
#include <vector>
#include <algorithm>

using namespace CluE;

Matrix::Matrix(size_t r, size_t c, std::vector<double> const & e):rows(r),cols(c)
{
	if (e.size() != r*c)
		throw InvalidArgumentException(0, "Incompatible dimensions!", "e");
		
	this->entries = e;
}

Matrix::Matrix(Point const & p, bool isRow /*= false*/)
{
	size_t d = p.dimension();

	if (isRow)
	{
		this->rows = 1;
		this->cols = d;
	} else {
		this->rows = d;
		this->cols = 1;
	}
	
	this->entries = std::vector<double>(d);
	for (size_t i=0; i<d; ++i)
		this->entries[i] = p[i];
}

Matrix Matrix::transpose() const
{
	Matrix m(this->cols, this->rows);
	for (size_t i=0; i<this->rows; ++i)
		for (size_t j=0; j<this->cols; ++j)
			m(j,i) = (*this)(i,j);
	return m;
}

Matrix Matrix::cholesky() const
{
	if (this->rows!=this->cols)
		throw InvalidRuntimeConfigurationException(0, "Matrix is not quadratic.");
		
	size_t n = this->rows;
	Matrix c(n,n);
	
	for (size_t i=0; i<n; ++i)
	{
		for (size_t j=0; j<i; ++j)
		{
			double sum = (*this)(i,j);
			for (size_t k=0; k<j; ++k)
				sum -= c(i,k)*c(j,k);
			c(i,j) = sum/c(j,j);
		}

		double sum = (*this)(i,i);
		for (size_t k=0; k<i; ++k)
			sum -= c(i,k)*c(i,k);
		if (sum > 0)
			c(i,i) = sqrt(sum);
		else
		{
			std::cout << "cholesky(): i = " << i << " sum = " << sum << std::endl;
			throw InvalidRuntimeConfigurationException(1, "Matrix is not positive definite.");
		}
	}
	
	return c;
}
	
bool Matrix::isSPD() const
{
	if (this->rows!=this->cols)
		return false;
		
	size_t n = this->rows;
	Matrix c(n,n);
	
	for (size_t i=0; i<n; ++i)
	{
		for (size_t j=0; j<i; ++j)
		{
			double sum = (*this)(i,j);
			for (size_t k=0; k<j; ++k)
				sum -= c(i,k)*c(j,k);
			c(i,j) = sum/c(j,j);
		}

		double sum = (*this)(i,i);
		for (size_t k=0; k<i; ++k)
			sum -= c(i,k)*c(i,k);
		if (sum > 0)
			c(i,i) = sqrt(sum);
		else
			return false;
	}
	
	return true;
}
	
double Matrix::spdDeterminant() const
{
	Matrix m = this->cholesky();
	
	size_t n = m.numRows();
	double det = 1;
	for (size_t i=0; i<n; ++i)
		det *= m(i,i)*m(i,i);
		
	return det;
}

Matrix Matrix::spdInverse() const
{
	Matrix ic = ltInverse(this->cholesky());
	
	// multiplication of inverse of cholesky decomposition with its transpose
	size_t n = this->rows;
	Matrix inv(n,n);
	for (size_t i=0; i<n; ++i)
		for (size_t j=0; j<n; ++j)
		{
			double sum = 0;
			for (size_t k=0; k<=i && k<=j; ++k)
			{
				sum += ic(i,k)*ic(j,k);
			}
			inv(i,j) = sum;
		}
		
	return inv;
}


Matrix Matrix::gramSchmidt() const
{
	size_t r = this->numRows();
	size_t c = this->numColumns();

	Matrix gs(*this);
	for (size_t i=0; i<c; ++i)
	{
		for (size_t j=0; j<i; ++j)
		{
			double s = 0;
			for (size_t k=0; k<r; ++k)
				s += gs(k,j)*(*this)(k,i);		
			for (size_t k=0; k<r; ++k)
				gs(k,i) -= s*gs(k,j);
		}
		double sum = 0;
		for (size_t k=0; k<r; ++k)
			sum += gs(k,i)*gs(k,i);
		for (size_t k=0; k<r; ++k)
			gs(k,i) /= sqrt(sum);
	}
	
	return gs;
}
Matrix& Matrix::operator+=(Matrix const & m)
{
	size_t r = this->numRows();
	size_t c = this->numColumns();

	if(m.numRows() != r || m.numColumns() != c)
		throw InvalidArgumentException(0, "Incompatible dimensions!", "m");

	for(size_t i = 0; i < r; ++i)
		for(size_t j = 0; j < c; ++j)
			(*this)(i,j) += m(i,j);

	return *this;
}

Matrix& Matrix::operator-=(Matrix const & m)
{
	size_t r = this->numRows();
	size_t c = this->numColumns();

	if(m.numRows() != r || m.numColumns() != c)
		throw InvalidArgumentException(0, "Incompatible dimensions!", "m");

	for(size_t i = 0; i < r; ++i)
		for(size_t j = 0; j < c; ++j)
			(*this)(i,j) -= m(i,j);

	return *this;
}

Matrix Matrix::operator+(Matrix const & m) const
{
	return Matrix(*this) += m;
}

Matrix Matrix::operator-(Matrix const & m) const
{
	return Matrix(*this) -= m;
}

Matrix& Matrix::operator*=(Matrix const & m)
{
	(*this)=(*this)*m;
	return *this;
}

Matrix Matrix::operator*(Matrix const & m) const
{
	size_t r = this->numRows();
	size_t n = this->numColumns();
	size_t c = m.numColumns();

	if(m.numRows() != n)
		throw InvalidArgumentException(0, "Incompatible dimensions!", "m");
	
	Matrix ret(r,c);
	for(size_t i = 0; i < r; ++i)
		for(size_t j = 0; j < c; ++j)
			for(size_t k = 0; k < n; ++k)
				ret(i,j) += (*this)(i,k)*m(k,j);
	
	return ret;
}

std::ostream& CluE::operator<<(std::ostream& os, Matrix const& m)
{
	size_t r = m.numRows();
	size_t c = m.numColumns();
	os << "(";
	for(size_t i=0; i<r; ++i)
	{
		os << "(";
		for(size_t j=0; j<c; ++j)
		{
			os << m(i,j);
			if(j<c-1)
				os << ", ";
		}
		os << ")";
		if(i<r-1)
			os << ", ";
	}
	os << ")";
	return os;
}

Matrix CluE::operator*(double scalar, Matrix const & m)
{	
	Matrix ret(m);

	size_t r = ret.numRows();
	size_t c = ret.numColumns();

	for(size_t i = 0; i < r; ++i)
		for(size_t j = 0; j < c; ++j)
			ret(i,j) = scalar * m(i,j);

	return ret;
}

Point CluE::operator*(Point const& row, Matrix const& m)
{
	size_t r = m.numRows();
	size_t c = m.numColumns();

	if(row.dimension() != r)
		throw InvalidArgumentException(0, "Incompatible dimensions!", "row, m");
	
	Point ret(c);
	for(size_t j = 0; j < c; ++j)
		for(size_t k = 0; k < r; ++k)
			ret[j] += row[k]*m(k,j);
	
	return ret;
}

Point CluE::operator*(Matrix const& m, Point const& column)
{
	size_t r = m.numRows();
	size_t c = m.numColumns();

	if(column.dimension() != c)
		throw InvalidArgumentException(0, "Incompatible dimensions!", "column, m");
	
	Point ret(r);
	for(size_t i = 0; i < r; ++i)
		for(size_t k = 0; k < c; ++k)
			ret[i] += m(i,k)*column[k];
	
	return ret;
}

Matrix Matrix::identity(double d)
{
	Matrix m(d, d);
	for (size_t i=0; i<d; ++i)
		m(i,i)=1;
		
	return m;
}

Matrix Matrix::ltInverse(Matrix const& lt)
{
	size_t n = lt.rows;
	Matrix inv(n,n);
	
	// compute inverse of lt (backwards substitution)
	for (size_t i=n; i-- > 0; )
	{
		inv(i,i) = 1/lt(i,i);
		for (size_t j=i+1; j<n; ++j)
		{
			double sum = 0;
			for (size_t k=i; k<j; ++k)
			{
				sum += lt(j,k)*inv(k,i);
			}
			inv(j,i) = -sum*inv(j,j);
		}
	}
	
	return inv;
}
