#ifndef MATRIX_H
#define MATRIX_H

#include "point.h"

#include <iostream>
#include <vector>

namespace CluE
{

/**
 * @brief Weighted matrix of arbitrary dimension.
 * 
 * @ingroup pointrelated_classes
 */
class Matrix
{
public:
	/**
	 * @brief Constructs a matrix.
	 */
	Matrix(size_t r = 0, size_t c = 0):rows(r),cols(c),entries(std::vector<double>(rows*cols, 0))
	{
	}
	
	/**
	 * @brief Constructs a matrix.
	 */
	Matrix(size_t r, size_t c, std::vector<double> const & e);

	/**
	 * @brief Constructs a matrix.
	 */
	Matrix(Point const & p, bool isRow = false);

	virtual ~Matrix()
	{
	}
	
	Matrix transpose() const;
	
	Matrix cholesky() const;
	
	bool isSPD() const;
	
	double spdDeterminant() const;
	
	Matrix spdInverse() const;
	
	Matrix gramSchmidt() const;

	/**
	 * @throw InvalidArgumentException [0] Incompatible dimensions!
	 */
	Matrix& operator+=(Matrix const & m);
	
	/**
	 * @throw InvalidArgumentException [0] Incompatible dimensions!
	 */
	Matrix& operator-=(Matrix const & m);
	
	/**
	 * @throw InvalidArgumentException [0] Incompatible dimensions!
	 */
	Matrix& operator*=(Matrix const & m);
	
	/**
	 * @throw InvalidArgumentException [0] Incompatible dimensions!
	 */
	Matrix operator+(Matrix const & m) const;
	
	/**
	 * @throw InvalidArgumentException [0] Incompatible dimensions!
	 */
	Matrix operator-(Matrix const & m) const;
	
	/**
	 * @throw InvalidArgumentException [0] Incompatible dimensions!
	 */
	Matrix operator*(Matrix const & m) const;	
	
	/**
	 * Accesses one particular matrix entry.
	 */
	double& operator()(size_t row, size_t col)
	{
		return this->entries[row*this->cols+col];
	}
	
	/**
	 * Returns one particular matrix entry.
	 */
	double operator()(size_t row, size_t col) const
	{
		return this->entries[row*this->cols+col];
	}
	
	/**
	 * Returns the number of rows of this matrix.
	 */
	size_t numRows() const
	{
		return this->rows;
	}
	
	/**
	 * Returns the number of columns of this matrix.
	 */
	size_t numColumns() const
	{
		return this->cols;
	}
	
	static Matrix identity(double dimension);
	
	static Matrix ltInverse(Matrix const& lt);
		
private:
	size_t rows, cols;
	std::vector<double> entries;
};

std::ostream& operator<<(std::ostream&, Matrix const&);

Matrix operator*(double scalar, Matrix const& m);

Point operator*(Point const& row, Matrix const& m);

Point operator*(Matrix const& m, Point const& column);

}

#endif