#include <math.h>

#include "../point/pointgauss.h"
#include "../point/matrix.h"

#include <algorithm>

using namespace CluE;

PointGauss::PointGauss(Point const& m, Matrix const& cov) : mean(m), covariance(cov)
{
	size_t d = m.dimension();
	this->cholesky = cov.cholesky();
	this->inverseCholesky = Matrix::ltInverse(this->cholesky);

	std::vector<double> temp(d);
	for (size_t i=0; i<d; ++i)
		temp[i] = log(this->cholesky(i,i));
	sort(temp.begin(), temp.end());
	this->logSqrt = accumulate(temp.begin(), temp.end(), d*log(2*M_PI)/2);
}

double PointGauss::density(Point const& x) const
{
	size_t d = x.dimension();
	
	// transform the given point
	Point y = this->inverseCholesky*(x-this->mean);
	std::vector<double> temp(d);
	for (size_t i=0; i<d; ++i)
		temp[i] = y[i]*y[i];
	sort(temp.begin(), temp.end());
	double qf = accumulate(temp.begin(), temp.end(), .0);

	return exp(-0.5*qf - logSqrt);	
}

double PointGauss::nll(Point const& x) const
{
	size_t d = x.dimension();
	
	// transform the given point
	Point y = this->inverseCholesky*(x-this->mean);
	std::vector<double> temp(d);
	for (size_t i=0; i<d; ++i)
		temp[i] = y[i]*y[i];
	sort(temp.begin(), temp.end());
	double qf = accumulate(temp.begin(), temp.end(), .0);

	return 0.5*qf + logSqrt;	
}

double PointGauss::squaredMahalanobis(Point const& x) const
{
	size_t d = x.dimension();
	
	// transform the given point
	Point y = this->inverseCholesky*(x-this->mean);
	double ret = 0;
	for (size_t i=0; i<d; ++i)
		ret += y[i]*y[i];
	return ret;	
}
