#ifndef ADAPTIVESAMPLING_H
#define ADAPTIVESAMPLING_H

#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../base/algorithm.h"
#include "../base/dissimilaritymeasure.h"
#include "../base/solutionprovider.h"
#include "../datastructure/discreteproxysolution.h"
#include "../misc/randomness.h"
#include "../point/pointweightmodifier.h"
#include "../exception/invalidruntimeconfigurationexception.h"

#include <set>
#include <ctime>
#include <vector>
#include <limits>

namespace CluE
{

/**
 * @brief k-means++ sampling method
 *
 * This is an Algorithm<T> template implementing the sampling method from
 * "k-means++: The advantages of careful seeding" by Arthur and Vassilvitskii
 *
 * @ingroup sampling_algorithms
 */
template<typename T> class AdaptiveSampling : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	AdaptiveSampling(const std::vector<T*>* data = NULL, DissimilarityMeasure<T>* measure = NULL,
	                 unsigned int numberOfSamples = 0);

	AdaptiveSampling(const AdaptiveSampling<T>&);
	AdaptiveSampling<T>& operator=(const AdaptiveSampling<T>&);
	virtual ~AdaptiveSampling();

	/**
	* @brief Computes the sample set
	* @return Sample set of size number_of_samples if number_of_samples is at most the size of input
	* @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	* @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	* @throw InvalidRuntimeConfigurationException [2] Empty input set.
	* @throw InvalidRuntimeConfigurationException [3] Desired number of samples is 0.
	* @throw InvalidRuntimeConfigurationException [4] Desired number of samples is larger than size of input.
	*/
	virtual DiscreteProxySolution<T>* compute();

	virtual void setInput(std::vector<T*> const*);
	virtual void setMeasure(DissimilarityMeasure<T> const*);
	void setNumberOfSamples(unsigned int);
	void setWeightModifier(WeightModifier<T>* wm);

	/**
	* @brief Does a dynamic cast of the given Algorithm to AdaptiveSampling<T>
	* @return NULL if the Algorithm is not a AdaptiveSampling<T> instance
	*/
	static AdaptiveSampling<T>* toAdaptiveSampling(Algorithm* s);

private:
	std::vector<T*> const* input;
	DissimilarityMeasure<T>* measure;
	unsigned int number_of_samples;
	WeightModifier<T>* weightModifier;
};

template<typename T> AdaptiveSampling<T>::AdaptiveSampling(std::vector<T*> const* data,
        DissimilarityMeasure<T>* m, unsigned int n) : input(data), measure(m==NULL?NULL:m->clone()),
	number_of_samples(n),
	weightModifier(0)
{
}

template<typename T> AdaptiveSampling<T>::AdaptiveSampling(const AdaptiveSampling<T>& rhs) :
	Algorithm(rhs), input(rhs.input), measure(rhs.measure==NULL?NULL:rhs.measure->clone()),
	number_of_samples(rhs.number_of_samples),
	weightModifier(0)
{
}

template<typename T> AdaptiveSampling<T>& AdaptiveSampling<T>::operator=(
    const AdaptiveSampling<T>& rhs)
{
	Algorithm::operator=(rhs);
	DissimilarityMeasure<T>* dm = rhs.measure==NULL?NULL:rhs.measure->clone();
	delete this->measure;
	this->measure = dm;
	this->input = rhs.input;
	this->number_of_samples = rhs.number_of_samples;
	this->weightModifier = rhs.weightModifier == 0 ? 0 : rhs.weightModifier->clone();
	return *this;
}

template<typename T> AdaptiveSampling<T>::~AdaptiveSampling()
{
	delete this->measure;
}

template<typename T> DiscreteProxySolution<T>* AdaptiveSampling<T>::compute()
{
	if(this->input==NULL)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if(this->measure==NULL)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");

	time_t start, end;
	start = time(0);

	DiscreteProxySolution<T>* solution = new DiscreteProxySolution<T>();

	unsigned int N = this->input->size();
	if(N==0)
		throw InvalidRuntimeConfigurationException(2, "Empty input set.");
	unsigned int samplenum = this->number_of_samples;
	if(samplenum==0)
		throw InvalidRuntimeConfigurationException(3, "Desired number of samples is 0.");
	if(samplenum>N)
		throw InvalidRuntimeConfigurationException(4, "Desired number of samples is larger than size of input.");

	std::vector<T*> unchosen = *this->input; // copy the input vector
	solution->proxysets.push_back(std::vector<T*>());

	RandomGenerator rg = Randomness::getRandomGenerator();
	std::uniform_int_distribution<int> dis(0, N-1);
	int index = dis(rg);

	solution->proxysets[0].push_back(unchosen[index]); // choose first center at random from all elements
	unchosen.erase(unchosen.begin()+index); // remove the first center from the first dataset

	std::vector<double> weights(N-1, std::numeric_limits<double>::infinity());
	for(unsigned int i=1; i<samplenum; i++)
	{
		// compute weights to be the distance to the nearest center
		double total = .0;
		for(unsigned int j=0; j<N-i; j++)
		{
			double w = this->measure->dissimilarity(*unchosen[j], *solution->proxysets[0][i-1]);
			if(weightModifier != 0)
				w *= weightModifier->getWeight(*unchosen[j]);
			if(w<weights[j])
				weights[j] = w;
			total += weights[j];
		}

		RandomGenerator rg = Randomness::getRandomGenerator();
		std::uniform_real_distribution<> dis(0, total);
		double pos = dis(rg);

		index = -1;
		for(unsigned int j=0; j<N-i&&index<0; j++)
			if(pos<weights[j])
				index = j;
			else
				pos -= weights[j];
		if(index<0)
			index = 0;

		solution->proxysets[0].push_back(unchosen[index]);
		unchosen.erase(unchosen.begin()+index);
	}

	end = time(0);
	solution->seconds=end-start;
	//std::clog << "CluE::AdaptiveSampling<T>::compute() - finished" << std::endl;
	return solution;
}

template<typename T> void AdaptiveSampling<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void AdaptiveSampling<T>::setMeasure(DissimilarityMeasure<T> const* m)
{
	this->measure = m==NULL?NULL:m->clone();
}

template<typename T> void AdaptiveSampling<T>::setNumberOfSamples(unsigned int n)
{
	this->number_of_samples = n;
}

template<typename T> void AdaptiveSampling<T>::setWeightModifier(WeightModifier<T>* wm)
{
	if(wm != 0)
		weightModifier = wm->clone();
	else
		wm = 0;
}

template<typename T> AdaptiveSampling<T>* AdaptiveSampling<T>::toAdaptiveSampling(Algorithm* s)
{
	return dynamic_cast<AdaptiveSampling<T>*>(s);
}

}

#endif
