#ifndef FARTHESFIRSTSAMPLING_H
#define FARTHESFIRSTSAMPLING_H

#include <ctime>

#include "../base/inputsetter.h"
#include "../base/measuresetter.h"
#include "../base/algorithm.h"
#include "../datastructure/discreteproxysolution.h"
#include "../base/dissimilaritymeasure.h"
#include "../misc/dynamicbitset.h"

#include "../exception/invalidruntimeconfigurationexception.h"

namespace CluE
{

/**
 * @brief Farthest first traversal algorithm.
 *
 * 1. Pick a point to start with (first point) and label it "1". \n
 * 2. Choose the point furthest from it and label it "2". \n
 * 3. Choose the point furthest from the two and label it "3".  \n
 * (n+1). Choose the point furthest from the n and label it "n+1".  \n
 *
 * @sa Gonzales is a clustering algorithm using n farthest-first-labeled points as proxies.
 *
 * @ingroup sampling_algorithms
 */
template<typename T> class FarthestFirstTraversal : public Algorithm, public InputSetter<T>, public MeasureSetter<T>
{
public:
	/**
	* @brief Constructor for general usage.
	* @sa setMeasure
	* @sa setInput
	* @sa setNumberOfSamples
	* @sa setFirstPoint
	*/
	FarthestFirstTraversal(DissimilarityMeasure<T> const * measure = 0, std::vector<T*> const* input = 0, unsigned int numOfSamples = 0, unsigned int firstPoint = 0);

	FarthestFirstTraversal(const FarthestFirstTraversal<T>&);
	FarthestFirstTraversal<T>& operator= (const FarthestFirstTraversal<T>&);
	virtual ~FarthestFirstTraversal();

	/**
	 * @brief Returns an ordered DiscretesProxySolution. The first point is the starting point.
	 * @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	 * @throw InvalidRuntimeConfigurationException [1] Dissimilarity measure is NULL.
	 * @throw InvalidRuntimeConfigurationException [2] Empty input set.
	 * @throw InvalidRuntimeConfigurationException [3] Desired number of samples is 0.
	 * @throw InvalidRuntimeConfigurationException [4] Desired number of samples is larger than size of input.
	 */
	virtual SolutionProvider* compute();

	virtual void setInput(std::vector<T*> const*);

	virtual void setMeasure(DissimilarityMeasure<T> const * measure);

	/**
	 * @brief Sets the desired number of proxies.
	 */
	void setNumberOfSamples(unsigned int number);

	/**
	 * @brief Sets the starting point.
	 */
	void setFirstPoint(unsigned int index);

	/**
	* @brief does a dynamic cast of the given Algorithm to FarthestFirstTraversal<T>
	* @return NULL if the Algorithm is not a FarthestFirstTraversal<T> instance
	*/
	static FarthestFirstTraversal<T>* toFarthestFirstTraversal(Algorithm* s);

private:
	std::vector<T*> const* input;
	DissimilarityMeasure<T>* measure;
	unsigned int number_of_samples;
	unsigned int first_point;
};

template<typename T> FarthestFirstTraversal<T>::FarthestFirstTraversal(DissimilarityMeasure<T> const *measure, std::vector<T*> const* data, unsigned int n, unsigned int s) :
	input(data),
	measure(measure==0 ? 0 : measure->clone()),
	number_of_samples(n),
	first_point(s)
{
}

template<typename T> FarthestFirstTraversal<T>::FarthestFirstTraversal(const FarthestFirstTraversal<T>& fft) :
	input(fft.input),
	measure(fft.measure == 0 ? 0 : fft.measure->clone()),
	number_of_samples(fft.number_of_samples),
	first_point(fft.number_of_samples)
{
}

template<typename T> FarthestFirstTraversal<T>& FarthestFirstTraversal<T>::operator= (const FarthestFirstTraversal<T>& fft)
{
	if(measure != 0)
		delete measure;

	Algorithm::operator= (fft);

	input = fft.input;
	measure == fft.measure == 0 ? 0 : fft.measure->clone();
	number_of_samples = fft.number_of_samples;
	first_point = fft.number_of_samples;

	return *this;
}

template<typename T> FarthestFirstTraversal<T>::~FarthestFirstTraversal()
{
	if(measure != 0)
		delete measure;
}

template<typename T> SolutionProvider* FarthestFirstTraversal<T>::compute()
{
	DiscreteProxySolution<T>* solution = new DiscreteProxySolution<T>();
	unsigned int inputSize = this->input->size();

	if(this->input == 0)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	if(this->measure == 0)
		throw InvalidRuntimeConfigurationException(1, "Dissimilarity measure is NULL.");
	if(inputSize == 0)
		throw InvalidRuntimeConfigurationException(2, "Empty input set.");
	if(this->number_of_samples == 0)
		throw InvalidRuntimeConfigurationException(3, "Desired number of samples is 0.");
	if(this->number_of_samples > inputSize)
		throw InvalidRuntimeConfigurationException(4, "Desired number of samples is larger than input size.");

	time_t start, end;
	start = std::time(0);

	std::vector<T*> numberedPoints;
	unsigned int max = this->number_of_samples > 0 ? this->number_of_samples : inputSize;
	DynamicBitset<> isNumbered(inputSize);

	//First point
	T* firstPoint = (*this->input)[this->first_point];
	isNumbered.set(this->first_point, true);
	numberedPoints.push_back(firstPoint);

	//Min-distance caching
	std::vector<double> distances;
	for(unsigned int i = 0; i < inputSize; i++)
	{
		distances.push_back(this->measure->dissimilarity(*firstPoint, *(*this->input)[i]));
	}

	for(unsigned int i = 1; i < max; i++)
	{
		bool next_found = false;
		unsigned int next_number = 0;
		double next_distance = 0;

		for(unsigned int j = 0; j < inputSize; j++)
		{
			if(isNumbered[j])
				continue;

			//min(cached distance of candidate, distance between last numbered point and candidate)
			double testDist = measure->dissimilarity(*numberedPoints[i-1], * (*this->input)[j]);
			if(testDist < distances[j])
				distances[j] = testDist;

			if(distances[j] > next_distance || !next_found)
			{
				next_found = true;
				next_number = j;
				next_distance = distances[j];
			}
		}

		isNumbered.set(next_number,  true);
		numberedPoints.push_back((*this->input)[next_number]);
	}

	end = time(0);
	solution->proxysets.push_back(numberedPoints);
	solution->seconds = end-start;

	return solution;
}

template<typename T> void FarthestFirstTraversal<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void FarthestFirstTraversal<T>::setFirstPoint(unsigned int s)
{
	this->first_point = s;
}

template<typename T> void FarthestFirstTraversal<T>::setNumberOfSamples(unsigned int n)
{
	this->number_of_samples = n;
}

template<typename T> FarthestFirstTraversal<T>* FarthestFirstTraversal<T>::toFarthestFirstTraversal(Algorithm* s)
{
	return dynamic_cast<FarthestFirstTraversal<T>*>(s);
}

template<typename T> void FarthestFirstTraversal<T>::setMeasure(DissimilarityMeasure<T> const *measure)
{
	if(this->measure != 0)
		delete this->measure;
	
	if(measure == 0)
		this->measure = 0;
	else
		this->measure = measure->clone();
}

}
#endif
