#ifndef UNIFORMSAMPLING_H
#define UNIFORMSAMPLING_H

#include "../base/inputsetter.h"
#include "../base/algorithm.h"
#include "../base/solutionprovider.h"
#include "../datastructure/discreteproxysolution.h"
#include "../misc/randomness.h"
#include "../exception/invalidruntimeconfigurationexception.h"

#include <set>
#include <ctime>

namespace CluE
{

/**
 * @brief Uniform sampling
 * @ingroup sampling_algorithms
 */
template<typename T> class UniformSampling : public Algorithm, public InputSetter<T>
{
public:
	UniformSampling(std::vector<T*> const* data = NULL, unsigned int numberOfSamples = 0);

	virtual ~UniformSampling()
	{
	}

	/**
	* @brief Computes the sample set
	* @return Sample set of size number_of_samples if number_of_samples is at most the size of input
	* @throw InvalidRuntimeConfigurationException [0] Input is NULL.
	* @throw InvalidRuntimeConfigurationException [2] Empty input set.
	* @throw InvalidRuntimeConfigurationException [3] Desired number of samples is 0.
	* @throw InvalidRuntimeConfigurationException [4] Desired number of samples is larger than size of input.
	*/
	virtual DiscreteProxySolution<T>* compute();

	virtual void setInput(std::vector<T*> const*);
	void setNumberOfSamples(unsigned int);

	/**
	* @brief does a dynamic cast of the given Algorithm to UniformSampling<T>
	* @return NULL if the Algorithm is not a UniformSampling<T> instance
	*/
	static UniformSampling<T>* toUniformSampling(Algorithm* s);

private:
	std::vector<T*> const* input;
	unsigned int number_of_samples;
};

template<typename T> UniformSampling<T>::UniformSampling(std::vector<T*> const* data,
        unsigned int n) : input(data), number_of_samples(n)
{
}

template<typename T> DiscreteProxySolution<T>* UniformSampling<T>::compute()
{
	if(this->input==NULL)
		throw InvalidRuntimeConfigurationException(0, "Input is NULL.");
	
	time_t start, end;
	start = time(0);

	DiscreteProxySolution<T>* solution = new DiscreteProxySolution<T>();

	unsigned int N = this->input->size();
	if(N==0)
		throw InvalidRuntimeConfigurationException(2, "Empty input set.");
	unsigned int samplenum = this->number_of_samples;
	if(samplenum==0)
		throw InvalidRuntimeConfigurationException(3, "Desired number of samples is 0.");
	if(samplenum>N)
		throw InvalidRuntimeConfigurationException(4, "Desired number of samples is larger than size of input.");

	std::vector<T*> unchosen = *this->input; // copy the input vector
	solution->proxysets.push_back(std::vector<T*>());

	for(unsigned int i=0; i<samplenum; i++)
	{
		RandomGenerator rg = Randomness::getRandomGenerator();
		std::uniform_int_distribution<int> dis(0, N-i-1);
		int index = dis(rg);

		solution->proxysets[0].push_back(unchosen[index]); // choose at random from remaining elements
		unchosen.erase(unchosen.begin()+index); // remove chosen sample from dataset
	}

	end = time(0);
	solution->seconds=end-start;
	std::clog << "CluE::UniformSampling<T>::compute() - finished" << std::endl;
	return solution;
}

template<typename T> void UniformSampling<T>::setInput(std::vector<T*> const* data)
{
	this->input = data;
}

template<typename T> void UniformSampling<T>::setNumberOfSamples(unsigned int n)
{
	this->number_of_samples = n;
}

template<typename T> UniformSampling<T>* UniformSampling<T>::toUniformSampling(Algorithm* s)
{
	return dynamic_cast<UniformSampling<T>*>(s);
}

}

#endif
