#include "../base/partitionevaluation.h"
#include "../base/combinedevaluation.h"
#include "../base/proxygenerator.h"

namespace CluE
{

/**
 * @brief Wraps CombinedEvaluation into PartitionEvaluation by using a ProxGenerator instance.
 *
 * Provide a CombinedEvaluation and a suitable ProxyGenerator to use the CombinedEvaluation as a PartitionEvaluation.
 *
 * @remarks As this class is a wrapper class, it does not create any copies of the provided objects / pointers (e.g. CombinedEvaluation)
 *
 * @ingroup wrapper_classes
 */
template<class T> class CombinedToPartitionEvaluationWrapper : public PartitionEvaluation<T>
{
public:
	/**
	 * @brief Provide a CombinedEvaluation and a suitable ProxyGenerator.
	 */
	CombinedToPartitionEvaluationWrapper(CombinedEvaluation<T> const *evaluation, ProxyGenerator<T> const *generator);

	/**
	 * @brief See wrapped class's combinedcost() documentation.
	 */
	virtual double partitioncost(std::vector<std::vector<T*> > const &partitioning) const;
	/**
	 * @overload
	 */
	virtual double partitioncost(PartitionProvider<T> const &partitionProvider, unsigned int solutionIndex) const;

	/**
	 * @overload
	 */
	virtual double partitioncost(std::vector<T*> const &partition) const;
	/**
	 * @overload
	 */
	virtual double partitioncost(PartitionProvider<T> const &partitionProvider, unsigned int solutionIndex, unsigned int partitionIndex) const;

private:
	CombinedEvaluation<T> const * const evaluation;
	ProxyGenerator<T> const * const generator;
};

template<class T> CombinedToPartitionEvaluationWrapper<T>::CombinedToPartitionEvaluationWrapper(CombinedEvaluation<T> const *e, ProxyGenerator<T> const *g) : evaluation(e), generator(g)
{
	//TODO NULL-Exception, ...
}

template<class T> double CombinedToPartitionEvaluationWrapper<T>::partitioncost(std::vector<std::vector<T*> > const &partitioning) const
{
	std::vector<T> proxies;
	size_t n = partitioning.size();
	for(size_t i = 0; i < n; i++)
	{
		proxies.push_back(generator->generate(partitioning[i]));
	}
	return evaluation->combinedcost(partitioning, proxies);
}

template<class T> double CombinedToPartitionEvaluationWrapper<T>::partitioncost(PartitionProvider<T> const &partitionProvider, unsigned int solutionIndex) const
{
	std::vector<std::vector<T*> > partitioning = partitionProvider.clustering(solutionIndex);

	std::vector<T> proxies;
	size_t n = partitioning.size();
	for(size_t i = 0; i < n; i++)
	{
		proxies.push_back(generator->generate(partitioning[i]));
	}
	return evaluation->combinedcost(partitioning, proxies);
}

template<class T> double CombinedToPartitionEvaluationWrapper<T>::partitioncost(std::vector<T*> const &partition) const
{
	return evaluation->combinedcost(partition, generator->generate(partition));
}

template<class T> double CombinedToPartitionEvaluationWrapper<T>::partitioncost(PartitionProvider<T> const &partitionProvider, unsigned int solutionIndex, unsigned int partitionIndex) const
{
	std::vector<T*> partition = partitionProvider.cluster(solutionIndex, partitionIndex);
	return evaluation->combinedcost(partition, generator->generate(partition));
}
}
