#include <random>

#include "demoprocessor.h"

using std::cout;
using std::endl;
using namespace CluE;

namespace CluE
{

DemoProcessor::DemoProcessor(unsigned int dimension, unsigned int size, unsigned int clu, bool useGauss) : clusters(clu)
{
	cout << "Generating random points ... ";
	cout.flush();
	if(useGauss)
		genGaussInput(dimension, size);
	else
		genUniformInput(dimension, size);
	cout << "finished." << endl << endl;
}

DemoProcessor::DemoProcessor(std::vector<Point*> inp, unsigned int clu) : input(inp), clusters(clu)
{
}

void DemoProcessor::agglomerativeSingleLinkage()
{
	MaxDiamEvaluator<Point> mde(measure);

	SingleLinkage<Point> *link = new SingleLinkage<Point>(*measure);

	cout << "Agglomerative (single linkage) ... ";
	cout.flush();

	Agglomerative<Point> agg(&input, link);
	PartitionSolution<Point> *ps;
	ps = dynamic_cast<PartitionSolution<Point>*>(agg.compute());

	cout << "finished." << endl;

	double cost = mde.partitioncost(ps->partitions[ps->partitions.size()-1]);

	cout << "Calculation time: " << (ps->seconds) << endl;
	cout << "Maximum diameter cost: " << cost << endl;
	cout << endl;

	delete ps;
	delete link;
}

void DemoProcessor::agglomerativeAverageLinkage()
{
	MaxDiamEvaluator<Point> mde(measure);

	AverageLinkage<Point> *link = new AverageLinkage<Point>(*measure);

	cout << "Agglomerative (average linkage) ... ";
	cout.flush();

	Agglomerative<Point> agg(&input, link);
	PartitionSolution<Point> *ps;
	ps = dynamic_cast<PartitionSolution<Point>*>(agg.compute());

	cout << "finished." << endl;

	double cost = mde.partitioncost(ps->partitions[ps->partitions.size()-1]);

	cout << "Calculation time: " << (ps->seconds) << endl;
	cout << "Maximum diameter cost: " << cost << endl;
	cout << endl;

	delete ps;
	delete link;
}

void DemoProcessor::agglomerativeCompleteLinkage()
{
	MaxDiamEvaluator<Point> mde(measure);

	CompleteLinkage<Point> *link = new CompleteLinkage<Point>(*measure);

	cout << "Agglomerative (complete linkage) ... ";
	cout.flush();

	Agglomerative<Point> agg(&input, link);
	PartitionSolution<Point> *ps;
	ps = dynamic_cast<PartitionSolution<Point>*>(agg.compute());

	cout << "finished." << endl;

	double cost = mde.partitioncost(ps->partitions[ps->partitions.size()-1]);

	cout << "Calculation time: " << (ps->seconds) << endl;
	cout << "Maximum diameter cost: " << cost << endl;
	cout << endl;

	delete ps;
	delete link;
}

void DemoProcessor::lloydtypeKMeans(unsigned int iterations)
{
	PointCentroid *pc = new PointCentroid();
	KMeansEvaluator<Point> kme(measure);

	cout << "K-means ... ";
	cout.flush();

	UniformSampling<Point> us(&input, clusters);
	DiscreteProxySolution<Point> *dps;
	dps = dynamic_cast<DiscreteProxySolution<Point>*>(us.compute());

	Lloydtype<Point> lloyd(dps->proxysets[0], &input, pc, measure, iterations);
	DoubleSolution<Point> *ps;
	ps = dynamic_cast<DoubleSolution<Point>*>(lloyd.compute());

	cout << "finished." << endl;

	double cost = kme.proxycost(input, dps->proxysets[0]);

	cout << "Calculation time: " << (dps->seconds + ps->seconds) << endl;
	cout << "K-means cost: " << cost << endl;
	cout << endl;

	delete ps;

	delete pc;
}

void DemoProcessor::lloydtypeKMeansPP(unsigned int iterations)
{
	PointCentroid *pc = new PointCentroid();
	KMeansEvaluator<Point> kme(measure);

	cout << "K-means++ ... ";
	cout.flush();

	AdaptiveSampling<Point> as(&input, measure, clusters);
	DiscreteProxySolution<Point> *dps;
	dps = dynamic_cast<DiscreteProxySolution<Point>*>(as.compute());

	Lloydtype<Point> lloyd(dps->proxysets[0], &input, pc, measure, iterations);
	DoubleSolution<Point> *ps;
	ps = dynamic_cast<DoubleSolution<Point>*>(lloyd.compute());

	cout << "finished." << endl;

	double cost = kme.proxycost(input, dps->proxysets[0]);

	cout << "Calculation time: " << (dps->seconds + ps->seconds) << endl;
	cout << "K-means cost: " << cost << endl;
	cout << endl;

	delete ps;

	delete pc;
}

void DemoProcessor::gonzales()
{
	KMeansEvaluator<Point> kme(measure);
	
	cout << "Gonzales ... ";
	cout.flush();

	Gonzales<Point> gon(measure, &input, clusters);
	DiscreteDoubleSolution<Point> *ps;
	ps = dynamic_cast<DiscreteDoubleSolution<Point>*>(gon.compute());

	cout << "finished." << endl;

	double cost = kme.combinedcost(*ps, *ps, 0);
	
	cout << "Calculation time: " << ps->seconds << endl;
	cout << "K-means cost: " << cost << endl;
	cout << endl;
	
	delete ps;
}

void DemoProcessor::genGaussInput(unsigned int dimension, unsigned int size)
{
	RandomGenerator gen = Randomness::getRandomGenerator();
	std::normal_distribution<> dis(0.0, 1.0);
	cout << "(mean 0.0, sigma 1.0) ";

	std::vector<CluE::Point*> points;
	for (size_t i=0; i<size; i++)
	{
		std::vector<double> coords;
		for (size_t j=0; j<dimension; j++)
			coords.push_back(dis(gen));
		CluE::Point* p = new CluE::Point(coords, 1);
		input.push_back(p);
	}
}

void DemoProcessor::genUniformInput(unsigned int dimension, unsigned int size)
{
	RandomGenerator gen = Randomness::getRandomGenerator();
	std::uniform_real_distribution<> dis(0.0, 3.0);
	cout << "(range 0.0-3.0) ";

	std::vector<CluE::Point*> points;
	for (size_t i=0; i<size; i++)
	{
		std::vector<double> coords;
		for (size_t j=0; j<dimension; j++)
			coords.push_back(dis(gen));
		CluE::Point* p = new CluE::Point(coords, 1);
		input.push_back(p);
	}
}

void DemoProcessor::setMeasure(DissimilarityMeasure<Point> *m)
{
	measure = m;
}

}
