#include "helper.h"
#include "algorithmtest.h"
#include "exceptions.h"

#include <cstdlib>
#include <stdexcept>
#include <sstream>

#include <point/point.h>
#include <base/solutionprovider.h>

#include <clustering/gonzales.h>
#include <sampling/farthestfirsttraversal.h>
#include <clustering/dasguptalong.h>
#include <clustering/kumarsabharwalsen.h>
#include <clustering/balcanblumgupta.h>

#include <point/l1metric.h>
#include <point/squaredl1metric.h>
#include <point/l2metric.h>
#include <point/squaredl2metric.h>
#include <point/lpmetric.h>
#include <point/squaredlpmetric.h>

#include <exception/clueexception.h>

using std::string;
using std::endl;
using std::vector;

namespace CluE
{

AlgorithmTest::AlgorithmTest(std::string path, std::ostream& stream) : configuration(FileReader(path)), output(stream), testDiscreteProxySequence(false), testPartitioning(false), testEpsilonApproximation(false)
{
	if(configuration.hasValue("main.input"))
	{
		string inputpath = configuration.getConfigString("main.input", false);
		input =  FileReader::readPointFile(inputpath);
	}
	else
	{
		InvalidConfiguration ic;
		ic.message = "Main.Input: No input file specified.";
		throw ic;
	}

	size_t n = input.size();
	for(size_t i = 0; i < n; i++)
	{
		lookupIndex[input[i]] = i;
	}

	if(configuration.hasValue("main.algorithm"))
	{
		string alg = configuration.getConfigString("main.algorithm");
		if(alg.compare("farthestfirsttraversal") == 0)
			algorithm = AlgorithmType_FarthestFirstTraversal;
		else if(alg.compare("gonzales") == 0)
			algorithm = AlgorithmType_Gonzales;
		else if(alg.compare("dasguptalong") == 0)
			algorithm = AlgorithmType_DasguptaLong;
		else if(alg.compare("kumarsabharwalsen") == 0)
			algorithm = AlgorithmType_KumarSabharwalSen;
		else if(alg.compare("balcanblumgupta") == 0)
			algorithm = AlgorithmType_BalcanBlumGupta;
		else
		{
			InvalidConfiguration ic;
			ic.message = "Main.Algorithm: Invalid or unsupported algorithm specified.";
			throw ic;
		}
	}

	if(configuration.hasValue("main.measure"))
	{
		string m = configuration.getConfigString("main.measure");
		if(m.compare("l1") == 0)
			measure = new L1Metric();
		else if(m.compare("sl1") == 0)
			measure = new SquaredL1Metric();
		else if(m.compare("l2") == 0)
			measure = new L2Metric();
		else if(m.compare("sl2") == 0)
			measure = new SquaredL2Metric();
		else
		{
			InvalidConfiguration ic;
			ic.message = "Main.Measure: Invalid or unsupported measure specified.";
			throw ic;
		}
	}

	if(configuration.hasValue("main.solution"))
	{
		solution = configuration.getConfigUInt("main.solution");
	}
	else
	{
		solution = 0;
	}

	centroidGen = new PointCentroid();

	autoSetup();
}

AlgorithmTest::~AlgorithmTest()
{
	delete measure;
	delete centroidGen;

	size_t inputSize = input.size();
	for(size_t i = 0; i < inputSize; i++)
		delete input[i];
}

bool AlgorithmTest::run()
{
	bool success = true;

	runAlgorithm(true);

	if(configuration.hasValue("main.description"))
		output << " (" << configuration.getConfigString("main.description") << ")" << endl << endl;
	else
		output << endl << endl;

	if(solution >= result->number_of_solutions())
	{
		output << "ERROR: Requested solution is not available.";
		success = false;
	}
	
	if(testDiscreteProxySequence)
	{
		output << "--- DiscreteProxySequence test ---" << endl;
		bool suc = false;
		try
		{
			suc = doDiscreteProxySequenceTest();
		}
		catch(TestMessageException e)
		{
			std::cerr << "DiscreteProxySequence test failed: " << e.message << std::endl;
		}
		catch(CluEException e)
		{
			std::cerr << "DiscreteProxySequence test failed due to a CluE exception: " << e.message << std::endl;
		}
		catch(...)
		{
			std::cerr << "DiscreteProxySequence test failed due to an unexpected exception." << std::endl;
		}
		printSuccess(suc);
		success &= suc;
	}

	if(testPartitioning)
	{
		output << "--- Partitioning test ---" << endl;
		bool suc = false;
		try
		{
			suc = doPartitioningTest();
		}
		catch(TestMessageException e)
		{
			std::cerr << "Partitioning test failed: " << e.message << std::endl;
		}
		catch(CluEException e)
		{
			std::cerr << "Partitioning test failed due to a CluE exception: " << e.message << std::endl;
		}
		catch(...)
		{
			std::cerr << "Partitioning test failed due to an unexpected exception." << std::endl;
		}
		printSuccess(suc);
		success &= suc;
	}

	if(testEpsilonApproximation)
	{
		output << "--- Epsilon approximation test ---" << endl;
		bool suc = false;
		try
		{
			suc = doEpsilonApproximationTest();
		}
		catch(TestMessageException e)
		{
			std::cerr << "Epsilon approximation test: " << e.message << std::endl;
		}
		catch(CluEException e)
		{
			std::cerr << "Epsilon approximation test failed due to a CluE exception: " << e.message << std::endl;
		}
		catch(...)
		{
			std::cerr << "Epsilon approximation test failed due to an unexpected exception." << std::endl;
		}
		printSuccess(suc);
		success &= suc;
	}

	return success;
}

void AlgorithmTest::runAlgorithm(bool printOutput)
{
	if(printOutput)
		output << "Algorithm: ";
	switch(algorithm)
	{
		case AlgorithmType_FarthestFirstTraversal:
		{
			unsigned int start = configuration.getConfigUInt("farthestfirsttraversal.start");
			FarthestFirstTraversal<Point> ftt(measure, &input, input.size(), start);
			result = ftt.compute();
			if(printOutput)
				output << "FarthestFirstTraversal";
			break;
		}
		case AlgorithmType_Gonzales:
		{
			unsigned int start = configuration.getConfigUInt("gonzales.start");
			unsigned int number = configuration.getConfigUInt("gonzales.partitions");
			Gonzales<Point> gon(measure, &input, number, start);
			result = gon.compute();
			if(printOutput)
				output << "Gonzales";
			break;
		}
		case AlgorithmType_DasguptaLong:
		{
			unsigned int start = configuration.getConfigUInt("dasguptalong.start");
			double alpha = configuration.getConfigDouble("dasguptalong.alpha");
			double beta = configuration.getConfigDouble("dasguptalong.beta");
			DasguptaLong<Point> dl(measure, &input, start);
			dl.setGranularity(alpha, beta);
			result = dl.compute();
			if(printOutput)
				output << "DasguptaLong";
			break;
		}
		case AlgorithmType_KumarSabharwalSen:
		{
			unsigned int numOfClusters = configuration.getConfigUInt("kumarsabharwalsen.number");
			double supersetSize = configuration.getConfigDouble("kumarsabharwalsen.supersetsize");
			double subsetSize = configuration.getConfigDouble("kumarsabharwalsen.subsetsize");
			KumarSabharwalSen<Point> kss(measure, centroidGen, numOfClusters, &input);
			kss.setSetSizes(supersetSize, subsetSize);
			result = kss.compute();
			lastCost = kss.getLastCost();
			if(printOutput)
				output << "KumarSabharwalSen";
			break;
		}
		case AlgorithmType_BalcanBlumGupta:
		{
			unsigned int numOfClusters = configuration.getConfigUInt("balcanblumgupta.number");
			double alpha = configuration.getConfigDouble("balcanblumgupta.alpha");
			double beta = configuration.getConfigDouble("balcanblumgupta.beta");
			double epsilon = configuration.getConfigDouble("balcanblumgupta.epsilon");
			double omega = configuration.getConfigDouble("balcanblumgupta.omega");
			BalcanBlumGupta<Point> bbg(measure, &input, numOfClusters);
			bbg.setAlpha(alpha);
			bbg.setBeta(beta);
			bbg.setEpsilon(epsilon);
			bbg.setOmega(omega);
			result = bbg.compute();
			//LastCost
			if(printOutput)
				output << "BalcanBlumGupta";
		}
	}
}

void AlgorithmTest::autoSetup()
{
	if(configuration.getConfigFlag("main.testdiscreteproxysequence"))
	{
		string val = configuration.getConfigString("discreteproxysequence.sequence");
		vector<unsigned int> uintseq = Helper::convertToUIntSequence(val);

		setupDiscreteProxySequenceTest(uintseq);
		testDiscreteProxySequence = true;
	}

	if(configuration.getConfigFlag("main.testpartitioning"))
	{
		unsigned int num = configuration.getConfigUInt("partitioning.number");
		vector<vector<unsigned int> > partitions;
		for(unsigned int i = 0; i < num; i++)
		{
			std::stringstream key;
			key << "partitioning.partition" << i;
			if(configuration.hasValue(key.str()))
			{
				string val = configuration.getConfigString(key.str());
				vector<unsigned int> partition = Helper::convertToUIntSequence(val);
				partitions.push_back(partition);
			}
			else
			{
				InvalidConfiguration ic;
				ic.message = key.str() + ": missing.";
				throw ic;
			}
		}
		setupPartitioningTest(partitions);
		testPartitioning = true;
	}

	if(configuration.getConfigFlag("main.testepsilonapproximation"))
	{
		unsigned int laps = configuration.getConfigUInt("epsilonapproximation.laps");
		double optimalCost = configuration.getConfigDouble("epsilonapproximation.optimalcost");
		double propability = configuration.getConfigDouble("epsilonapproximation.propability");
		double epsilon = configuration.getConfigDouble("epsilonapproximation.epsilon");
		
		setupEpsilonApproximationTest(laps, optimalCost, propability, epsilon);
		testEpsilonApproximation = true;
	}
}

}
