#include "helper.h"
#include "evaluationtest.h"
#include "exceptions.h"

#include <cstdlib>
#include <stdexcept>
#include <sstream>
#include <unordered_map>

#include <point/point.h>

#include <evaluation/kmeansevaluator.h>
#include <evaluation/kcenterevaluator.h>
#include <evaluation/maxdiamevaluator.h>

#include <point/l1metric.h>
#include <point/squaredl1metric.h>
#include <point/l2metric.h>
#include <point/squaredl2metric.h>

using std::string;
using std::endl;
using std::vector;

namespace CluE {

EvaluationTest::EvaluationTest(std::string path, std::ostream& stream) : configuration(FileReader(path)), output(stream), testProxyEvaluation(false), testPartitionEvaluation(false), testCombinedEvaluation(false) {
	if(configuration.hasValue("main.input")) {
		string inputpath = configuration.getConfigString("main.input", false);
		input =  FileReader::readPointFile(inputpath);
	} else {
		InvalidConfiguration ic;
		ic.message = "Main.Input: No input file specified.";
		throw ic;
	}

	size_t n = input.size();
	for(size_t i = 0; i < n; i++) {
		lookupIndex[input[i]] = i;
	}

	if(configuration.hasValue("main.algorithm")) {
		string alg = configuration.getConfigString("main.algorithm");
		if(alg.compare("kmeans") == 0)
			algorithm = EvaluationType_KMeans;
		else if(alg.compare("kcenter") == 0)
			algorithm = EvaluationType_KCenter;
		else if (alg.compare("maxdiam") == 0)
			algorithm = EvaluationType_MaxDiam;
		else {
			InvalidConfiguration ic;
			ic.message = "Main.Algorithm: Invalid or unsupported algorithm specified.";
			throw ic;
		}
	}

	if(configuration.hasValue("main.measure")) {
		string m = configuration.getConfigString("main.measure");
		if(m.compare("l1") == 0)
			measure = new L1Metric();
		else if(m.compare("sl1") == 0)
			measure = new SquaredL1Metric();
		else if(m.compare("l2") == 0)
			measure = new L2Metric();
		else if(m.compare("sl2") == 0)
			measure = new SquaredL2Metric();
		else {
			InvalidConfiguration ic;
			ic.message = "Main.Measure: Invalid or unsupported measure specified.";
			throw ic;
		}
	}

	if(configuration.hasValue("main.solution")) {
		solution = configuration.getConfigUInt("main.solution");
	} else {
		solution = 0;
	}

	try {
		autoSetup();
	} catch(TestMessageException e) {
		std::cerr << "Auto setup failed: " << e.message << std::endl;
		throw;
	} catch(...) {
		std::cerr << "Auto setup failed due to an unexpected exception." << std::endl;
		throw;
	}
}

EvaluationTest::~EvaluationTest() {
	if(measure != 0)
		delete measure;

	size_t inputSize = input.size();
	for(size_t i = 0; i < inputSize; i++)
		delete input[i];
}

bool EvaluationTest::run() {
	bool success = true;

	output << "Algorithm: ";
	switch(algorithm) {
	case EvaluationType_KMeans: {
		evaluation = new KMeansEvaluator<Point>(measure);
		output << "KMeansEvaluator";
		break;
	}
	case EvaluationType_KCenter: {
		evaluation = new KCenterEvaluator<Point>(measure);
		output << "KCenterEvaluator";
		break;
	}
	case EvaluationType_MaxDiam: {
		evaluation = new MaxDiamEvaluator<Point>(measure);
		output << "MaxDiamEvaluator";
		break;
	}
	}
	if(configuration.hasValue("main.description"))
		output << " (" << configuration.getConfigString("main.description") << ")" << endl << endl;
	else
		output << endl << endl;

	if(testProxyEvaluation) {
		output << "--- ProxyEvaluation test ---" << endl;
		bool suc = false;
		try {
			suc = doProxyEvaluationTest();
		} catch(TestMessageException e) {
			std::cerr << "ProxyEvaluation test failed: " << e.message << std::endl;
		} catch(...) {
			std::cerr << "ProxyEvaluation test failed due to an unexpected exception." << std::endl;
		}
		printSuccess(suc);
		success &= suc;
	}

	if(testPartitionEvaluation) {
		output << "--- PartitionEvaluation test ---" << endl;
		bool suc = false;
		try {
			suc = doPartitionEvaluationTest();
		} catch(TestMessageException e) {
			std::cerr << "PartitionEvaluation test failed: " << e.message << std::endl;
		} catch(...) {
			std::cerr << "PartitionEvaluation test failed due to an unexpected exception." << std::endl;
		}
		printSuccess(suc);
		success &= suc;
	}

	if(testCombinedEvaluation) {
		output << "--- CombinedEvaluation test ---" << endl;
		bool suc = false;
		try {
			suc = doCombinedEvaluationTest();
		} catch(TestMessageException e) {
			std::cerr << "CombinedEvaluation test failed: " << e.message << std::endl;
		} catch(...) {
			std::cerr << "CombinedEvaluation test failed due to an unexpected exception." << std::endl;
		}
		printSuccess(suc);
		success &= suc;
	}

	return success;
}

void EvaluationTest::autoSetup() {
	if(configuration.getConfigFlag("main.testproxyevaluation")) {
		string proxVal = configuration.getConfigString("proxyevaluation.proxies");
		string costVal = configuration.getConfigString("proxyevaluation.proxycosts");
		vector<unsigned int> proxies = Helper::convertToUIntSequence(proxVal);
		vector<double> costs = Helper::convertToDoubleSequence(costVal);
		setupProxyEvaluationTest(proxies, costs);
		testProxyEvaluation = true;
	}

	if(configuration.getConfigFlag("main.testpartitionevaluation")) {
		unsigned int num = configuration.getConfigUInt("partitionevaluation.number");
		vector<vector<unsigned int> > partitions;
		for(unsigned int i = 0; i < num; i++) {
			std::stringstream key;
			key << "partitionevaluation.partition" << i;
			if(configuration.hasValue(key.str())) {
				string val = configuration.getConfigString(key.str());
				vector<unsigned int> partition = Helper::convertToUIntSequence(val);
				partitions.push_back(partition);
			} else {
				InvalidConfiguration ic;
				ic.message = key.str() + ": missing.";
				throw ic;
			}
		}
		string costVal = configuration.getConfigString("partitionevaluation.partitioncosts");
		vector<double> costs = Helper::convertToDoubleSequence(costVal);
		setupPartitionEvaluationTest(partitions, costs);
		testPartitionEvaluation = true;
	}

	if(configuration.getConfigFlag("main.testcombinedevaluation")) {
		string proxVal = configuration.getConfigString("combinedevaluation.proxies");
		vector<unsigned int> proxies = Helper::convertToUIntSequence(proxVal);

		unsigned int num = configuration.getConfigUInt("combinedevaluation.number");
		vector<vector<unsigned int> > partitions;
		for(unsigned int i = 0; i < num; i++) {
			std::stringstream key;
			key << "combinedevaluation.partition" << i;
			if(configuration.hasValue(key.str())) {
				string val = configuration.getConfigString(key.str());
				vector<unsigned int> partition = Helper::convertToUIntSequence(val);
				partitions.push_back(partition);
			} else {
				InvalidConfiguration ic;
				ic.message = key.str() + ": missing.";
				throw ic;
			}
		}
		string costVal = configuration.getConfigString("combinedevaluation.partitioncosts");
		vector<double> costs = Helper::convertToDoubleSequence(costVal);

		setupCombinedEvaluationTest(partitions, costs, proxies);
		testCombinedEvaluation = true;
	}
}

}
